/*$*********************************************************************\

Source File     : tpol.h

Description     : contains keyword definitions for transverse polarimeter

Comments        : first test version
                                                
Date            : 14.11.2001

Updates         : 

Authors         : Stefan Schmitt (sschmitt@mail.desy.de)

Remarks         : 

The time-stamps for the polarisation measurement are calculated by
averaging the start- and stop time of the relevant laser-on cycle.

Polarisation measurements are taken in cycles of approximately one
minute. There is not much use to poll for any information from the
TPOL more frequent than ONCE PER MINUTE.

\*************************************************************************/

/* version number of this header file*/
#define TPOLVER 101

/*----------------------------------------------------------------*\
  this structure is used to hold quantities which have an error 
  \*--------------------------------------------------------------*/
typedef struct {
  float x;
  float dx;
} Measured_t;

/*-----------------------------------------------------------------------*\
 beam polarisation and other quantities averaged over many (or all) bunches
   usually the following numbers correspond to the

   >>> average over 1 minute of data-taking <<<

   however, please check the value for onTime+offTime to get more
   accurate information

   The time-stamp corresponds to the middle of the laser-on interval,
   since this period is sensitive to the beam polarisation.

   TPOLData_f.pol    transverse beam polarisation all bunches
   TPOLData_f.ncoll  transverse beam polarisation pilot bunches
   TPOLData_f.coll   transverse beam polarisation colliding bunches

   TPOLData_f.lumi   luminosity (in percent)
   TPOLData_f.onTime  total measuring time with laser on in seconds
   TPOLData_f.offTime total measuring time with laser off in seconds

   TPOLData_f.beamY   vertical position of the Compton beam in mm
   TPOLData_f.beamSY  vertical size of the Compton beam ellipsis in mm

\*-----------------------------------------------------------------------*/

typedef struct {
  Measured_t pol;
  Measured_t ncoll;
  Measured_t coll;
  Measured_t lumi;
  float onTime;
  float offTime;
  Measured_t beamY;
  Measured_t beamSY;
} TPOLData_t;

TPOLData_t TPOLData_f;

NETKEY TPOLData={"TPOLDAQ","TPOLDATA",CF_FLOAT,
		 sizeof(TPOLData_t)/sizeof(float),
		 0,RC_INI,T_INIT,ST_INI,TPOLVER,(void*)&TPOLData_f};

/*-----------------------------------------------------------------------*\
 beam polarisation of single bunches

   these polarisation numbers are calculated as an

   >>> average over 5 minutes of data taking <<<

   The time-stamp corresponds to the mean time of all
   contributing laser-on measurement intervals

   Note that the exact measurement time might vary, but is guranteed to be
   larger than 300s.

   TPOLBunch_f[0] transverse beam polarisation of first bunch with error
   TPOLBunch_f[1] transverse beam polarisation of second bunch with error
     ...
   TPOLBunch_f[219] transverse beam polarisation of bunch number 220
   
\*-----------------------------------------------------------------------*/

#define N_TPOLBunch 220
Measured_t TPOLBunch_f[N_TPOLBunch];

NETKEY
TPOLBunch={"TPOLDAQ","TPOLBUNCH",CF_FLOAT,
	   N_TPOLBunch*sizeof(Measured_t)/sizeof(float),
	   0,RC_INI,T_INIT,ST_INI,TPOLVER,(void*)TPOLBunch_f };

/*-----------------------------------------------------------------------*\
 transverse polarimeter status information
  isDAQup      0 if the DAQ is down
               1 if the DAQ programs running

  isDAQactive  2 online polarisation measurement is active
               1 online DAQ is taking polarisation data, but the data
                  might be compromised (because some tuning is going on)
               0 online polarisation measurement not active
              -1 online DAQ is claiming to take polarisation data,
                  but something is wrong
              -2 online DAQ is taking polarisation data,
                  but someone is doing tests 
              -3 online DAQ is taking polarisation data,
              -9 online DAQ is down
  
  isAutopilotActive
               0 if the autopilot is not active (or the DAQ is down)
               1 if the autopilot is partially enabled
               2 if the autopilot has full control
               <0 if someone is doing tests:
                   -1 test flag active with autopilot disabled
                   -2 test flag active with autopilot partially enabled
                   -3 test flag active with autopilot fully enabled
               -9 if the DAQ is down

  runNumber      0 if no data is recorded permanently
                >0 the run-number of the data which is currently being recorded
                   NOTE!!! a run number may be assigned even if no polarisation
                           data is collected (e.g. for laser-measurements)
                -9 if the DAQ is down
    
  runType         0  if no data is being recorded permanently
                  1  beam polarisation measurement
                  2  luminosity optimisation
                  3  beam tilt measurement
                  4  beam focus measurement
                  5  table centering
                  6  detector calibration
                 21  light polarisation measurement
                 22  mirror optimisation
                101  test run
                 -1  unknown
                 -9  the DAQ is down
    
  collimators     0 if the collimators are closed
                         -> TPOL is ready for injection or beam-dump
                  1 if the collimators are open
                         -> TPOL is ready to take data:
                 <0 if the collimators are in a special state
                    -1 the collimators are about to be opened
                    -2 the collimators are about to be closed
                    -7 the status of the collimators is undefined

                 -8 the status of the collimators is not known
                 -9 the DAQ is down

\*-----------------------------------------------------------------------*/

typedef struct {
  int isDAQup;
  int isDAQactive;
  int isAutopilotActive;
  int runNumber;
  int runType;
  int collimators;
} TPOLStatus_t;

TPOLStatus_t TPOLStatus_i;
NETKEY TPOLStatus={"TPOLDAQ","TPOLSTATUS",CF_INT4,
		   sizeof(TPOLStatus_t)/sizeof(float),
		   0,RC_INI,T_INIT,ST_INI,TPOLVER,(void*)&TPOLStatus_i };

