#include "cprolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#ifdef SVR5
#include <sys/wait.h>
#endif
#include "errors.h"
#include "rpctypes.h"
#include "unixcmd.h"

/*
	Please use SystemDelay() in place of sleep()

	callback functions must have the prototype:

	void function(int id,int cc);

	where id is the subscription id and cc is the completion code
	(or interim status)

	mode is the data acquisition mode.
	Please note the possible acquistion modes:

		CM_CANCEL      cancels a data link
		CM_SINGLE      requests a single (asynchronous) transaction
		CM_REFRESH     requests that data be sent only upon change
									 (and at 1 minute intervals - hearbeat)
		CM_POLL        requests that data be sent at the given polling rate
		CM_REGISTER    requests that the link be registered but not activated
									 (normally for internal use only)

*/

extern char *erlst[];
extern int NGdebug;
extern time_t RPCtimestamp;

char *getFormatAsString(int fmt)
{
  static char fmtstr[32];
  switch (fmt)
  {
    case CF_BYTE:   strcpy(fmtstr,"BYTE"); break;
    case CF_SHORT:  strcpy(fmtstr,"short"); break;
    case CF_LONG:   strcpy(fmtstr,"long"); break;
    case CF_FLOAT:  strcpy(fmtstr,"float"); break;
    case CF_DOUBLE: strcpy(fmtstr,"double"); break;
    case CF_TEXT:   strcpy(fmtstr,"char"); break;
    case CF_STRUCT: strcpy(fmtstr,"struct (user)"); break;
    case CF_CHAR8:  strcpy(fmtstr,"char*8"); break;
    case CF_CHAR16:  strcpy(fmtstr,"char*16"); break;
    case CF_CHAR32:  strcpy(fmtstr,"char*32"); break;
    case CF_CHAR48:  strcpy(fmtstr,"char*48"); break;
    case CF_CHAR16FI:  strcpy(fmtstr,"char*16-float-int"); break;
    case CF_SHTINT:  strcpy(fmtstr,"short-int"); break;
    case CF_LNGINT:  strcpy(fmtstr,"long-int"); break;
    case CF_FLTINT:  strcpy(fmtstr,"float-int"); break;
    case CF_DBLINT:  strcpy(fmtstr,"double-int"); break;
    case CF_CHAR8I:  strcpy(fmtstr,"char8-int"); break;
    case CF_CHAR16I:  strcpy(fmtstr,"char16-int"); break;
    case CF_CHAR32I:  strcpy(fmtstr,"char32-int"); break;
    case CF_CHAR48I:  strcpy(fmtstr,"char48-int"); break;
    case CF_TDS:  strcpy(fmtstr,"time-float-int"); break;
    default:  strcpy(fmtstr,"undefined"); break;
  }
  return fmtstr;
}

int propID,devID,stateID;
#define NPROPSMAX 50
typedef struct
{
  char name[12];
  char numModules[5];
  char format[3];
  char size[10];
  char desc[50];
} PROPSTRUCT;
PROPSTRUCT srvPropsTextQuery[NPROPSMAX]; /* Legacy query for OLD servers */

PropertyQueryStruct srvProps[NPROPSMAX]; /* Modern query structure */

#define NDEVICESMAX 300

NAME16 devTags[NDEVICESMAX];

#define NTAGSMAX 300

NAME16 srvTags[NTAGSMAX];

int getTagList(char *taglist)
{
  int cc,i;
  char name[17];
  DTYPE dout;

  /* Get Tags from name server (synchronous call) */
  dout.dFormat = CF_STRING16;
  dout.dArrayLength = NTAGSMAX;
  dout.data.vptr = srvTags;
  if ((cc=ExecLink("NAMEQRY",taglist,&dout,NULL,CA_READ)) != 0)
  {
    printf("%s: %s\n>",taglist,RPCLastErrorString);
    exit(1);
  }
  printf("Available %s names:\n>\n>",taglist);
  for (i=0; i<NTAGSMAX; i++)
  {
    if (srvTags[i].name[0] == 0) break;
    strncpy(name,srvTags[i].name,16); name[16] = 0;
    printf((i+1)%10 ? "%16s " : "%16s \n>",name);
  }
  printf("\n>\n>");
  return i;
}

typedef struct
{
  short lumi;
  short Prunning;
  short Pbeam;
  short Pramping;
  short Erunning;
  short Ebeam;
  short Eramping;
} STATE;

STATE heraState;
#define NSTATEPARAMS 7

#define ccerr(i) { cc = i; goto err; }
#define callout(id,cc) {\
  if (cc) {\
     printf("LINK %d error: %s\n>",id,RPCLastErrorString);\
     return;\
  }\
}
void showstate(int id,int cc)
{
  int i;
  callout(id,cc);
  /* this is just an example, print this out at your discretion... */
  printf("State of the HERA Machine: %s>",ctime(&RPCtimestamp));
  printf("Proton  active  :\t%s\n>",heraState.Prunning ? "YES" : "NO");
  printf("Proton  beam    :\t%s\n>",heraState.Pbeam    ? "YES" : "NO");
  printf("Proton  ramping :\t%s\n>",heraState.Pramping ? "YES" : "NO");
  printf("Electron active :\t%s\n>",heraState.Erunning ? "YES" : "NO");
  printf("Electron beam   :\t%s\n>",heraState.Ebeam    ? "YES" : "NO");
  printf("Electron ramping:\t%s\n>",heraState.Eramping ? "YES" : "NO");
  printf("Lumi run        :\t%s\n>",heraState.lumi     ? "YES" : "NO");

  printf("\n>");
}
void showpropsLegacy(int id,int cc)
{
  int i,nummod,siz;
  char *c,scratch[16],desc[50];
  callout(id,cc);
  printf("Queried Properties:\n>");

  for (i=0; i<NPROPSMAX; i++)
  {
    if ((c=strchr(srvPropsTextQuery[i].name,' ')) != NULL) *c = 0;
    if (srvPropsTextQuery[i].name[0] == 0) break;
    strncpy(scratch,srvPropsTextQuery[i].numModules,5); scratch[5] = 0;
    nummod = atoi(scratch);
    strncpy(scratch,srvPropsTextQuery[i].size,10); scratch[10] = 0;
    siz = atoi(scratch);
    strncpy(scratch,srvPropsTextQuery[i].format,3); scratch[3] = 0;
    strncpy(desc,srvPropsTextQuery[i].desc,50); desc[49] = 0;
    printf("%12s (%d modules, maximum %d %s(s)) : %s\n>",
	   srvPropsTextQuery[i].name,nummod,siz,
	   getFormatAsString(atoi(scratch)),desc);
  }
  printf("\n>\n>");
}
void showprops(int id,int cc)
{
  int i;
  callout(id,cc);
  printf("Queried Properties:\n>");

  for (i=0; i<NPROPSMAX; i++)
  {
    if (srvProps[i].prpName[0] == 0) break;
    printf("%32s (maximum %d %s(s)) : %s\n>",
	   srvProps[i].prpName,srvProps[i].prpSize,
	   getFormatAsString(LFMT(srvProps[i].prpFormat)),
           srvProps[i].prpDescription);
  }
  printf("\n>\n>");
}
void showdevices(int id,int cc)
{
  int i;
  callout(id,cc);
  printf("Queried Devices:\n>");

  for (i=0; i<NDEVICESMAX; i++)
  {
    if (devTags[i].name[0] == 0) break;
    printf((i+1)%2 ? "%32s " : "%32s\n>",devTags[i].name);
  }
  printf("\n>\n>");
}
void displayInfo(char *srvnam)
{
  DTYPE dout;
  int i,cc = 0;
  short num;

  /* use synchronous calls here in this simple example:  */
  
  dout.dFormat = CF_SHORT;
  dout.dArrayLength = 1;
  dout.data.sptr = &num;
  if ((cc=ExecLink(srvnam,"NPROPS",&dout,NULL,CA_READ)) != 0) ccerr(cc);
  
  if (num > NPROPSMAX) num = NPROPSMAX; // truncate to what I can buffer
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = num*PROPERTYQUERYSTRUCT_SIZE;
  dout.data.vptr = srvProps;
  if ((cc=ExecLink(srvnam,"PROPS",&dout,NULL,CA_READ)) != 0) ccerr(cc);
  for (i=0; i<num; i++)
  {
    if (srvProps[i].prpName[0] == 0) break;
    printf("%32s (maximum %d %s(s)) : %s\n>",
	   srvProps[i].prpName,srvProps[i].prpSize,
	   getFormatAsString(LFMT(srvProps[i].prpFormat)),
           srvProps[i].prpDescription);
  }
  printf("\n>\n>");

  dout.dFormat = CF_SHORT;
  dout.dArrayLength = 1;
  dout.data.sptr = &num;
  if ((cc=ExecLink(srvnam,"NDEVICES",&dout,NULL,CA_READ)) != 0) ccerr(cc);

  if (num > NDEVICESMAX) num = NDEVICESMAX;
  dout.dFormat = CF_NAME16;
  dout.dArrayLength = num;
  dout.data.vptr = devTags;
  if ((cc=ExecLink(srvnam,"DEVICES",&dout,NULL,CA_READ)) != 0) ccerr(cc);
  for (i=0; i<num; i++)
  {
    if (devTags[i].name[0] == 0) break;
    printf((i+1)%4 ? "%16s " : "%16s\n>",devTags[i].name);
  }
  printf("\n>\n>");

err:
  if (cc)
  {
    printf("%s: <%d> %s\n>",srvnam,cc,RPCLastErrorString);
    return;
  }
  return;
}
int main(int argc, char *argv[])
{
  int i,ntags,nalmtags,nfecs,cycle=0;
  char *srvnam;
  DTYPE dout;
  switch (argc)
  {
    case 2:
      NGdebug = atoi(argv[1]);
      break;
    default:
      NGdebug = 0;
  }

  nfecs    = getTagList("FECS");
  ntags    = getTagList("TAGS");
  nalmtags = getTagList("ALMTAGS");

  /* "REFRESH link to get the HERA state information (Asynchronous) */
  dout.dFormat = CF_SHORT;
  dout.dArrayLength = NSTATEPARAMS;
  dout.data.sptr = (UINT16 *)&heraState;
  if ((stateID=AttachLink("GLOBQRY","STATE",&dout,NULL,CA_READ,
			  1000,showstate,CM_REFRESH)) < 0)
  {
    printf("STATE: %s\n>",RPCLastErrorString);
    exit(1);
  }

  printf("'set CommandString = <srvName>' to query properties\n>");
  for (;;)
  {
    SystemCycle(FALSE);
    if (strlen(CommandString))
    {
      for (srvnam=CommandString; *srvnam == ' '; srvnam++); /* trim away leading blanks */
      displayInfo(srvnam);
      CommandString[0] = 0;
      printf("'set CommandString = <srvName>' to query properties\n>");
    }
    if (NGdebug > 1) printf("loop cycle: %d (1-second intervals)\n",cycle++);
    SystemDelay(1);
  }

  return 0;
}
