package TINE;
import java.util.Date;
/** 
 * TLink.java
 *
 * @brief User interface calls to access tine data links.
 *  
 * The methods presented here deal with linking local application data with
 * remote data sources.  Primarily you will need to create a TLink object,
 * passing the necessary identifying parameters of the remote server to which 
 * the link is to be established.  When the TLink object is instantiated, 
 * you will then need to retrieve the data using one of the retrieval 
 * methods, e.g. execute() for synchronous acquistion, or attach() for  
 * asynchronous acquisition, or receive() for the acquisition of network
 * globals.
 */
public final class TLink
{
  String cntName;
  String expName;
  String devName;
  String devProperty;
  TDataType dOutput;
  TDataType dInput;
  TDataType dError;
  private static int maxOutputBufferSize = 0xffff;
  private String arrayDelimiter;
  short devAccess;
  short devTimeout;
  int defaultTimeout = 800;  // milliseconds
  short srvId; // for network subscriptions
  boolean active;
  boolean terminate;
  boolean needsNotification = false;
  boolean isInsideCallback = false;
  boolean useErrValue = false;
  boolean useErrObject = false;
  boolean adjustDefaultValues = false;
  boolean retryOnTimeoutError = true;
  TModAddress modAddress;
  TContract con;
  TSubscription sub;
  TPublishedData phs;
  Theader hdr;
  int linkStatus;
  int linkCounter;
  boolean linkStale;
  double linkErrValue;
  String linkErrString;
  byte[] linkStatusStringBuffer = new byte[32];
  TCallback tcb;
  int callbackId;
  TLinkCallback tlcb;
  /**
   * @brief Returns the current status for this link
   * 
   * @return The Link status as tine error code
   */
  public int getLinkStatus()
  {
    return linkStatus;
  }
  /**
   * @brief Returns the context of this link
   * 
   * @return The Context of the current link
   */
  public String getContext()
  {
    return cntName;
  }
  /**
   * @brief Returns the device server addressed in this link
   * 
   * @return The device server addressed in the current link
   */
  public String getDeviceServer()
  {
    return expName;
  }
  /**
   * @brief Returns the device (module) name addressed in this link
   * 
   * @return The device (module) name addressed in the current link
   */
  public String getDeviceName()
  {
    return devName;
  }
  /**
   * @brief Returns the device property addressed in this link
   * 
   * @return The device property addressed in the current link
   */
  public String getProperty()
  {
    return devProperty;
  }
  /**
   * @brief Returns the fully qualified device name addressed in this link
   * 
   * @return The the fully qualified device name addressed in the current link
   */
  public String getFullDeviceName()
  {
    return new String("/"+cntName+"/"+expName+"/"+devName);
  }
  /**
   * @brief Returns the output data object as a TDataType
   * 
   * @return The output data type 
   */
  public TDataType getOutputDataObject()
  {
    return dOutput;
  }
  /**
   * @brief Returns the output data input as a TDataType
   * 
   * @return The input data type 
   */
  public TDataType getInputDataObject()
  {
    return dInput;
  }
  /**
   * @brief The last link timestamp
   * 
   * @return The last link timestamp as a UTC java long value (milliseconds
   * since 1.1.1970
   */
  public long getLastTimeStamp()
  {
    return (long)(dOutput.timestamp) * 1000 + (long)(dOutput.timestampMSEC);
  }
  /**
   * @brief The last link status in plain text
   * 
   * @return The last link status in plain text
   */
  public String getLastError()
  {
	  return getError(linkStatus);
  }      
  /**
   * @brief The specified error code in plain text
   * 
   * @param code Is the tine error code for which the error text is desired
   * 
   * @return The specified error code in plain text
   */
  public String getError(int code) 
  {
    if (code >= 512)
      return new String(linkStatusStringBuffer).trim();
    else
	    return TErrorList.getErrorString(code);
  }
  Date  date = new Date();
  /**
   * @brief Specifies an error value to fill in associated data buffers 
   * in case of link error.
   * 
   * @param errValue Is the error value to use in case of link errors
   * 
   * 0 if successful
   */
  public int setErrorValue(double errValue)
  {
    if (sub.mode == TMode.CM_CANCEL) return TErrorList.ask_error;
    linkErrValue = errValue;
    useErrValue = true;
    return 0;
  }
  /**
   * @brief Specifies an error string to fill in associated data buffers 
   * in case of link error.
   * 
   * @param errString Is the string to use in case of link errors
   * 
   * 0 if successful
   */
  public int setErrorValue(String errString)
  {
    if (sub.mode == TMode.CM_CANCEL) return TErrorList.ask_error;
    linkErrString = errString;
    useErrValue = true;
    return 0;
  }
  private int TCreateLink(String devname, String devproperty, TDataType dout, TDataType din, short devaccess)
  {
    String fullname = new String(devname);
    active = true;
    if (dout != null) dOutput = dout; else dOutput = new TDataType();
    if (din != null) dInput = din; else dInput = new TDataType();
    if (arrayDelimiter != null) dOutput.setArrayDelimiter(arrayDelimiter);
    
    int indx = fullname.indexOf('/');
    tf = TFactory.getInstance();
    linkId = tf.registerLink(this);
    if (linkId < 0) return TErrorList.out_of_client_memory;
    cntName = "DEFAULT";
    if (indx == 0)  // context given
    {
      fullname = fullname.substring(1);
      indx = fullname.indexOf('/');
      if (indx != -1) 
      {
        cntName = fullname.substring(0,indx);
        fullname = fullname.substring(indx+1);
        indx = fullname.indexOf('/');
      }
    }
    if (indx == -1)
    {
      expName = fullname;
      devName = "";
    }
    else
    {
      expName = fullname.substring(0,indx);
      devName = fullname.substring(indx+1);
    }
    fullname = indx == -1 ? "" : fullname.substring(indx+1);
    devProperty = new String(devproperty);
    devAccess = devaccess;
    devTimeout = (short)1000;
    if (expName.compareTo("GLOBAL") != 0)
    {
      modAddress = new TModAddress(tf,expName,cntName);
      if (modAddress.srvAddress == null && tf.debugLevel > 0)
        System.out.println("cannot resolve server address for " + "/" + cntName + "/" + expName);
    }
    con = new TContract(this);
    sub = new TSubscription(con,this); // going out ...
    hdr = new Theader(System.getProperty("user.name"));
    String defaultTimeoutStr;
    if ((defaultTimeoutStr = System.getProperty("default.timeout")) != null)
    {
      defaultTimeout = Integer.parseInt(defaultTimeoutStr);
    }
    tf.linkReady(this);
    return 0;
  }

  /** 
   * @brief Defines a tine Data link
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param devname is the fully specified device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;
   * or &lt;server&gt;/&lt;device&gt;
   * @param devproperty is the desired property
   * @param dout is a reference to the output data object (which will receive
   * the data from the server)
   * @param din is a reference to the input data object (which will send data
   * to the server)
   * @param devaccess is the requested control access mode (TMode.CA_READ, TMode.CA_WRITE
   * etc.)
   *
   * @return A valid TLink instance if successful or null if not
   *
   * \b Example:
   * 
   * \include eg_TLink1.java
   *
   * @see TINE.TLink#attach
   * @see TINE.TLink#execute
   * @see TINE.TDataType#TDataType 
   */
  public TLink(String devname, String devproperty, TDataType dout, TDataType din, short devaccess)
  {
    TCreateLink(devname,devproperty,dout,din,devaccess);
  }
  /** 
   * @brief Defines a tine Data link
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param fullDeviceNameAndProperty is the fully specified, property-appended
   * device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   * or &lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   * @param dout is a reference to the output data object (which will receive
   * the data from the server)
   * @param din is a reference to the input data object (which will send data
   * to the server)
   * @param devaccess is the requested control access mode (TMode.CA_READ, TMode.CA_WRITE
   * etc.)
   *
   * @return A valid TLink instance if successful or null if not
   *
   * \b Example:
   * 
   * \include eg_TLink2.java
   *
   * @see TINE.TLink#attach
   * @see TINE.TLink#execute
   * @see TINE.TDataType#TDataType 
   */
  public TLink(String fullDeviceNameAndProperty, TDataType dout, TDataType din, short devaccess)
  {
    String devname, devproperty;
    int cnt = 1;
    int indx = fullDeviceNameAndProperty.indexOf('/');
    if (indx == 0) cnt = 3;
    for (int i=0; i<cnt; i++)
    {
      indx +=  fullDeviceNameAndProperty.substring(indx+1).indexOf('/') + 1;
    }
    devname = new String(fullDeviceNameAndProperty.substring(0,indx));
    devproperty = new String(fullDeviceNameAndProperty.substring(indx+1));
    
    TCreateLink(devname,devproperty,dout,din,devaccess);
  }

  /** 
   * @brief Defines a read-only tine Data link with default behavior.
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param devname is the fully specified device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;
   * or &lt;server&gt;/&lt;device&gt;
   * @param devproperty is the desired property
   *
   * @return A valid TLink instance if successful or null if not
   *
   * \b Example:
   * 
   * \include eg_TLink3.java
   *
   * @see TINE.TLink#attach
   * @see TINE.TLink#execute
   */
  public TLink(String devname, String devproperty)
  {
    TDataType dout = new TDataType(this.maxOutputBufferSize,TFormat.CF_DEFAULT);
    TDataType din = new TDataType();
    TCreateLink(devname,devproperty,dout,din,TAccess.CA_READ);
  }

  /** 
   * @brief Defines a read-only tine Data link with default behavior.
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param fullDeviceNameAndProperty is the fully specified, property-appended
   * device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   * or &lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   *
   * @return A valid TLink instance if successful or null if not
   *
   * \b Example:
   * 
   * \include eg_TLink4.java
   *
   * @see TINE.TLink#attach
   * @see TINE.TLink#execute
   */
  public TLink(String fullDeviceNameAndProperty)
  {
    String devname, devproperty;
    int cnt = 1;
    int indx = fullDeviceNameAndProperty.indexOf('/');
    if (indx == 0) cnt = 3;
    for (int i=0; i<cnt; i++)
    {
      indx +=  fullDeviceNameAndProperty.substring(indx+1).indexOf('/') + 1;
    }
    devname = new String(fullDeviceNameAndProperty.substring(0,indx));
    devproperty = new String(fullDeviceNameAndProperty.substring(indx+1));
    
    TDataType dout = new TDataType(this.maxOutputBufferSize,TFormat.CF_DEFAULT);
    TDataType din = new TDataType();
    TCreateLink(devname,devproperty,dout,din,TAccess.CA_READ);
  }

  /** 
   * @brief Defines a tine GLobal Data link
   * 
   * This defines a tine Data Link and assumes that the requested parameter
   * is a network global, available on the tine multicast group.
   * This method does not establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param globalKeyword is the desired network global parameter
   * @param dout is a reference to the output data object (which will receive
   * the network global data).
   *
   * @return A valid TLink instance if successful or null if not
   *
   * \b Example:
   * 
   * \include eg_TLink5.java
   *
   * @see TINE.TLink#attach
   * @see TINE.TLink#receive
   * @see TINE.TDataType#TDataType 
   */
  public TLink(String globalKeyword, TDataType dout)
  {
    TCreateLink("GLOBAL",globalKeyword,dout,null,TAccess.CA_READ);
  }

  /** 
   * @brief Attaches the data sets to the created link at the default polling rate (1000 msec)
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   *  
   * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
   * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
   * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TCallback() class interface.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_attach1.java
   *
   * @see TINE.TLink#TLink
   * @see TINE.TLink#TLinkCallback
   * @see TINE.TLink#execute
   */
  public synchronized int attach(short mode,TCallback f)
  {
    return attach(mode,f,1000,-1);
  }
 
  /**
   * Stop thread execution until this link has been signalled.
   */
	   
  public synchronized int Twait()
  {
	 if (sub.id < 0) return -1;
	 try
	 {
		while (!this.terminate)
		{
		  wait();
//		  if (!this.terminate && 
//		       this.needsNotification &&
//		        tcb != null) tcb.callback(sub.id,linkStatus);
//		  this.needsNotification = false;
		}
	 }
	 catch (InterruptedException e) {}
//	 sub.mode = TMode.CM_CANCEL;
	 return -1;
  }            
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for 1000 msec
   * or until the call completes.
   *  
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpretted using getError() or getLastError().
   *
   * \b Example:
   * 
   * \include eg_execute1.java
   *
   * @see TINE.TLink#TLink
   * @see TINE.TLink#attach
   */
  public synchronized int execute()
  {
	 return execute(1000);
  }                    
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   * @param appendMode will be appended to the tine Control Mode, which is 
   * TMode.CM_SINGLE. (e.g. TMode.CM_CONNECT).
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpretted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute2.java
   *
   * @see TINE.TLink#TLink
   * @see TINE.TLink#attach
   */
  public synchronized int execute(int timeout, short appendMode)
  {
    return execute(timeout,appendMode,false);
  }                    
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   * @param retryOnError if 'true' will instuct the tine engine to issue
   * an automatic retry following a connection timeout. 
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpretted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute3.java
   *
   * @see TINE.TLink#TLink
   * @see TINE.TLink#attach
   */
  public synchronized int execute(int timeout, boolean retryOnError)
  {
    return execute(timeout,(short)0,retryOnError);
  }                    
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   * @param appendMode will be appended to the tine Control Mode, which is 
   * TMode.CM_SINGLE. (e.g. TMode.CM_CONNECT).
   * @param retryOnError if 'true' will instuct the tine engine to issue
   * an automatic retry following a connection timeout. 
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpretted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute4.java
   *
   * @see TINE.TLink#TLink
   * @see TINE.TLink#attach
   */
  public synchronized int execute(int timeout, short appendMode,boolean retryOnError)
  {
    Date dt;
    if (this.linkId < 0)
    {
      this.linkStatus = TErrorList.out_of_client_memory;
      return -1;
    }
    if (timeout < 100) timeout = 100;
    devTimeout = (short)timeout;
    devAccess |= TAccess.CA_SYNC;
    this.sub.mode = TMode.CM_SINGLE;
    if ((appendMode & TMode.CM_CONNECT) == TMode.CM_CONNECT) this.sub.mode |= appendMode;
    retryOnTimeoutError = retryOnError;
    linkStatus = TErrorList.link_not_open;

    if (tf.establishLink(this) != 0)
    {
      if (tf.debugLevel > 0) System.out.println("can't establish link");
      return -1;
    }
   
    long ltime = System.currentTimeMillis();
    dOutput.timestamp = (int)(ltime/1000);
    this.sub.linkStartTime = ltime;
    this.sub.linkLastTime = ltime;
      
    try
    {
      wait();
      this.active = false;
      return linkStatus;
    }
    catch (InterruptedException e) {}
    return linkStatus;
  }                    
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpretted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute5.java
   *
   * @see TINE.TLink#TLink
   * @see TINE.TLink#attach
   */
  public synchronized int execute(int timeout)
  {
   return execute(timeout,(short)0);
  }                            
  /**
   * Terminates an active link
   */
  public synchronized int cancel()
  {
	  return tf.cancel(this);
  }              

  public static void main (String[] args)
  {
	  int cc;
/*
    final class TEST1
    {
      float[] fval = new float[3];
      int[] ival = new int[2];
      short sval1;
      short sval2;
      char[] s = new char[32];
      int sizeInBytes = 3 * 4 + 2 * 4 + 2 * 2 + 32;
      byte[] hByteArray = new byte[sizeInBytes];
      TEST1()
      {
        System.out.println("TEST1 buffer : " + hByteArray.length + " bytes");
      }
      public void toStruct(byte[] data)
      {
        try
        {
          ByteArrayInputStream dis = new ByteArrayInputStream(data);
          DataInputStream ds = new DataInputStream(dis);
          ds.read(data, 0, 4*3);
          byte[] b = new byte[32];
          for (int i=0; i<3; i++)
          { 
            System.arraycopy(data, i*4, b, 0, 4);
            fval[i] = Swap.Float(b);
          }
          for (int i=0; i<2; i++) ival[i] = Swap.Long(ds.readInt());
          sval1 = Swap.Short(ds.readShort());
          sval2 = Swap.Short(ds.readShort());
          ds.read(b,0,32);
          System.arraycopy(b, 0, s, 0, 32);
          ds.close(); dis.close();
        }
        catch (IOException e)
        {
          e.printStackTrace();
          System.exit(1);
        }   
      }
    }
    
    TStruct.fill("TEST1",TFormat.CF_FLOAT,3);
    TStruct.fill("TEST1",TFormat.CF_LONG,2);
    TStruct.fill("TEST1",TFormat.CF_SHORT,1);
    TStruct.fill("TEST1",TFormat.CF_SHORT,1);
    TStruct.fill("TEST1",TFormat.CF_TEXT,32);
    TStruct.fill("TEST1",TFormat.CF_NULL,1);

    TEST1 test1 = new TEST1();
    TDataType struct_dout = new TDataType(test1.hByteArray,"TEST1");
    TDataType struct_din = new TDataType();
    TLink struct1 = new TLink("/TEST/SINE/SINEDEV_0","STRUCTTEST",struct_dout,struct_din,TAccess.CA_READ);
    int rc = struct1.execute(500,true);
    System.out.println("STRUCTTEST returned " + rc);
    System.out.print(TStruct.toString("TEST1", test1.hByteArray));

    System.exit(0);
*/
    int[] sinbuf = new int[100];
    int[] echobuf = new int[512];
    int[] echoin = new int[512];
    StringBuffer strbuf = new StringBuffer(64);
//    char[] strbuf = new char[64];
    for (int i=0; i<512; i++) echoin[i] = i;
    for (int i=0; i<512; i++) echobuf[i] = 10;
//    TDataType dout = new TDataType(sinbuf);
    NAME16[] echostr = new NAME16[50];
//    TDataType dout = new TDataType(echobuf);
    TDataType dout = new TDataType(echostr);
    TDataType din = new TDataType(echoin);
//    TDataType din = new TDataType();

    TLink sin2 = new TLink("/TEST/SINE/SINEDEV_0","ECHOX",dout,din,TAccess.CA_READ);
    System.out.println("First call");
    sin2.setErrorValue(113);
    sin2.setErrorValue("Uh-Oh");
    if ((cc=sin2.execute()) != 0)  // synchronous access ...
    {
      String msg = "err: " + cc;
      System.out.println(msg);
    }
    
    System.out.println("echo -> " + dout.toString());
    sin2.cancel();
  for (int i=0; i<48; i += 8)
  {
    System.out.println(echostr[i] + " " + echostr[i+1] + " " + echostr[i+2] + " " + echostr[i+3] + " " +
    echostr[i+4] + " " + echostr[i+5] + " " + echostr[i+6] + " " + echostr[i+7] + " ");
  }
//    for (int i=0; i<512; i += 8)
//    {
//      System.out.println(echobuf[i] + " " + echobuf[i+1] + " " + echobuf[i+2] + " " + echobuf[i+3] + " " +
//                         echobuf[i+4] + " " + echobuf[i+5] + " " + echobuf[i+6] + " " + echobuf[i+7] + " ");
//    }
    System.exit(0);

    dout = new TDataType(strbuf);
    din = new TDataType();
    TLink sin1 = new TLink("/TEST/SINE/SINEDEV_0","SRVVERSION",dout,din,TAccess.CA_READ);
    System.out.println("First call");
    if ((cc=sin1.execute()) == 0)  // synchronous access ...
    {
      String sinout = "ver: " + strbuf.toString().trim(); //(new String(strbuf)).toString().trim();
//      String sinout = "sin: " + sinbuf[0] + ", " + sinbuf[1] + ", " + sinbuf[2];
      System.out.println(sinout);
//      sinout = "sin: " + sinbuf[97] + ", " + sinbuf[98] + ", " + sinbuf[99];
//      System.out.println(sinout);
    }
    else 
    {
      String msg = "err: " + cc;
      System.out.println(msg);
    }
    sin1.cancel();
    dout = new TDataType(sinbuf);
    sin1 = new TLink("/TEST/SINE/SINEDEV_1","TEST",dout,din,TAccess.CA_READ);
    System.out.println("First call");
    if ((cc=sin1.execute()) == 0)  // synchronous access ...
    {
      String sinout = "sin: " + sinbuf[0] + ", " + sinbuf[1] + ", " + sinbuf[2];
      System.out.println(sinout);
      sinout = "sin: " + sinbuf[97] + ", " + sinbuf[98] + ", " + sinbuf[99];
      System.out.println(sinout);
    }
    else 
    {
      String msg = "err: " + cc;
      System.out.println(msg);
    }
    sin1.cancel();    

    PropertyQueryEx[] pqx1 = TQuery.getDevicePropertyInformation("HERA","ADDA","#0","PROP10");
    if (pqx1 != null)
      for (int i=0; i<pqx1.length; i++) 
      {
        System.out.println("Name:" + pqx1[i].prpName);
        System.out.println("Description:" + pqx1[i].prpDescription);
        System.out.println("Calling info:" + pqx1[i].prpSize + " " + TFormat.toString(pqx1[i].prpFormat));
        System.out.println("Short history depth: " + pqx1[i].prpHistoryDepthShort);
        System.out.println("Long history depth: " + pqx1[i].prpHistoryDepthLong);
        System.out.println("Range:" + pqx1[i].prpMinValue + " to " + pqx1[i].prpMaxValue + " " + pqx1[i].prpUnits);
      }
    pqx1 = TQuery.getDevicePropertyInformation("HERA","ADDA","#0","PROP26");
    if (pqx1 != null)
      for (int i=0; i<pqx1.length; i++) 
      {
        System.out.println("Name:" + pqx1[i].prpName);
        System.out.println("Description:" + pqx1[i].prpDescription);
        System.out.println("Calling info:" + pqx1[i].prpSize + " " + TFormat.toString(pqx1[i].prpFormat));
        System.out.println("Short history depth: " + pqx1[i].prpHistoryDepthShort);
        System.out.println("Long history depth: " + pqx1[i].prpHistoryDepthLong);
        System.out.println("Range:" + pqx1[i].prpMinValue + " to " + pqx1[i].prpMaxValue + " " + pqx1[i].prpUnits);
      }
    //System.exit(0);

/*
//    TLink sinx = new TLink("/TEST/SINE/SINEDEV_0/TEST");
    USTRING[] egu = new USTRING[1];
    egu[0] = new USTRING();
    TDataType egudout = new TDataType(egu);
//    TDataType din = new TDataType(echobuf);
    TDataType egudin = new TDataType();
    TLink tl = new TLink("/HERA/HEVACN/PR-WEG:PR367","PRESSURE.EGU",egudout,egudin,TAccess.CA_READ);
    cc = tl.execute(1000);
    tl.cancel();
    System.out.println("completion " + cc + " egu " + egu[0].str + " max : " + egu[0].f2val + " min : " + egu[0].f1val);
*/
    String[] hproperties = TQuery.getDeviceProperties("HERA","HEHISTORY","#0");
    if (hproperties != null)
      for (int i=0; i<hproperties.length; i++) 
        System.out.println(hproperties[i]);
    else 
      System.out.println("No properties ...");


    TLink sinx = new TLink("/HERA/HEEIDC/GEAR0/Strom");
    cc = sinx.execute(1000);
    int dsize = sinx.dOutput.getLastDataSize();
    int dfmt = sinx.dOutput.getFormat();
    System.out.println("completion " + cc + " size " + dsize + " format " + dfmt);
    sinx.setArrayDelimiter(", ");
    System.out.println(sinx.dOutput.toString());    


    float[] tstval = new float[2];
    dout = new TDataType(tstval);
    TLink tst = new TLink("/TEST/SINE/SINEDEV_2","TEST",dout,din,TAccess.CA_READ);
    tst.attach((short)(TMode.CM_POLL | TMode.CM_MCAST),(TCallback)null,500);    
    
    try
    {
      for (int i=0; i<10; i++)
      {
        Thread.currentThread().sleep(1000);
        System.out.println("test : " + tstval[1]);
      }
    }
    catch (InterruptedException e) {};
    
    System.exit(0);
/*
//	  double[] sinbuf = new double[1000];
	  int[] sinbuf = new int[100];
	  int[] echobuf = new int[100];
	  for (int i=0; i<100; i++) echobuf[i] = (100+i)*2;
	  TDataType dout = new TDataType(sinbuf);
//	  TDataType din = new TDataType(echobuf);
	  TDataType din = new TDataType();
    String[] devices = TQuery.getDeviceNames("HERA","HERAKRYONORTH");
    if (devices != null)
      for (int i=0; i<devices.length; i++) 
        System.out.println(devices[i]);
    System.exit(0);

    String[] contexts = TQuery.getContexts();
    if (contexts != null)
      for (int i=0; i<contexts.length; i++) 
        System.out.println(contexts[i]);
    String[] dservers = TQuery.getDeviceServers(contexts[0]);
    if (dservers != null)
      for (int i=0; i<dservers.length; i++) 
        System.out.println(dservers[i]);
    devices = TQuery.getDeviceNames("TTF","BPM.INJ");
    if (devices != null)
      for (int i=0; i<devices.length; i++) 
        System.out.println(devices[i]);
    else 
      System.out.println("No devices ...");
    String[] properties = TQuery.getDeviceProperties("TTF","BPM.INJ","");
    if (properties != null)
      for (int i=0; i<properties.length; i++) 
        System.out.println(properties[i]);
    else 
      System.out.println("No properties ...");
    if (devices != null)
    {
      PropertyQueryEx[] pqx = TQuery.getDevicePropertyInformation("HERA","HEPBPM",devices[0],"ORBIT.X");
      if (pqx != null)
        for (int i=0; i<pqx.length; i++) 
        {
          System.out.println("Name:" + pqx[i].prpName);
          System.out.println("Description:" + pqx[i].prpDescription);
          System.out.println("Calling info:" + pqx[i].prpSize + " " + TFormat.toString(pqx[i].prpFormat));
          System.out.println("history depth: " + pqx[i].prpHistoryDepth);
          System.out.println("Range:" + pqx[i].prpMinValue + " to " + pqx[i].prpMaxValue + " " + pqx[i].prpUnits);
        }
    }
    System.exit(0);

    
    
// done for execute	  // make an engine which loops on recv with 1000 msec timeout (separate thread)
// done	  // make engine handle long packets : blkid, numblk, mtu, etc.
	  // event notification ?
	  // get username from environment ...
	  // get address from file ...
	  // system formats ...
	  TLink sin = new TLink("/TEST/SINE/SINEDEV_0","TEST",dout,din,TAccess.CA_READ);
	  System.out.println("First call");
	  if ((cc=sin.execute()) == 0)  // synchronous access ...
	  {
      String sinout = "sin: " + sinbuf[0] + ", " + sinbuf[1] + ", " + sinbuf[2];
      System.out.println(sinout);
      sinout = "sin: " + sinbuf[97] + ", " + sinbuf[98] + ", " + sinbuf[99];
      System.out.println(sinout);
	  }
	  else 
	  {
		  String msg = "err: " + cc;
		  System.out.println(msg);
	  }
	  try 
	  {
	    Thread.currentThread().sleep(1000);
	  }
	  catch (InterruptedException e) {};
	  System.out.println("Second call");
	  if ((cc=sin.execute()) == 0)  // synchronous access ...
	  {
      String sinout = "sin: " + sinbuf[0] + ", " + sinbuf[1] + ", " + sinbuf[2];
      System.out.println(sinout);
      sinout = "sin: " + sinbuf[97] + ", " + sinbuf[98] + ", " + sinbuf[99];
      System.out.println(sinout);
	  }
	  else 
	  {
		  String msg = "err: " + cc;
		  System.out.println(msg);
	  }

    try 
    {
      Thread.currentThread().sleep(1000);
    }
    catch (InterruptedException e) {};

    TLink sin1 = new TLink("/TEST/SINE/SINEDEV_1","TEST",dout,din,TAccess.CA_READ);
    System.out.println("TCP call");
    if ((cc=sin1.execute(500,TMode.CM_CONNECT)) == 0)  // synchronous access ...
    {
      String sinout = "sin: " + sinbuf[0] + ", " + sinbuf[1] + ", " + sinbuf[2];
      System.out.println(sinout);
      sinout = "sin: " + sinbuf[97] + ", " + sinbuf[98] + ", " + sinbuf[99];
      System.out.println(sinout);
    }
    else 
    {
      String msg = "err: " + cc;
      System.out.println(msg);
    }

    int[] tstbuf = new int[10];
    dout = new TDataType(tstbuf);
    TLink tst1 = new TLink("/TEST/MSTLXGLB/#3","TEST",dout,din,TAccess.CA_READ);
    System.out.println("ERR call");
    if ((cc=tst1.execute(500)) == 0 || 
        (cc & TErrorList.CE_SENDDATA) == TErrorList.CE_SENDDATA)
    {
      String sinout = "tst: " + tstbuf[0] + ", " + tstbuf[1] + ", " + tstbuf[2];
      System.out.println(sinout);
    }
    if (cc != 0)
    {
      String msg = "err: " + cc + " -> " + tst1.getLastError();
      System.out.println(msg);
    }
    
    dout = new TDataType(sinbuf);
    TLink sinr = new TLink("/TEST/MSTLXGLB/SINEDEV_1","SINE",dout,din,TAccess.CA_READ);
    System.out.println("redir call");
    if ((cc=sinr.execute()) == 0)  // synchronous access ...
    {
      String sinout = "sin: " + sinbuf[0] + ", " + sinbuf[1] + ", " + sinbuf[2];
      System.out.println(sinout);
      sinout = "sin: " + sinbuf[97] + ", " + sinbuf[98] + ", " + sinbuf[99];
      System.out.println(sinout);
    }
    else 
    {
      String msg = "err: " + cc;
      System.out.println(msg);
    }
    
    
    System.out.println("globals call");
    float[] glbcurrent = new float[1];
    dout = new TDataType(glbcurrent);
//    TLink glbcur = new TLink("GLOBAL","HECur",dout,din,TAccess.CA_READ);
    TLink glbcur = new TLink("HECur",dout);
    glbcur.receive((TCallback)null);
    
    float[] tstval = new float[2];
    dout = new TDataType(tstval);
    TLink tst = new TLink("/TEST/SINE/SINEDEV_2","TEST",dout,din,TAccess.CA_READ);
    tst.attach((short)(TMode.CM_POLL | TMode.CM_MCAST),(TCallback)null,1000);    
    
    try
    {
      for (int i=0; i<10; i++)
      {
        Thread.currentThread().sleep(1000);
        System.out.println("current : " + glbcurrent[0]);
        System.out.println("test : " + tstval[1]);
      }
    }
    catch (InterruptedException e) {};
	  System.exit(0);
  */
	}
  public int linkId;  
  int linkTimeouts;  
  boolean pending = false;  
  //static TErrorList tel;  
  TFactory tf; 
  TFactory.TBucket tb; 
  TLink() {}   
  protected TLink(TFactory tfactory, int tid, String devname, String devproperty, TDataType dout, TDataType din, short devaccess)
  {
    String fullname = new String(devname);
    active = true;
    if (dout != null) dOutput = dout; else dOutput = new TDataType();
    dInput = din;
    int indx = fullname.indexOf('/');
    tf = TFactory.getInstance();
    linkId = tid=tf.registerLink(this);
    
    cntName = "DEFAULT";
    if (indx == 0)  // context given
    {
      fullname = fullname.substring(1);
      indx = fullname.indexOf('/');
      if (indx != -1) 
      {
        cntName = fullname.substring(0,indx);
        fullname = fullname.substring(indx+1);
        indx = fullname.indexOf('/');
      }
    }
    if (indx == -1)
    {
      expName = fullname;
      devName = "";
    }
    else
    {
      expName = fullname.substring(0,indx);
      devName = fullname.substring(indx+1);
    }
    fullname = indx == -1 ? "" : fullname.substring(indx+1);
//	devName = new String(devname);
    devProperty = new String(devproperty);
    devAccess = devaccess;
    devTimeout = (short)2000;
//	tfactory.tfThrd.interrupt();
    if (expName.compareTo("ENS") == 0)
    {
      modAddress = new TModAddress("",expName,cntName);
      if (modAddress.srvAddress == null) System.out.println("Cannot resolve Equipment Name Server Address");
    }
    else 
    {
      modAddress = new TModAddress(tfactory,expName,cntName);
      if (modAddress.srvAddress == null && tf.debugLevel > 0) 
        System.out.println("Cannot resolve Server Address for /" + cntName + "/" + expName);
    }
    con = new TContract(this);
    sub = new TSubscription(con,this); // going out ...
//	phs = new TPublishedData(); // coming in ...
    hdr = new Theader(System.getProperty("user.name"));
    tf.linkReady(this);
  }

/** 
 * @brief Attaches the data sets to the created link at the given polling rate
 * 
 * Asynchronous. This method attaches the data sets to the created link using the calling
 * parameters provided at instantiation.  The arriving data will automatically
 * update the output data set and then fire the callback method provided.
 * If the callback is not needed, a null can be passed for this parameter.
 *  
 * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
 * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
 * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
 * @param f is the 'callback function' to be called whenever either new data 
 * arrive or a change in link status occurs.
 * The callback function should instantiate a class you create which 
 * implements the TCallback() class interface.
 * @param pollrate is the desired polling rate (at the server) in milliseconds.
 *
 * @return A positive link handle if successful, otherwise -1.  On 
 * failure the linkStatus property of TLink can be examined.
 * 
 * \b Example:
 * 
 * \include eg_attach2.java
 *
 * @see TINE.TLink#TLink
 * @see TINE.TLink#TLinkCallback
 * @see TINE.TLink#execute
 */
  public synchronized int attach(short mode,TCallback f,int pollrate)
  {
	  return attach(mode,f,pollrate,-1);
  }

/** 
 * @brief Attaches the data sets to the created link at the given polling rate
 * and signals the callback routine using the give notification parameter. 
 * 
 * Asynchronous. This method attaches the data sets to the created link using the calling
 * parameters provided at instantiation.  The arriving data will automatically
 * update the output data set and then fire the callback method provided.
 * If the callback is not needed, a null can be passed for this parameter.
 *  
 * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
 * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
 * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
 * @param f is the 'callback function' to be called whenever either new data 
 * arrive or a change in link status occurs.
 * The callback function should instantiate a class you create which 
 * implements the TCallback() class interface.
 * @param pollrate is the desired polling rate (at the server) in milliseconds.
 * @param notificationId is the desired identifier to be passed to the
 * supplied callback routine when it is fired.
 *
 * @return A positive link handle if successful, otherwise -1.  On 
 * failure the linkStatus property of TLink can be examined.
 * 
 * \b Example:
 * 
 * \include eg_attach3.java
 *
 * @see TINE.TLink#TLink
 * @see TINE.TLink#TCallback
 * @see TINE.TLink#execute
 */
  public synchronized int attach(short mode,TCallback f,int pollrate,int notificationId)
  {
    if (this.linkId < 0)
    {
      this.linkStatus = TErrorList.out_of_client_memory;
      return -1;
    }
    this.sub.mode = mode;
    tcb = f;
    if (pollrate < 100) pollrate = 100;
    devTimeout = (short)pollrate;
    this.sub.pollingRate = devTimeout;
    if (tf.establishLink(this) != 0) return -1;

    long ltime = System.currentTimeMillis();
    dOutput.timestamp = (int)(ltime/1000);
    this.sub.linkStartTime = ltime;
    this.sub.linkLastTime = ltime;
    if (notificationId >= 0)
      callbackId = notificationId;
    else
      callbackId = sub.id;
 
    if ((mode & TMode.CM_USE_ON_ERROR) == TMode.CM_USE_ON_ERROR) 
    {
      dError = (TDataType)dOutput.clone();
      if (dError != null) useErrObject = true;
    } 
    if ((mode & TMode.CM_WAIT) == TMode.CM_WAIT) waitForLinkCompetion();
 
    return sub.id;
  }

/** 
 * @brief Attaches the data sets to the created link at the given polling rate
 * and signals the callback routine using a reference to the TLink instance. 
 * 
 * Asynchronous. This method attaches the data sets to the created link using the calling
 * parameters provided at instantiation.  The arriving data will automatically
 * update the output data set and then fire the callback method provided.
 * If the callback is not needed, a null can be passed for this parameter.
 * This is the preferred AttachLink method to use.  The Callback instantiates
 * a TLinkCallback() interface which provides a TLink object as parameter.
 *  
 * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
 * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
 * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
 * @param f is the 'callback function' to be called whenever either new data 
 * arrive or a change in link status occurs.
 * The callback function should instantiate a class you create which 
 * implements the TLinkCallback() class interface.
 * @param pollrate is the desired polling rate (at the server) in milliseconds.
 *
 * @return A positive link handle if successful, otherwise -1.  On 
 * failure the linkStatus property of TLink can be examined.
 * 
 * \b Example:
 * 
 * \include eg_attach4.java
 *
 * @see TINE.TLink#TLink
 * @see TINE.TLink#TLinkCallback
 * @see TINE.TLink#execute
 */
  public synchronized int attach(short mode,TLinkCallback f,int pollrate)
  {
    if (this.linkId < 0)
    {
      this.linkStatus = TErrorList.out_of_client_memory;
      return -1;
    }
    this.sub.mode = mode;
    tlcb = f;
    if (pollrate < 100) pollrate = 100;
    devTimeout = (short)pollrate;
    if (tf.establishLink(this) != 0) return -1;

    long ltime = System.currentTimeMillis();
    dOutput.timestamp = (int)(ltime/1000);
    this.sub.linkStartTime = ltime;
    this.sub.linkLastTime = ltime;
 
    if ((mode & TMode.CM_USE_ON_ERROR) == TMode.CM_USE_ON_ERROR) 
    {
      dError = (TDataType)dOutput.clone();
      if (dError != null) useErrObject = true;
    } 
    if ((mode & TMode.CM_WAIT) == TMode.CM_WAIT) waitForLinkCompetion();
 
    return sub.id;
  }
  /** 
   * @brief Attaches the data sets to the global link
   * and signals the callback routine using the assigned link id. 
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   * Using the receive() method assumes that the data requested are global available.
   *  
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TCallback() class interface.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_receive1.java
   *
   * @see TINE.TLink#TLink
   * @see TINE.TLink#TLinkCallback
   * @see TINE.TLink#receive(TCallback f,int notificationId)
   * @see TINE.TLink#receive(TLinkCallback f)
   */
  public synchronized int receive(TCallback f)
  {
    return attach(TMode.CM_GLOBAL,f,1000);
  }
  /** 
   * @brief Attaches the data sets to the global link
   * and signals the callback routine using a reference to the
   * given notifcation ID in the callback
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   * Using the receive() method assumes that the data requested are global available.
   *  
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TCallback() class interface.
   * @param notificationId is the desired identifier to be passed to the
   * supplied callback routine when it is fired.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_receive2.java
   *
   * @see TINE.TLink#TLink
   * @see TINE.TLink#TLinkCallback
   * @see TINE.TLink#receive(TCallback f)
   */
  public synchronized int receive(TCallback f,int notificationId)
  {
    return attach(TMode.CM_GLOBAL,f,1000,notificationId);
  }
  /** 
   * @brief Attaches the data sets to the global link
   * and signals the callback routine using a reference to the
   * referenced TLink object.
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   * Using the receive() method assumes that the data requested are global available.
   * This is the preferred receive method to use.  The Callback instantiates
   * a TLinkCallback() interface which provides a TLink object as parameter.
   *  
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TLinkCallback() class interface.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_receive3.java
   *
   * @see TINE.TLink#TLink
   * @see TINE.TLink#TLinkCallback
   * @see TINE.TLink#receive(TCallback f)
   */
  public synchronized int receive(TLinkCallback f)
  {
    return attach(TMode.CM_GLOBAL,f,1000);
  }
  /**
   * @brief Stops thread execution until this link has completed.
   * Useful for grouped asynchronous links.
   */
  public void waitForLinkCompetion() 
  {
    try
    {
  	 while (this.linkStatus < 0)
  	 {
  	   Thread.sleep(100);
  	 }
    }
    catch (InterruptedException e) 
    {
    	System.out.println(" waitForLinkCompletion : " + e);
    }
  }
  /**
   * @brief Gets the array delimiter for use in the toString() method 
   * of associated TDataType objects.
   * 
   * @return The delimiter in use in separating data
   * variables in a string dump (default ",")
   */
  public String getArrayDelimiter()
  {
    return arrayDelimiter;
  }
  /**
   * @brief Sets the array delimiter for use in the toString() method 
   * of associated TDataType objects.
   * 
   * @param delimiter is the string delimiter to use in separating data
   * variables in a string dump (default ",")
   */
  public void setArrayDelimiter(String delimiter)
  {
    if (arrayDelimiter != null && arrayDelimiter.compareTo(delimiter) == 0) return;
    arrayDelimiter = new String(delimiter);
    this.dOutput.setArrayDelimiter(arrayDelimiter);
  }
}