package TINE;
/**
 * @brief tine Query routines
 * 
 * This is a static class which can be used to make system queries without
 * the need for instantiating a TQuery object.
 */
public final class TQuery
{
  public static final int nicePropertyQuerySize = 10;
  private static int maxQueryBufferSize = 65000;
  private static byte[] hByteBlob = new byte[nicePropertyQuerySize * PropertyQueryEx.sizeInBytes];
  private static  byte[] hLegacyByteBlob;
  private static String lastQueriedContext = null;
  private static String lastQueriedServer = null;
  private static String lastQueriedDevice = null;
  private static StringBuffer host = new StringBuffer(32);
  private static StringBuffer query = new StringBuffer(32);
  private static short[] numout = new short[1];
  private static TDataType numoutData = new TDataType(numout);
  private static String[] getTagList(String context,String tagtype)
  {
    int cc,n,i;
    TDataType dout;
    TLink tl;
    NAME16[] taglist;
    String[] strlist;
    
    host.delete(0,31); query.delete(0,31);
    if (context != null && context.length() != 0)
    {
      host.insert(0,"ENS/" + context);
    }
    else 
    {
      host.insert(0,"ENS");
    }
    query.insert(0,"N" + tagtype);
    
    tl = new TLink(host.toString(),query.toString(),numoutData,null,TAccess.CA_READ);
    cc = tl.execute(tl.defaultTimeout,true);
    tl.cancel(); 
    if (cc != 0) return null;
//  /* Get Tags from name server (synchronous call) */
    query.delete(0,31);
    query.insert(0,tagtype);
    n = numout[0];
    taglist = new NAME16[n];
    for (i=0; i<n; i++) taglist[i] = new NAME16();
    if (taglist.length == 0) return null;
    if (numout[0] > taglist.length) numout[0] = (short)taglist.length;  
    dout = new TDataType(taglist);
    tl = new TLink(host.toString(),query.toString(),dout,null,TAccess.CA_READ);
    cc = tl.execute(tl.defaultTimeout,true);
    tl.cancel();
    if (cc != 0) return null;
    strlist = new String[n];
    for (i=0; i<n; i++) strlist[i] = taglist[i].name;
    return strlist;
  }
  private static int getNumberOf(String property,String target)
  {
    int cc;
    short[] n = new short[1];
    TDataType dout;
    TLink tl;
    String propertyq;
    
    if (property == null || target == null) return -TErrorList.argument_list_error;
   
    dout = new TDataType(n);
    propertyq = new String("N" + property); 
    tl = new TLink(target,propertyq,dout,null,TAccess.CA_READ);
    cc = tl.execute(tl.defaultTimeout,true);
    tl.cancel();
    return cc > 0 ? -cc : (int)n[0];    
  }
  /**
   * @brief Sets the maximum size for use in tine queries.
   *
   * @param size is the size in bytes of the buffer used to hold the query
   * information (65000 bytes is the default)
   *  
   * @return 0 if successful 
   */
  public static short setMaximumQueryBufferSize(int size)
  {
    if (size < 4000 || size > 256000) return TErrorList.out_of_range;
    maxQueryBufferSize = size;
    return 0;
  }
  /**
   * @brief Returns a list of Contexts (obtained from the name services)
   * 
   * @return A string array containing a list of available contexts as 
   * obtained from the Equipment Name Server (ENS)
   */
  public static String[] getContexts()
  {
    return getTagList(null,"CONTEXTS");
  }
  /**
   * @brief Returns a list of device servers associated with the specified
   * context (obtained from the name services)
   * 
   * @param context Is the context for which the device server name list is
   * desired
   * 
   * @return A string array containing a list of device servers associated
   * with the given context as obtained from the Equipment Name Server (ENS)
   */
  public static String[] getDeviceServers(String context)
  {
    return getTagList(context,"TAGS");
  }
  /**
   * @brief Returns a list of front end computers associated with the specified
   * context (obtained from the name services)
   * 
   * @param context Is the context for which the device server name list is
   * desired
   * 
   * @return A string array containing a list of front end computers associated
   * with the given context as obtained from the Equipment Name Server (ENS)
   */
  public static String[] getServers(String context)
  {
    return getTagList(context,"FECS");
  }
  /**
   * @brief Returns a list of properties associated with the specified
   * context and device server and device (obtained from the device server)
   * 
   * @param context Is the context of the device server for which the property list is
   * desired.
   * @param server Is the device server for which the property list is
   * desired.
   * @param device Is the device name (module name) for which the property list
   * is desired. 
   * 
   * @return A string array containing a list of properties
   * as obtained from the given device server
   */
  public static String[] getDeviceProperties(String context,String server,String device)
  {
    return getDeviceProperties(context,server,device,null);
  }
  /**
   * @brief Returns a list of properties associated with the specified
   * context and device server, device, and property (sub)string 
   * (obtained from the device server)
   * 
   * @param context Is the context of the device server for which the property list is
   * desired.
   * @param server Is the device server for which the property list is
   * desired.
   * @param device Is the device name (module name) for which the property list
   * is desired.
   * @param property Can be a queriable string with wildcard 
   * parameter '*'.  e.g. property = 'P*' will return all properties whose
   * first letter begins with 'P'.
   * 
   * @return A string array containing a list of properties
   * as obtained from the given device server
   */
  public static String[] getDeviceProperties(String context,String server,String device,String property)
  {
    TLink tl;
    TDataType dout;
    TDataType din;
    String tgt;
    int cc = 0,n,i,j;
    NAME32[] properties;
    String[] strprops;

    tgt = new String("/" + context + "/" + server + "/" + device);
    n = getNumberOf("PROPS",tgt);
    if (n <= 0) return null;
    if (n*32 > maxQueryBufferSize) n = maxQueryBufferSize/32;
    properties = new NAME32[n];
    for (i=0; i<n; i++) properties[i] = new NAME32();
    if (properties.length <= 0) return null;
    dout = new TDataType(properties);
    if (property != null) din = new TDataType(property); else din = null;
    tl = new TLink(tgt,"PROPS",dout,din,TAccess.CA_READ);
    cc = tl.execute(tl.defaultTimeout,true);
    tl.cancel();
    if (cc == 0)
    {
      for (i=0,n=0; i<properties.length; i++) if (properties[i].name.length() > 0) n++;
      strprops = new String[n];
      for (i=0,j=0; j<n && i<properties.length; i++) 
      {
        if (properties[i].name.length() == 0) continue;
        strprops[j++] = properties[i].name;
      }
      return strprops;
    }
    else if (cc != TErrorList.link_not_open)
    {
      if (lastQueriedContext == context && lastQueriedServer == server &&
          lastQueriedDevice == device && lastQueriedServer != null)
      {
        cc = 0;
      }
      else
      {
        hLegacyByteBlob = new byte[n * PropertyQuery.sizeInBytes];
        dout = new TDataType(hLegacyByteBlob,"");
        tl = new TLink(tgt,"PROPS",dout,din,TAccess.CA_READ);
        cc = tl.execute(tl.defaultTimeout,false);
        tl.cancel();
      }
      if (cc == 0)
      {
        lastQueriedContext = context;
        lastQueriedServer = server;
        lastQueriedDevice = device;
        strprops = new String[n];
        PropertyQuery[] pq = new PropertyQuery[n];
        byte b[];
        hLegacyByteBlob = new byte[n * PropertyQuery.sizeInBytes];
        for (i=0; i<n; i++)
        {
          pq[i] = new PropertyQuery();
          b = pq[i].toByteArray();
          System.arraycopy(hLegacyByteBlob,i*PropertyQuery.sizeInBytes,b,0,PropertyQuery.sizeInBytes);
          pq[i].toStruct();
          strprops[i] = pq[i].name;
        }        
        return strprops;
      }
    }
    return null;
  }
  /**
   * @brief Returns a list of device names associated with the specified
   * context and device server 
   * (obtained from the device server)
   *
   * As the call is directed to the device server, it assumes the return 
   * list applies to all properties
   *  
   * @param context Is the context of the device server for which the property list is
   * desired.
   * @param server Is the device server for which the property list is
   * desired.
   * 
   * @return A string array containing a list of devices
   * as obtained from the given device server
   */
  public static String[] getDeviceNames(String context,String server)
  {
    return getDeviceNames(context,server,null);
  }
  /**
   * @brief Returns a list of device names associated with the specified
   * context, device server and property 
   * (obtained from the device server)
   *
   * @param context Is the context of the device server for which the device list is
   * desired.
   * @param server Is the device server for which the device list is
   * desired.
   * @param property Is the property for which the device list is
   * desired.
   * 
   * @return A string array containing a list of devices
   * as obtained from the given device server
   */
  public static String[] getDeviceNames(String context,String server,String property)
  {
    TLink tl;
    TDataType dout;
    int cc = 0,n,i,j;
    String prp, tgt, tmp;
    NAME16[] devices;
    String[] strdevs;
    PropertyQueryEx[] pqx;

    tgt = new String("/" + context + "/" + server + "/#0");
    if (property != null && property.length() > 0)
    {
      pqx = getDevicePropertyInformation(context,server,"#0",property);
      n = pqx[0].prpSize;    
      prp = property + ".NAM";
    }
    else
    {
      n = getNumberOf("DEVICES",tgt);
      if (n < 0) return null;
      n++;
      prp = "DEVICES"; 
    }

    if (n*16 > maxQueryBufferSize) n = maxQueryBufferSize/16;
    devices = new NAME16[n];
    for (i=0; i<n; i++) devices[i] = new NAME16();
    if (devices.length == 0) return null;
    if (n == 1)
    { 
      devices[0].name = "#0"; // server claimed not to have device names
    }
    else
    {
      dout = new TDataType(devices);
      tl = new TLink(tgt,prp,dout,null,TAccess.CA_READ);
      cc = tl.execute(tl.defaultTimeout,true);
      tl.cancel();
      if (cc != 0) return null;
    }
    // check for long names (the Channel Access kluge)
    for (i=0,n=0; i<devices.length; i++) 
      if (devices[i].name.length() > 0 && !devices[i].name.endsWith("&")) n++;
    strdevs = new String[n];
    for (i=0,j=0; i<devices.length && j<n; i++) 
    {
      tmp = devices[i].name;
      if (tmp.length() == 0) continue;
      if (tmp.endsWith("&"))
      {
        i++;
        tmp = tmp.substring(0, tmp.length() - 1) + devices[i].name;
      }
      strdevs[j++] = tmp;
    }
    return strdevs;
  }
  /**
   * @brief Returns a list of extended property query information objects associated
   * with the specified context, device server, device. and target property 
   * (obtained from the device server).
   *
   * Usually a call to getDeviceProperties() returns a property list
   * A secondary call to getDevicePropertyInformation() returns a list of 
   * all property information pertaining to the specified property including
   * all property overloads (maximum 10).  If 'property' is null then the
   * call returns information for the first 10 items.
   *  
   * @param context Is the context of the device server for which the property
   * information is desired.
   * @param server Is the device server for which the property information is
   * desired.
   * @param device Is the device name (module name) for which the property 
   * information is desired. 
   * @param property Is the property for which the property information is
   * desired.
   * 
   * @return An array of PropertyQueryEx (extended Property Query) objects,
   * as obtained from the given device server
   */
  public static synchronized PropertyQueryEx[] getDevicePropertyInformation(String context,String server,String device,String property)
  {
    TLink tl;
    TDataType dout, din;
    int cc = 0,i,n;
    String tgt;
    PropertyQueryEx[] pqx = null;
    PropertyQueryEx pqx1 = new PropertyQueryEx();

    if (context == null || context.length() == 0)
      tgt = new String("/DEFAULT/" + server + "/" + device);
    else
      tgt = new String("/" + context + "/" + server + "/" + device);
    dout = new TDataType(hByteBlob,"");
    din = new TDataType(property);
    tl = new TLink(tgt,"PROPERTIES",dout,din,TAccess.CA_READ);
    cc = tl.execute(tl.defaultTimeout,true);
    tl.cancel();
    if (cc == 0) // it's at lease a 3.20 server
    {
      byte b[] = pqx1.toByteArray();
      System.arraycopy(hByteBlob,0,b,0,PropertyQueryEx.sizeInBytes);
      pqx1.toStruct();
      if (pqx1.prpNumOverloads < 1) pqx1.prpNumOverloads = 1; // can't be less than 1 !
      int npqx = pqx1.prpNumOverloads;
      if (npqx > nicePropertyQuerySize) npqx = nicePropertyQuerySize;
      pqx = new PropertyQueryEx[npqx];
      pqx[0] = pqx1;
      for (i=1; i<npqx; i++)
      {
        pqx[i] = new PropertyQueryEx();
        b = pqx[i].toByteArray();
        System.arraycopy(hByteBlob,i*PropertyQueryEx.sizeInBytes,b,0,PropertyQueryEx.sizeInBytes);
        pqx[i].toStruct();
      }
    }
    else if (cc != TErrorList.link_not_open) // legacy server
    {
      String[] prps = getDeviceProperties(context,server,device);
      if (prps == null) return null;
      PropertyQuery[] pq = new PropertyQuery[prps.length];
      pqx = new PropertyQueryEx[1];
      byte b[];
      hLegacyByteBlob = new byte[prps.length * PropertyQuery.sizeInBytes];
      for (i=0; i<prps.length; i++)
      {
        pq[i] = new PropertyQuery();
        b = pq[i].toByteArray();
        System.arraycopy(hLegacyByteBlob,i*PropertyQuery.sizeInBytes,b,0,PropertyQuery.sizeInBytes);
        pq[i].toStruct();
        if (pq[i].name.compareTo(property) != 0) continue;
        pqx[0] = new PropertyQueryEx();
        pqx[0].prpFormat = pq[i].prpFormat;
        pqx[0].prpSize = pq[i].prpSize;
        pqx[0].prpAccess = pq[i].prpAccess;
        pqx[0].prpDescription = pq[i].prpDesc;
        pqx[0].prpName = pq[i].name;
        pqx[0].prpNumOverloads = 1;
        if ((pq[0].prpAccess & TAccess.CA_WRITE) == TAccess.CA_WRITE)
        {
           pqx[0].prpFormatIn = pq[i].prpFormat;
           pqx[0].prpSizeIn = pq[i].prpSize;
        }
        else     
        {
           pqx[0].prpFormatIn = (byte)TFormat.CF_NULL;
           pqx[0].prpSizeIn = 0;
        }
        USTRING[] egu = new USTRING[1];
        egu[0] = new USTRING();
        dout = new TDataType(egu);
        din = new TDataType();
        String prpegu = new String(property + ".EGU");
        tl = new TLink(tgt,prpegu,dout,din,TAccess.CA_READ);
        cc = tl.execute(tl.defaultTimeout,false);
        tl.cancel();
        if (cc == 0)
        {
          pqx[0].prpMinValue = egu[0].f1val;
          pqx[0].prpMaxValue = egu[0].f2val;
          pqx[0].prpUnits = egu[0].str;
          pqx[0].prpGraphType = (byte)egu[0].ival;  
        }
        break;
      }        
    }
    return pqx;
  }
  
}
