/**
 * sedaclib.c
 * 
 * #define SEDAC_READIO to use memory mapped polling of sedac card (otherwise sedac is in
 * driver mode).
 * #define MULTITHREADED if this is a multithreaded application.
 * #define N_SEDAC_LINES <nr> to be the number of sedac lines used by the application
 *
 */
#include "prolog.h"
#include <stdio.h>
#include <errno.h>
#include <time.h>

/***   For mmap function   ***/
#include <unistd.h>
#include <sys/mman.h> 

/***   For open function   ***/
#include <sys/types.h>
#include <sys/stat.h>  
#include <fcntl.h> 

/***   For system function errors "errno", for example ***/
#include <stdlib.h>

#ifdef SEDAC_READIO
# include <asm/io.h>
# define PAGE_SIZE       0x1000
#else
# include "sedacDrv.h"
#endif

/* simple for(;;) loop without hardware read (2000 countes): 600us and 100 us	*/ 
#define N_POLLS	500		/* 640us for 130Mhz pentium, 950us for 33Mhz 486	*/

#include "rpcdefs.h"
#include "sedac.h"
#include "errors.h"
#include "unixcmd.h"

int nSedacLines = N_SEDAC_LINES;;
int SedacSimulate = 0;
u16 *ATC40Address = NULL;

u16 sedacSimulationBuffer[N_SEDAC_LINES];
u16 sedacSimulationCount[N_SEDAC_LINES] = {0,0,0,0};
#ifdef SEDAC_READIO
SEDAC_LINE volatile *line[N_SEDAC_LINES];
#else
PSEDAC_DATA da[N_SEDAC_LINES] = {0,0,0,0};		// dynamic allocation
int dataSize[N_SEDAC_LINES] = {0,0,0,0};
int sedac_fd[N_SEDAC_LINES] = {-1,-1,-1,-1};
#endif

#ifdef MULTITHREADED
#include "pthread.h"
pthread_mutex_t sedacMutex;
pthread_mutexattr_t sedacAttr;
#endif
/***************************************/
void initSedac(void)
{ 
  int i, j;
# ifdef SEDAC_READIO
  int mem_fd;
  caddr_t atc40_mem;				/* Pointer to "reflected" ATC40 memory */
  long rdSedacDrvMem[3];   // deviceType, io and memSize
# define BASE_ATC40_ADDR rdSedacDrvMem[1]
# define SIZE_ATC40_MEM rdSedacDrvMem[2]
# else
  char fname[80];
# endif

  if (SedacSimulate) return;
# ifdef SEDAC_READIO
  feclog("ioctl: using SEDAC in polling mode");
# else
  feclog("ioctl: using SEDAC in driver mode");
# endif
# ifdef MULTITHREADED
  pthread_mutexattr_settype(&sedacAttr,PTHREAD_MUTEX_RECURSIVE_NP);
  pthread_mutex_init(&sedacMutex,&sedacAttr);
# endif

  for (i=0; i<nSedacLines; i++)		// look for up to # SEDAC lines
  {
#   ifdef SEDAC_READIO
    if( ATC40Address == NULL ) 
    {    /*** Open /dev/mem   ***/   
      if(( mem_fd = open("/dev/sedac0", O_RDWR) ) < 0 ) 
      {
        feclog("ioctl error: ATC40 memory-window can't open /dev/mem \n");
        exit(-1);
      }

      for ( j=0; j<2; j++ )   // -10: io; -11: devicetype
      {
        if ( ioctl(mem_fd,j-11,&rdSedacDrvMem[j]) < 0 )
        {
          feclog("ioctl error: %d; %s\n",errno, strerror(errno));
          exit(1);
        }
      }

      if ( rdSedacDrvMem[0] < 1 || rdSedacDrvMem[0] > 2 )
      {
        feclog("ioctl error: cannot map sedac Memory, Wrong sedac device type!");
        exit(1);
      }
      rdSedacDrvMem[2] = (rdSedacDrvMem[0] == 1 ? 0x300 : 0x4000);	// atc40- or pci-ipcard
      feclog("ioctl: sedac type %d  mem addr %x  mem size %x",
        rdSedacDrvMem[0],rdSedacDrvMem[1],rdSedacDrvMem[2]);

      /***   mmap ATC40 memory   ***/
      if ((atc40_mem = malloc( rdSedacDrvMem[2] + 2*(PAGE_SIZE-1) )) == NULL) 
      {
        feclog("ioctl error: ATC40 memory allocation error");  
        exit(-1);
      }

      if(( unsigned long)atc40_mem % PAGE_SIZE )
        atc40_mem += PAGE_SIZE - ((unsigned long)atc40_mem % PAGE_SIZE );

      atc40_mem = (caddr_t)mmap(   /* void *mmap */
        (void *)atc40_mem,         /* void *start */
        (size_t)rdSedacDrvMem[2],
        PROT_READ|PROT_WRITE,      /* int prot */
        MAP_SHARED|MAP_FIXED,      /* int flags */
        (int)mem_fd,               /* int fd */
        (off_t)rdSedacDrvMem[1]);  /* size_t */

      if( (long)atc40_mem < 0 ) 
      { 
        feclog("ioctl error: ATC40 memory map error: atc40_mem = %x", atc40_mem);  
        feclog("ioctl error: mmap failed, errno: %d: %s", errno, strerror(errno));
        exit(-1);
      }
      ATC40Address = (u16 *)atc40_mem;
    }
    if ( rdSedacDrvMem[0] == 2 )     // pci ip card
      line[i] = (SEDAC_LINE *) (ATC40Address+0x800+(i/2)*0x1000 + (i%2)*0x04);
    else
      line[i] = (SEDAC_LINE *) (ATC40Address + (i/2)*0x100 + (i%2)*0x04);

    feclog("Initializing SEDAC line %d at base address %x, %x, %x",i+1, &line[i]->csr, &line[i]->igr, &line[i]->dat);
    
    line[i]->csr = 0;		            // init sedac ip module
    line[i]->igr = 0;
#   else
    sprintf(fname,"/dev/sedac%d",i);
    if((sedac_fd[i]=open(fname, O_RDWR)) < 0 )
      fprintf(stderr, "%s: %s; error Number: %d\n",fname,strerror(errno),errno);
#   endif
  }
 
  RegisterUserCommand("nlines",NULL,&nSedacLines,NULL,CA_READ);
  RegisterUserCommand("sedac",cmdReadSedac,NULL,NULL,CA_READ);
  RegisterUserCommand("sedac",cmdWriteSedac,NULL,NULL,CA_WRITE);
  RegisterUserCommand("crates",cmdCrates,NULL,NULL,CA_READ);
  RegisterUserCommand("addr",cmdAddr,NULL,NULL,CA_READ);

  return;
}
int exitSedac(void)
{ 
  int ret, i;
  u16 IOaddress;

  if (SedacSimulate) return 0;

  for (i=0; i<nSedacLines; i++)		// look for up to # SEDAC lines
  {
#   ifndef SEDAC_READIO
    if( sedac_fd[i] < 0 ) return sedac_fd[i];
    close(sedac_fd[i]);
    sedac_fd[i] = -1;
    if (da[i])
    {
      free(da[i]);
      da[i] = 0;
    }
#   endif
  }
  return 0;
}

#ifdef SEDAC_READIO
// ============  poll sedac active bit ===================================
int sedacReady(u16 ln, u16 *data)
{ 
  int i,j;
  u16 ret;

  for (i=0; i<N_POLLS ; i++ ) 
  {
    ret = line[ln]->csr;
    if ( (ret & _RBE) == 0 ) break;
  }
  if (data && i < N_POLLS)
  {
    *data = line[ln]->dat;
  }
  return (i < N_POLLS ? (ret & ERRALL) : -2) ;
}
// ================  read Async routine  ===================================
int rsedacAsync(u16 ln, u16 crate, u16 subadd)
{
  if (ln<1 || ln>nSedacLines) return -1;
  ln--;
  if ( SedacSimulate) return 0;

# ifdef MULTITHREADED
  pthread_mutex_lock(&sedacMutex);
# endif
  line[ln]->cmd = 0x8000 | (crate << 8) | subadd;
  return 0;
}
// ================  read Async Poll routine  ===================================
int rsedacPoll(u16 ln, u16 *data)
{ int i, ret;

  if (ln<1 || ln>nSedacLines) return -1;
  ln--;

  if (SedacSimulate) 
  {
    *data = sedacSimulationBuffer[ln] + sedacSimulationCount[ln]++;
    return 0;
  }
  ret = sedacReady(ln, data);
# ifdef MULTITHREADED
  pthread_mutex_unlock(&sedacMutex);
# endif
  return ret;
}
// ============  Block routines ==========================================
void sedac4(SedacInfo *sedacList, int RWF, int num, int (*fcn)(int))
{ int i, l;

# ifdef MULTITHREADED
  pthread_mutex_lock(&sedacMutex);
# endif
  switch( RWF )
  {
    case READ:
      for (i=0; i<num; i++) 
      {
        for (l=0; !SedacSimulate && l<nSedacLines; l++) line[l]->cmd = 0x8000 | sedacList[l].csa[i];
        if (fcn) fcn(i);				// service function
        for (l=0; l<nSedacLines; l++)   // SEDAC line in
          if (SedacSimulate) 
          {
            sedacList[l].data[i] = sedacSimulationBuffer[l] + sedacSimulationCount[l]++;
            sedacList[l].error[i] = 0;
          } 
          else 
          {
            sedacList[l].error[i] = sedacReady(l, &sedacList[l].data[i]);
          }
      }
      if (fcn) fcn(i);
      break;
    case WRITE:
      for (i=0; i<num; i++) 
      {
        for (l=0; !SedacSimulate && l<nSedacLines; l++) 
        {
          line[l]->dat = sedacList[l].data[i];
          line[l]->cmd = 0x4000 | sedacList[l].csa[i];
        }
        if (fcn) fcn(i);
        for (l=0; l<nSedacLines; l++) 
        {   // SEDAC lines in
          if (SedacSimulate) 
          {
            sedacSimulationBuffer[l] = sedacList[l].data[i];
            sedacList[l].error[i] = 0;
          } 
          else 
          {						  /* poll RBE bit */
            sedacList[l].error[i] = sedacReady(l, 0);
          }
        }
      }
      if (fcn) fcn(i);                                    // i = num
      break;
  }
# ifdef MULTITHREADED
  pthread_mutex_unlock(&sedacMutex);
# endif
  return;
}
#else // not SEDAC_READIO => Sedac Driver
// ============  getBuffer  ===================================
PSEDAC_DATA getBuffer(u16 ln, int number)
{ 
  PSEDAC_DATA buf;

  if ( da[ln] )
  {
	free(da[ln]);
	da[ln] = 0;
  }

  buf = (PSEDAC_DATA)malloc( SEDAC_DATA_SIZE * (number + 1) );
  if ( buf != NULL ) dataSize[ln] = number;

  printf("get buffer: %x; size: %d\n",buf,number);
  return buf;
}
// ================  read/write nwords telegram  ===================================
int execSedac(u16 ln, u16 *crate, u16 *subadd, int *RWF, u16 *data, int nwords, int flag)
{ int i;
  int ret;
  PSEDAC_HEADER phe;

  if (ln<1 || ln>nSedacLines) return -2;
  ln--;

  if ( sedac_fd[ln] < 0 ) return -3;

  if ( SedacSimulate == 0 )
  {
    if ( da[ln] == 0 || dataSize[ln] < nwords )
    if ( (da[ln] = getBuffer(ln, nwords)) == 0 ) return -4;
  }

  if ( SedacSimulate )
  {
    for (i=0; i<nwords; i++, data++)
    {
      ret = flag ? RWF[i] : RWF[0];
      if ( ret == READ )
      *data = sedacSimulationBuffer[ln] + sedacSimulationCount[ln]++;
      else
      sedacSimulationBuffer[ln] = *data;
    }
    return 0;
  }
  for (i=0; i<nwords; i++) 
  {
    ret = flag ? RWF[i] : RWF[0];
    if ( ret == WRITE ) da[ln][i+1].Data = flag ? data[i] : data[0];
    da[ln][i+1].ExcessMode = ret == READ ? SE_READ : SE_WRITE;
    da[ln][i+1].Line = ln;
    da[ln][i+1].SubAdr = flag ? subadd[i] : subadd[0];
    da[ln][i+1].CaAdr = flag ? crate[i] : crate[0];
  }
  ret = ioctl(sedac_fd[ln], nwords, da[ln]);
  if (ret == nwords + 1)
  {
    for (i=0; i<nwords; i++)
      if ( (flag ? RWF[i] : RWF[0]) == SE_READ ) data[i] = da[ln][i+1].Data;
    phe = (PSEDAC_HEADER)da[ln];
    return phe->NumberOfError;
  }
  if (ret < 0)
  {
    fprintf(stderr, "ioctl error (request #Telegrams= %d): %s; return code: %d \n",nwords, strerror(errno), ret);
    return -5;
  }

  phe = (PSEDAC_HEADER)da[ln];
  printf("timeout! return code: %d; #sedac error: %d; #Requested: %d; #Procecced: %d\n",ret, phe->NumberOfError, phe->numberRequested, phe->numberProcessed);
  return -1;
}
int nsedac(u16 ln, u16 *crate, u16 *subadd, int *RWF, u16 *data, int nwords)
{
  return execSedac(ln, crate, subadd, RWF, data, nwords, 1);
}
#endif // SEDAC_READIO
// ============  write routine  ===================================
int wsedac(u16 ln, u16 crate, u16 subadd, u16 data)
{ 
# ifndef SEDAC_READIO
  PSEDAC_HEADER phe;
# endif
  int i, ret;

  if (ln<1 || ln>nSedacLines) return -1;
  ln--;

  if (SedacSimulate) 
  {
    sedacSimulationBuffer[ln] = data;
    return 0;
  }

# ifdef MULTITHREADED
  pthread_mutex_lock(&sedacMutex);
# endif

# ifdef SEDAC_READIO
  line[ln]->dat = data;
  line[ln]->cmd = 0x4000 | (crate << 8) | subadd;
  ret = sedacReady(ln, 0);
# else
  if (sedac_fd[ln] < 0) return -3;
  if (da[ln] == 0) if ((da[ln]=(PSEDAC_DATA)getBuffer(ln, 1)) == 0 ) return -4;
  da[ln][1].Data = data;
  da[ln][1].ExcessMode = SE_WRITE;
  da[ln][1].Line = ln;
  da[ln][1].SubAdr = subadd;
  da[ln][1].CaAdr = crate;

  ret = ioctl(sedac_fd[ln], 1, da[ln]);

  if ( ret == 2 ) return da[ln][1].CommandReturn;

  if ( ret < 0 )
  {
    fprintf(stderr, "ioctl error (request #Telegrams= 1): %s; return code: %d \n",strerror(errno), ret);
    return -5;
  }
  phe = (PSEDAC_HEADER)da[ln];
  printf("timeout! return code: %d; #sedac error: %d; #Requested: %d; #Procecced: %d\n",ret, phe->NumberOfError, phe->numberRequested, phe->numberProcessed);
  return -1;
# endif

# ifdef MULTITHREADED
  pthread_mutex_unlock(&sedacMutex);
# endif

  return ret;
}
// ================  read routine  ===================================
int rsedac(u16 ln, u16 crate, u16 subadd, u16 *data)
{
# ifndef SEDAC_READIO
  PSEDAC_HEADER phe;
# endif
  int i, ret;

  if (ln<1 || ln>nSedacLines) return -1;
  ln--;

  if (SedacSimulate) 
  {
    *data = sedacSimulationBuffer[ln] + sedacSimulationCount[ln]++;
    return 0;
  }
# ifdef MULTITHREADED
  pthread_mutex_lock(&sedacMutex);
# endif

# ifdef SEDAC_READIO
  line[ln]->cmd = 0x8000 | (crate << 8) | subadd;
  ret = sedacReady(ln, data);
# else
  if (sedac_fd[ln] < 0) return -3;

  if (da[ln] == 0) if ((da[ln]=getBuffer(ln, 1)) == 0 ) return -4;
  
  da[ln][1].ExcessMode = SE_READ;
  da[ln][1].Line = ln;
  da[ln][1].SubAdr = subadd;
  da[ln][1].CaAdr = crate;

  ret = ioctl(sedac_fd[ln], 1, da[ln]);

  if ( ret == 2 )
  {
    *data = da[ln][1].Data;
    return da[ln][1].CommandReturn;
  }
  if ( ret < 0 )
  {
    fprintf(stderr, "ioctl error (request #Telegrams= 1): %s; return code: %d \n",strerror(errno), ret);
    return -5;
  }
  phe = (PSEDAC_HEADER)da[ln];
  printf("timeout! return code: %d; #sedac error: %d; #Requested: %d; #Procecced: %d\n",ret, phe->NumberOfError, phe->numberRequested, phe->numberProcessed);
  return -1;
# endif

# ifdef MULTITHREADED
  pthread_mutex_unlock(&sedacMutex);
# endif
  return ret;
}
// ================  read/write nwords telegram  ===================================
int sedac(int ln, u16 csa, int RWF, u16 *data, int nwords)
{ 
  int i, err = -1;
  u16 addr,crate;

# ifdef MULTITHREADED
  pthread_mutex_lock(&sedacMutex);
# endif

# ifdef SEDAC_READIO
  if (ln<1 || ln>nSedacLines) return -1;
  ln--;
  switch( RWF )
  {
    case READ:
	    addr = 0x8000 | csa;
	    for (err=0, i=0; i<nwords; i++, data++) 
      {
        if (SedacSimulate) 
        {
          *data = sedacSimulationBuffer[ln] + sedacSimulationCount[ln]++;
        } 
        else 
        {
	        line[ln]->cmd = addr;
	        if ( sedacReady(ln, data) ) err++;
        }
      }
      break;
    case WRITE:
      addr = 0x4000 | csa;
      for (err=0,i=0; i<nwords; i++, data++ ) 
      {
        if (SedacSimulate) 
        {
          sedacSimulationBuffer[ln] = *data;
        }
        else 
        {
          line[ln]->dat = *data;
          line[ln]->cmd = addr;
          if ( sedacReady(ln, 0) ) err++;
        } 
      }
      break;
    default:
      break;
  }
# else
  crate = csa >> 8; addr = csa & 0xff;
  return execSedac(ln, &crate, &addr, &RWF, data, nwords, 0);
# endif

# ifdef MULTITHREADED
  pthread_mutex_unlock(&sedacMutex);
# endif
  return err;
}
// ================  find crates  ===================================
int F_crates(int *ncrates, int rwflag, u16 ln)
{   
  u16 data;
  int i,n;
  char str[256];

  if (ln < 1 || ln > nSedacLines) return no_such_line;

  str[0] = 0;
  switch (rwflag)
  {
    case  READ:
      for (i=0,n=0; i<32; i++) if (!rsedac(ln, i, 252, &data))
      {
        sprintf(&str[strlen(str)],"  %2d",i);
        if ((++n%10) == 0)
        {
          strcat(str,"\n>"); ttyoutput(str);
          str[0] = 0;
        }
      }
      strcat(str,"\n>"); ttyoutput(str);
      *ncrates = n;
      return 0;
    default:
      return illegal_read_write;
  }
}
// ================  find sub addres (16*n) ===================================
int F_addr(int *naddr, int rwflag, u16 ln, u16 crate)
{ 
  u16 data;
  int i,n;
  char str[256];

  if (ln < 1 || ln > nSedacLines) return no_such_line;

  str[0] = 0;
  switch (rwflag)
  {
    case READ:
      for (i=0,n=0; i<252; i+=16) if (!rsedac(ln,crate,i,&data))
      {
        sprintf(&str[strlen(str)],"  %2d",i);
        if ((++n%10) == 0)
        {
          strcat(str,"\n>"); ttyoutput(str);
          str[0] = 0;
        }
      }
      strcat(str,"\n>"); ttyoutput(str);
      *naddr = n;
      return 0;
    default:
      return illegal_read_write;
  }
}
int cmdReadSedac(int ln,int cr,int sa,int val)
{
  UINT16 data;
  int cc;
  char str[128];

  if (nSedacLines)
  {
    if ((cc=rsedac(ln,cr,sa,&data)) != 0) sprintf(str,"SEDAC error");
    else sprintf(str,"lin %d cra %d bsa %d : %d (0x%x)\n>",ln,cr,sa,data,data);
    ttyoutput(str);
    return cc;
  }
  return -1;
}
int cmdWriteSedac(int ln,int cr,int sa,int val)
{
  UINT16 data = val;
  int cc=0;
  char str[128];

  if (nSedacLines)
  {
    if ((cc=wsedac(ln,cr,sa,data)) != 0) sprintf(str,"SEDAC error");
    else sprintf(str,"lin %d cra %d bsa %d wrote : %d (0x%x)\n>",ln,cr,sa,data,data);
    ttyoutput(str);
    return cc;
  }
  return -1;
}
int cmdCrates(int ln,int cr,int sa,int access)
{
  UINT16 data;
  int cc,n=0;
  char str[128];

  if (nSedacLines)
  {
    if ((cc=F_crates(&n,1,ln)) != 0) sprintf(str,"line %d : %s",ln,erlst[cc]);
    else sprintf(str,"line %d : %d crates\n>",ln,n);
    ttyoutput(str);
    return n;
  }
}
int cmdAddr(int ln,int cr,int sa,int access)
{
  UINT16 data;
  int cc,n=0;
  char str[128];

  if (nSedacLines)
  {
    if ((cc=F_addr(&n,1,ln,cr)) != 0) sprintf(str,"lin %d cra %d : %s",ln,cr,erlst[cc]);
    else sprintf(str,"lin %d cra %d : %d sub-addresses\n>",ln,cr,n);
    ttyoutput(str);
    return n;
  }
}
