/** \file client.c
 * \brief TINE Client-side API Routines
 *      
 * Below are the principal client-side API routines which allow the user to obtain data from
 * elements in the control system.  Data acquistion can either be synchronous (e.g. ExecLink()) 
 * where execution is blocked until a call completes, or asynchronous (e.g. AttachLink()) where
 * the results of a call are given to a callback routine upon completion.  Data can also be 
 * 'monitored' asynchronously either at a supplied polling rate or upon data change.  
 *
 */
#ifndef NO_DOXYGEN
#include "cprolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include "servdef.h"
#include "brdcast.h"
#include "database.h"
#include "srvip.h"
#if defined(IPXNW) || defined(IPXBSD)
#  include "srvipx.h"
#endif
#if defined(UNIX) && !defined(NOREMOTE)
#  define UNIX_TTY
#  include "unixcmd.h"
#endif
#if defined(MSDOS)
#  include "doscmd.h"
#endif
#if defined(WINNT)
#  include "ntcmd.h"
#endif

int PrepRequest(int num, int *id);
int doClient(int num, int *id);
void CloseLink(int i);
void doIPXConsumer(void);
int initIPClient(unsigned short port);
void doUDPConsumer(void);
void DataDump(char *txt,char *data,int siz,short fmt,short dir);

short RPCLastErrorSource = 0;
int doConsumerSemaphore = 0;
int ExecLinkSemaphore = 0;
int AttachLinkSemaphore = 0;
int recvNetGlobalSemaphore = 0;
short RPCLastError = 0;
int redirectedEntrySlot = -1;
int connectionTableID, nConnectionTableEntries = 0;
unsigned long nConnectionTimeouts = 0,nArrivals = 0;
int SystemLocalDataTimeSynchronization = FALSE;
int SystemSupressHeartbeatNotification = FALSE;
char fecaddr[30],fecprot[4];
extern char timebuf[9],databuf[9];
int MaxNumConnections = MAX_ENTRIES;
ConnTableEntry **connTable;
GroupTableEntry *grpTable = NULL;
extern ConnTableEntry connNameServer;
#ifndef clnWorkArea
BYTE *clnWorkArea = NULL;
unsigned long clnWorkAreaSize = MIN_WORK_SIZE;
#endif
BYTE *clientbuf;

char *modetoa[] = { "CANCEL","SINGLE","REFRESH","POLL","REGISTER","INVALID","INVALID","INVALID" };

GlobalListStruct **NGlobalList;
int GlobalHeartbeat = GLOBALHEARTBEAT;
/* 
 * General Purpose Routines:
 */
int outputConnectionList(FILE *fp)
{
  int k,baseMode;
  char Buffer[256],linkstatus[5];
  ConnTableEntry *c;

  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," LINK"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[24],"Contracts\n");
  fprintf(fp,Buffer);
# ifdef UNIX_TTY
  if (fp == stdout) rmtoutput(Buffer);
# endif

  for (k=0; k<nConnectionTableEntries; k++)
  {
     memset(Buffer,' ',80); c = connTable[k];
     sprintf(&Buffer[0]," %s (%d)",RPCFec[c->FECindex].Name,c->RPCprotocol);
     Buffer[strlen(Buffer)] = ' ';
     baseMode = BASEMODE(c->mode);
     strcpy(linkstatus, baseMode < CM_SINGLE ? "-" : c->tocounter ? "DOWN" : "UP");
     sprintf(&Buffer[24],"[%d] %s %s %s %ld msec (%s) - %s\n",
	     k,c->name,c->deviceName,c->prop,c->pollRate,modetoa[baseMode],linkstatus);
     fprintf(fp,Buffer);
#    ifdef UNIX_TTY
     if (fp == stdout) rmtoutput(Buffer);
#    endif
  }
  return 0;
}
int outputConnectionStats(int k)
{
  int baseMode;
  char Buffer[256],linkstatus[5];
  ConnTableEntry *c;

  if (k < 0 || k >= nConnectionTableEntries)
  {
    sprintf(Buffer,"connection %d is not available\n>",k);
    ttyoutput(Buffer);
    sprintf(Buffer,"available connections are %d to %d\n>",k,nConnectionTableEntries-1);
    ttyoutput(Buffer);
    return 0;
  }
  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," Front End"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[24],"Contract\n>");
  ttyoutput(Buffer);

  memset(Buffer,' ',80); c = connTable[k];
  sprintf(&Buffer[0]," %s (%d)",RPCFec[c->FECindex].Name,c->RPCprotocol);
  Buffer[strlen(Buffer)] = ' ';
  baseMode = BASEMODE(c->mode);
  strcpy(linkstatus, baseMode < CM_SINGLE ? "-" : c->tocounter ? "DOWN" : "UP");
  sprintf(&Buffer[24],"[%d] %s %s %s %ld msec (%s) - %s\n>",
	     k,c->name,c->deviceName,c->prop,c->pollRate,modetoa[baseMode],linkstatus);
  ttyoutput(Buffer);

  ttyoutput("\n>");
  sprintf(Buffer,"%15s%15s%15s%12s%8s%11s","Link Starttime","Last Arrived","Last Timestamp","Is Pending","Counter","Timeouts\n>");
  ttyoutput(Buffer);
  sprintf(Buffer,"%15ld%15ld%15ld%12s%8d%9d\n>",
     c->starttime,c->time,c->dtimestamp,c->pending ? "YES" : "NO",c->counter,c->tocounter);
  ttyoutput(Buffer);

  ttyoutput("\n>");
  sprintf(Buffer,"%15s%15s%15s%12s%14s","Link Heartbeat","Array Size","Num Blocks","Is Grouped","Callback ID\n>");
  ttyoutput(Buffer);
  sprintf(Buffer,"%15d%15ld%15d%12s%12lu\n>",
     (int)c->heartbeat,c->sizeOut,c->numblks,c->grouped ? "YES" : "NO",c->callbackID);
  ttyoutput(Buffer);

  return 0;
}
int dumpConnectionList(void)
{
  return outputConnectionList(stdout);
}
int dumpConnectionStats(int i)
{
  return outputConnectionStats(i);
}
int initClient(void)
{
  static int initialized = 0;
  int cc;

  if (initialized) return 0;
  if (startuptime == 0) startuptime = time(NULL);
  whoami(gUserName);
  if ((connTable=(ConnTableEntry **)SystemCalloc(MaxNumConnections,sizeof(ConnTableEntry *))) == NULL)
    return out_of_client_memory;
  if ((cc=initFecTable()) != 0) return cc;/* first call initializes the Fec Table */
  if ((cc=initTransport(0)) != 0) return cc;
  if (clnWorkArea == NULL)
  {
    feclog("Client Work Space not allocated !!");
    return un_allocated;
  }

  initialized = 1;
  return 0;
}
void closeClients(void)
{
  int i,n;

  /* For packing efficiency do this in two steps: */
  for (n=0; n<nConnectionTableEntries; n++) if (connTable[n]->RPCprotocol > 3) CloseLink(n);
  for (n=0; n<nConnectionTableEntries; n++)
  {
    CheckSubscriptions(i);
    if (ReturnToClientCycle)
    {
      if (ReturnToClientCycle%2) hardDelay(20);
      ReturnToClientCycle = 0;
      n--;
    }
  }
  ShutdownClients(TRUE);
}
GroupTableEntry *findGroup(UINT32 id)
{
  GroupTableEntry *gt;

  for (gt=grpTable; gt && gt->GrpID != id; gt=gt->next);

  if (!gt)
  {
    if ((gt=(GroupTableEntry *)SystemCalloc(1,sizeof(GroupTableEntry))) == NULL) return NULL;
    gt->GrpID = id;
    gt->next = grpTable;
    grpTable = gt;
  }
  return gt;
}
extern int NameServerLoaded;
int isNameServerRequest = FALSE;
int DataTagging = FALSE;
int lastOutputDataLink = -1;
REDIRECTED_LINK *RedirectedLinkList = NULL;
REDIRECTED_LINK *appendRedirectedLinkList(char *con,char *srv,char *prp,char *dev,char *dstsrv,char *dstprp)
{
  REDIRECTED_LINK *r;
  if ((r=(REDIRECTED_LINK *)calloc(1,sizeof(REDIRECTED_LINK))) == NULL) return NULL;
  strncpy(r->srcContext,con,16);
  strncpy(r->srcServerName,srv,16);
  strncpy(r->srcProperty,prp,32);
  strncpy(r->srcDeviceName,dev,32);
  strncpy(r->dstServerName,dstsrv,16);
  if (strlen(dstprp) > 0) strncpy(r->dstProperty,dstprp,32);
  else strncpy(r->dstProperty,prp,32);
  r->nxt = RedirectedLinkList;
  if (RedirectedLinkList != NULL) RedirectedLinkList->prv = r;
  RedirectedLinkList = r;
  if (NGdebug) 
  {
    dbglog("append /%s/%s/%s/%s -> /%s/%s/%s/%s to redirection table",
      con,srv,prp,dev,con,dstsrv,r->dstProperty,dev);
  }
  return r;
}
int removeRedirectedLink(REDIRECTED_LINK *r,int cc)
{
  if (NGdebug) 
  {
    dbglog("removing /%s/%s/%s/%s from redirection table",
      r->srcContext,r->dstServerName,r->dstProperty,r->srcDeviceName);
  }
  if (r->prv != NULL) r->prv->nxt = r->nxt;
  else if (r == RedirectedLinkList) RedirectedLinkList = r->nxt;
  free(r);
  return 0;
}
int redirectLink(char *con,char *srv,char *prp,char *dev)
{
  REDIRECTED_LINK *r=NULL;

  for (r=RedirectedLinkList; r != NULL; r = r->nxt)
  {
    if (strncmp(r->srcContext,con,16)) continue;
    if (strncmp(r->srcServerName,srv,16)) continue;
    if (strncmp(r->srcProperty,prp,32)) continue;
    if (strncmp(r->srcDeviceName,dev,32)) continue;
    strncpy(srv,r->dstServerName,16);
    strncpy(prp,r->dstProperty,32);
    if (NGdebug)
    {
      dbglog("redirecting /%s/%s/%s/%s -> /%s/%s/%s/%s from redirection table",
        r->srcContext,r->srcServerName,r->srcProperty,dev,
        r->srcContext,r->dstServerName,r->dstProperty,dev);
    }
    return -1;
  }
  return 0;
}
void markTCPConnectionAsIdle(int slot)
{
  int i;
  if (connTable[slot]->TCPsocket > 0)       /* mark the connection as idle ? */
  {
     for (i=0; i<nConnectionTableEntries; i++)    /* run thru the list again */
     {
       if (i != slot && connTable[i]->TCPsocket == connTable[slot]->TCPsocket) break;
     }
     if (i == nConnectionTableEntries)                    /* didn't find it */
     {                /* the last entry holding the socket is being re-used */
       AddIdleConnection(connTable[slot]->TCPsocket,connTable[slot]->FECindex);
     }
  }
  return;
}
int getConnectionEntry(DTYPE *dout,DTYPE *din,char *devcontext,
    char *devgrp,char *devname,char *property)
{
  static int initialized = 0;
  UINT32 doutsize,dinsize;
  UINT16 pktlen;
  int cc,i,fmtsizein,rpcidx = -1,fecidx = -non_existent_fec,slot = 0;
  short EqNo = 0;
  struct structStruct *s;
  BYTE *dataPtr;
  char fecname[20],eqpname[8];
  char lgroup[20],lproperty[36];
  time_t t = time(NULL);
  UINT16 datamtu;

  if (NGdebug) feclog("connect to %s %s %s",devgrp,devname,property);
  if (!initialized)
  {
    initialized = 1;
    if ((cc=initClient()) != 0) return -cc;
  }
  if (devname && devname[0] == '#') EqNo = atoi(&devname[1]);
  if (NameServerLoaded && !strcmp(devgrp,CSHOSTS)) /* Name Server request */
  {
    /* NOTE: direct CSHOSTS requests do not handle input data sets (no swapping, etc) */
    prepNameServerEntry(0);
    return 0;
  }
  /* has this been redirected ? */
  strncpy(lgroup,devgrp,16); lgroup[16] = 0;
  strncpy(lproperty,property,32); lproperty[32] = 0;
  redirectLink(devcontext,lgroup,lproperty,devname);
  /* resolve FEC name and EQP name: */
  if ((fecidx=locateTagName(lgroup,devcontext,&rpcidx)) < 0) return fecidx;
  if (rpcidx < 0) return -non_existent;
  strncpy(fecname,RPCFec[fecidx].Name,16);
  strncpy(eqpname,RPCData[rpcidx].EqName,EQP_NAME_SIZE);
  /* check to see if in connection Table */
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (i > 0 && slot == 0 &&
        connTable[i]->heartbeat < SUB_PENDING &&
        connTable[i]->needsNotification == FALSE &&
        connTable[i]->mode == CM_CANCEL) slot = i;     /* slot is available */
    if (strncmp(devname,connTable[i]->deviceName,32)) continue;
    if (devcontext != NULL && strncmp(devcontext,connTable[i]->context,16)) continue;
    if ((!strncmp(fecname,RPCFec[connTable[i]->FECindex].Name,16) &&
         !strncmp(eqpname,RPCData[connTable[i]->RPCindex].EqName,EQP_NAME_SIZE) &&
         !strncmp(lproperty,connTable[i]->prop,EQP_PROP_SIZE)) ||
        (connTable[i]->mode != CM_CANCEL && connTable[i]->redirectID > 0 &&
         !strncmp(connTable[i]->redirectName,devgrp,16) &&
         !strncmp(connTable[i]->redirectProp,property,32)))
    {
      if (dout)
      {
        if (dout->dFormat != connTable[i]->formatOut) continue;
        if ((dout->dFormat == CF_STRUCT || DataTagging) &&
	          strncmp(dout->dTag,connTable[i]->strTagOut,8)) continue;
        if (dout->dFormat == CF_STRUCT && (s=getStructFromTag(dout->dTag)) != NULL)
        {
          if (dout->dArrayLength*s->rawlen != connTable[i]->sizeOut) continue;
        }
        else
        {
          if (dout->dArrayLength != connTable[i]->sizeOut) continue;
        }
      }
      else if (connTable[i]->sizeOut) continue;

      if (din)
      {
        if (connTable[i]->dataIn == NULL) continue;
        if (din->dFormat != connTable[i]->formatIn) continue;
        if ((din->dFormat == CF_STRUCT || DataTagging) &&
            strncmp(din->dTag,connTable[i]->strTagIn,8)) continue;
        if (din->dFormat == CF_STRUCT && (s=getStructFromTag(din->dTag)) != NULL)
        {
          if (din->dArrayLength*s->rawlen != connTable[i]->sizeIn) continue;
          AssignClnStructDataPtr(dataPtr,connTable[i]->dataIn,RTOL);
          if (memcmp(din->data.vptr,dataPtr,(size_t)din->dArrayLength*s->size)) continue;
        }
        else
        {
          if (din->dArrayLength != connTable[i]->sizeIn) continue;
          dinsize = connTable[i]->sizeIn*fmtsizeof(connTable[i]->formatIn);
          if (memcmp(din->data.vptr,connTable[i]->dataIn,(size_t)dinsize)) continue;
        }
      }
      else if (connTable[i]->sizeIn) continue;

      if (dout != NULL)
      {
        if (BASEMODE(connTable[i]->mode) != CM_CANCEL)  /* active link */
        {
          if (NGdebug) feclog("re-direct active %s link",modetoa[BASEMODE(connTable[i]->mode)]);
          lastOutputDataLink = i;
          return -link_exists;                          /* redirect output data */
        }
        else                                            /* not active link */
        {
          if (dout->data.vptr != connTable[i]->dataOut) /* different stored buffer */
            continue;
        }
      }
      if (NGdebug) feclog("re-use inactive link");
      connTable[i]->redirectedLink = NULL;
      connTable[i]->redirectID = 0;
      connTable[i]->redirectName[0] = connTable[i]->redirectProp[0] = 0;
      connTable[i]->callbackID = 0;
      return i;                                         /* re-use inactive link */
    }
  }
  /* brace ourselves for a new entry: */
  if (redirectedEntrySlot >= 0 && redirectedEntrySlot < nConnectionTableEntries)
  {
    slot = redirectedEntrySlot;
    markTCPConnectionAsIdle(slot);
    connTable[slot]->redirectedLink = NULL;
  }
  else if (slot == 0)
  {
    slot = nConnectionTableEntries;
    if ((cc=AssignConnTableMemory(nConnectionTableEntries)) != 0) return -cc;
  }
  else
  {
    markTCPConnectionAsIdle(slot);
    if (connTable[slot]->allocated && connTable[slot]->dataOut)
      SystemFree(connTable[slot]->dataOut);
    if (connTable[slot]->dataIn) SystemFree(connTable[slot]->dataIn);
    if (connTable[slot]->bufOut) SystemFree(connTable[slot]->bufOut);
    if (connTable[slot]->bufOutBlks) SystemFree(connTable[slot]->bufOutBlks);
    if (connTable[slot]->bufCmp) SystemFree(connTable[slot]->bufCmp);
    if (connTable[slot]->dtf) SystemFree(connTable[slot]->dtf);
    memset(connTable[slot],0,sizeof(ConnTableEntry));
    connTable[slot]->dataOut = NULL;
    connTable[slot]->dataIn  = NULL;
    connTable[slot]->bufOut  = NULL;
    connTable[slot]->bufOutBlks= NULL;
    connTable[slot]->bufCmp  = NULL;
    connTable[slot]->dtf = NULL;
    connTable[slot]->redirectedLink = NULL;
    connTable[slot]->redirectID = 0;
    connTable[slot]->redirectName[0] = connTable[slot]->redirectProp[0] = 0;
    connTable[slot]->callbackID = 0;
  }
  /* assign data indices and other parameters in connection Table: */
  if (devcontext != NULL)
    strncpy(connTable[slot]->context,(char *)devcontext,16);
  else connTable[slot]->context[0] = 0;
  strncpy(connTable[slot]->name,(char *)lgroup,16);
  strncpy(connTable[slot]->deviceName,(char *)devname,32);
  strncpy(connTable[slot]->prop,(char *)lproperty,32);
  connTable[slot]->EqNo = EqNo;
  connTable[slot]->formatIn = din ? din->dFormat : CF_NULL;
  connTable[slot]->formatOut = dout ? dout->dFormat : CF_NULL;
  if (din && (din->dFormat == CF_STRUCT || DataTagging))
    strncpy(connTable[slot]->strTagIn,din->dTag,8);
  if (dout && (dout->dFormat == CF_STRUCT || DataTagging))
    strncpy(connTable[slot]->strTagOut,dout->dTag,8);
  connTable[slot]->RPCindex = rpcidx;
  connTable[slot]->FECindex = fecidx;
  connTable[slot]->sizeIn  = din ? din->dArrayLength : 0;
  connTable[slot]->sizeOut = dout ? dout->dArrayLength : 0;
  connTable[slot]->RPCprotocol = 5;
  connTable[slot]->mode = CM_REGISTER;
  connTable[slot]->blknumIn  = 1;
  connTable[slot]->numblksIn = 1;
  connTable[slot]->nextDataPtr = 0;
  connTable[slot]->blkidIn = 0;
  pktlen = RPCFec[fecidx].Prot == IPX ? IPX_BUFFER_SIZE : UDP_BUFFER_SIZE;
  connTable[slot]->datamtu = pktlen-SUBSCRIPTION_SIZE-HEADERSTRUCT_SIZE;
  connTable[slot]->protocol = (RPCFec[fecidx].Prot & IPX) ? IPX : UDP ;
  if (connTable[slot]->starttime == t) connTable[slot]->starttime++;
  else connTable[slot]->starttime = t;
  connTable[slot]->dtimestamp = 0;
  connTable[slot]->dtimestampMSEC = 0;
  connTable[slot]->dataFlg = link_not_open;
  if (redirectedEntrySlot >= 0 && redirectedEntrySlot < nConnectionTableEntries)
  {
    redirectedEntrySlot = -1;
    return slot;  /* return here and don't mess with the memory ...*/
  }
  if (din && din->data.vptr != NULL)
  {
    if (din->dFormat == CF_STRUCT && (s=getStructFromTag(din->dTag)) != NULL)
    {
      connTable[slot]->sizeIn = din->dArrayLength*s->rawlen;
    }
    fmtsizein = fmtsizeof(din->dFormat);
    if ((dinsize=connTable[slot]->sizeIn*fmtsizein) == 0)
    {
      connTable[slot]->dataIn = NULL;
    }
    else 
    {
      if ((connTable[slot]->dataIn=(char *)SystemCalloc(1,(size_t)dinsize)) == NULL)
        return -out_of_client_memory;
      if (fmtsizein > 0) connTable[slot]->datamtu -= connTable[slot]->datamtu % fmtsizein;
      cpyClnStructData(connTable[slot]->dataIn,din->data.vptr,(size_t)dinsize);

      connTable[slot]->BytesTotal = dinsize;
      connTable[slot]->numblksIn = (UINT16)(dinsize/connTable[slot]->datamtu +
                                            (dinsize%connTable[slot]->datamtu ? 1 : 0));
      connTable[slot]->BytesRemaining = connTable[slot]->BytesTotal;
      connTable[slot]->nextDataSize = MIN(connTable[slot]->datamtu,connTable[slot]->BytesRemaining);
    }
  }
  else connTable[slot]->dataIn = NULL;
  if (dout)
  {
    if (dout->dFormat == CF_STRUCT && (s=getStructFromTag(dout->dTag)) != NULL)
    {
      connTable[slot]->sizeOut = dout->dArrayLength*s->rawlen;
      doutsize = dout->dArrayLength*s->size;
    }
    else
    {
      doutsize = connTable[slot]->sizeOut*fmtsizeof(dout->dFormat);
      if (dout->dFormat == CF_SPECTRUM) doutsize =+ 96;
    }
    if (dout->data.vptr == NULL) /* unassigned pointer */
    {
      if ((connTable[slot]->dataOut=(char *)SystemCalloc(1,(size_t)doutsize)) == NULL)
        return -out_of_client_memory;
      dout->data.vptr = connTable[slot]->dataOut;
      connTable[slot]->allocated = TRUE;
    }
    else connTable[slot]->dataOut = (char *)dout->data.vptr;
    datamtu = UDP_BUFFER_SIZE - PRODUCER_SIZE - sizeof(UINT16);
    connTable[slot]->numblks = (UINT16)(doutsize/datamtu + (doutsize%datamtu ? 1 : 0));
    if (connTable[slot]->numblks)
    {
      connTable[slot]->bufOut = (char *)SystemCalloc(1,(size_t)doutsize);
      connTable[slot]->bufOutBlks = (BYTE *)SystemCalloc(connTable[slot]->numblks,1);
    }
  }
  else connTable[slot]->dataOut = NULL;

  if (slot == nConnectionTableEntries) IncrementConnections(nConnectionTableEntries);
  return slot;            /* everythings OK, return list No. */
}
int hotRpc(char **Values, char *eqpname, int EqNo, char *property, int rwFlg,
		 int format, int pollingRate, void (*function)(int,int))
{ /* legacy cover routine */
  int id;

  if ((id=Subscribe(Values,eqpname,EqNo,property,rwFlg,format,
		   pollingRate,function,CM_POLL)) > 0) return id;
  return -id;
}
int syncCount = 0;                /* no synchronous activity */
void SyncAsync(int id,int cc)
{
  if (id < nConnectionTableEntries && id >= 0) 
  {
    connTable[id]->dataFlg = cc;
  }
  syncCount--;
}
int saRPC(char *Values, char *eqpname, int EqNo, char *property, int rwFlg, int format, int datasize)
{ /* legacy cover routine */
  int cc,id,mode;
  int to = 1,isRetry=0;
  time_t interim = time(NULL);

  if (abs(rwFlg)*fmtsizeof(format) > datasize) return -buffer_too_small;
retry:
  mode = CM_SINGLE;
  if (isRetry) mode |= CM_RETRY;
  if ((id=Subscribe(&Values,eqpname,EqNo,property,rwFlg,format,0,SyncAsync,mode)) < 0) return -id;
  for (syncCount=1; syncCount>0 && time(NULL) <= interim+to;) WaitForSyncCount(syncCount);
  if (syncCount)
  {
    isRetry = TRUE;
    to += REPLY_PENDING;
    interim = time(NULL);
    goto retry;
  }
  if ((cc=connTable[id]->dataFlg) == use_stream_transport)
    connTable[id]->protocol = connTable[id]->protocol == IPX ? SPX : TCP;
  if (NGdebug && cc) dbglog("RPC: %s",RPCLastErrorString);
  return cc;
}
int aRPC(char *Values, char *eqpname, int EqNo, char *property, int rwFlg,
	 int format, int datasize, void (*function)(int,int))
{ /* legacy cover routine */
  if (abs(rwFlg)*fmtsizeof(format) > datasize) return -buffer_too_small;
  return Subscribe(&Values,eqpname,EqNo,property,rwFlg,format,0,function,CM_SINGLE);
}
int RPC(char *Values, char *eqpname, int EqNo, char *property, int rwFlg, int format, int datasize)
{ /* legacy cover routine */
  return saRPC(Values,eqpname,EqNo,property,rwFlg,format,datasize);
}
int getRpc(char *Values, char *eqpname, int EqNo, char *property, int rwFlg, int datasize)
{ /* legacy cover routine */
  return RPC(Values,eqpname,EqNo,property,rwFlg,CF_DOUBLE,datasize);
}
int PutRPCErrorValue(int i, void *errValue)
{
  if (i > nConnectionTableEntries || i < 0) return argument_list_error;
  if (connTable[i]->mode == CM_CANCEL) return ask_error;
  memcpy(connTable[i]->errValue,errValue,64);
  connTable[i]->useErrValue = TRUE;
	return 0;
}
int Subscribe(char **Values,char *eqpname,int EqNo,char *property,int rwFlg,
	      int format,int pollingRate,void (*function)(int,int),int mode)
{
  DTYPE dout,din;
  int i;
  short access;
  char devname[16];

  if (mode > CM_SINGLE && rwFlg < 0) return -illegal_read_write;
  dout.dArrayLength = abs(rwFlg);
  dout.dFormat = format;
  dout.data.vptr = (void *)(*Values);
  if (rwFlg < 0)
  {
    if (*Values == NULL) return -out_of_range;
    din.data.vptr = (void *)(*Values); din.dArrayLength = abs(rwFlg);
    din.dFormat = format;
    access = CA_READ|CA_WRITE;
  }
  else
  {
    din.data.vptr = NULL; din.dArrayLength = 0;
    din.dFormat = CF_NULL;
    access = CA_READ;
  }

  sprintf(devname,"%s/#%d",eqpname,EqNo);
  pollingRate = pollingRate ? pollingRate : 1000;
  mode = mode > CM_REGISTER ? CM_REGISTER : mode < CM_CANCEL ? CM_CANCEL : mode;
  i = AttachLinkEx(devname,property,&dout,&din,access,pollingRate,function,mode,0);
  *Values = connTable[i]->dataOut;
  return i;
}
int PrepSubBuffer(BYTE *data)
{
  HeaderStruct hTmp,*h = &hTmp;
  short FECindex = -1,i,j,k,pktlen=0,dsize,totalsize = HEADERSTRUCT_SIZE;
  short stepsize=0,prot=0;
  struct subscription4 sTmp, *s = &sTmp;
  struct subscription  sTmpEx, *sEx = &sTmpEx;
  ConnTableEntry *c=NULL;
  time_t timestamp = time(NULL);
  char tin[9],tout[9];
  TCPBCKT *bckt;
  IDLE_CONNECTION *ic=NULL;
  int sck=0,transport=0;
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  char ipcpipe[64];
# endif

  if (data == NULL) return -1;
  AssignOutgoingHeader(h,data);
  for (i=0,j=0; i<nConnectionTableEntries; i++) if ((c=connTable[i])->RPCprotocol >= 4)
  {
    if (c->heartbeat < SUB_PENDING) continue;
    if (!j)
    {
      FECindex=c->FECindex;
      pktlen = RPCFec[FECindex].Prot == IPX ? IPX_BUFFER_SIZE : UDP_BUFFER_SIZE;
      memset(data,0,pktlen);
      if ((prot=c->RPCprotocol) == 4)
      {
        stepsize = SUBSCRIPTION4_SIZE;
        c->datamtu = pktlen-stepsize-HEADERSTRUCT_SIZE;
        c->numblksIn = 1;
      }
      else stepsize = SUBSCRIPTION_SIZE;
    }
    else if (FECindex != c->FECindex) continue;

    if (c->mode & CM_CONNECT && RPCFec[FECindex].Prot != PIPE)
    {
      if (transport == UDP || transport == IPX) continue;
      if (c->TCPsocket == 0)                      /* not yet connected */
      {
        for (k=0; k<nConnectionTableEntries; k++)
          if (connTable[k]->FECindex == FECindex &&
              connTable[k]->TCPsocket != 0)
        {
          c->TCPsocket = connTable[k]->TCPsocket; /* some one else is ! */
          break;
        }
      }
      if (c->TCPsocket == 0)                      /* still not connected */
      {
        if ((ic=getIdleConnection(c->FECindex)) != NULL)
        {
          c->TCPsocket = ic->sck;
          RemoveIdleConnection(ic);               /* no longer idle */
        }
      }
      if (c->TCPsocket == 0)                      /* still not connected */
      {
        if (connectTCP(gtTCPPort,i) == 0)         /* success */
        {
          if ((bckt=getBucket(c->TCPsocket)) != NULL)
          {
            bckt->idx = FECindex;
            RPCFec[FECindex].Prot |= TCP;
            c->protocol = TCP;
          }
          else CloseConnectionSocket(c->TCPsocket);
        }
      }
      sck = c->TCPsocket;
      transport = TCP;
    }
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    else if (RPCFec[FECindex].Prot == PIPE) 
    {
      memcpy(&sck,RPCFec[FECindex].Network,4);
      if (sck == 0)
      {
        sprintf(ipcpipe,"%s%s-server.ipc",IPC_PATH,RPCFec[FECindex].Name);
        if ((sck=CMDconnect(ipcpipe)) < 0) 
        {
          perror("IPC Socket");
          sck = 0;
        }
        else
        {
          memcpy(RPCFec[FECindex].Network,&sck,4);
          if (NGdebug) dbglog("FEC %s: attach pipe\n>",RPCFec[FECindex].Name);
          if ((bckt=getBucket(sck)) != NULL) bckt->idx = FECindex;
          else close(sck);
        }
      }
      transport = PIPE;
    }
#   endif
    else
    {
      if (transport == TCP || transport == PIPE) continue;
      sck = 0;              /* let SendSubBuffer figure it out */
      transport = (RPCFec[FECindex].Prot & IPX) ? IPX : UDP;
    }
    if (c->sizeIn && c->dataIn)
    {
      dsize = (short)c->nextDataSize;
      if ((totalsize + stepsize + dsize) > pktlen) break;
      PrepOutgoingData((char *)&data[totalsize+stepsize],&c->dataIn[(int)c->nextDataPtr],(size_t)dsize,c->formatIn);
    }
    else dsize = 0;
    if (dsize % 2 != 0) dsize++;                         /* make sure of even boundaries */

    AssignOutgoingSubscription(prot,i,c,RPCData[c->RPCindex].EqName,&data[totalsize]);
    PrepOutgoingSubscription(prot,&data[totalsize]);
    j++;
    totalsize += stepsize + dsize;
    if (NGdebug)
    {
      memset(tin,0,9);
      if (c->strTagIn[0] != 0) strncpy(tin,c->strTagIn,8);
      memset(tout,0,9);
      if (c->strTagOut[0] != 0) strncpy(tout,c->strTagOut,8);
      dbglog("subs %s %s %s (%ld %d %s %ld %d %s %d of %d) [%s] (%ld ms): %s",
              c->name,c->prop,c->deviceName,
              c->sizeOut,c->formatOut,tout,c->sizeIn,c->formatIn,tin,
              c->blknumIn,c->numblksIn,
              RPCFec[FECindex].Name,c->pollRate,modetoa[BASEMODE(c->mode)]);
    }
    if (NGdebug > 3 && c->dataIn)
      DataDump("data sent: ",c->dataIn,(int)c->sizeIn,c->formatIn,LTOR);
    ReturnToClientCycle = 0;
    if (c->blknumIn == c->numblksIn)
    {
      c->blknumIn = 1;
      c->nextDataPtr = 0;
      c->blkidIn = (++(c->blkidIn))%1000;
      c->BytesRemaining = c->BytesTotal;
      c->heartbeat = BASEMODE(c->mode) != (UINT16)CM_REFRESH ? (UINT32)(2+c->pollRate/1000) :
         c->dataFlg == (short)link_not_open ? (UINT32)REPLY_PENDING : (UINT32)(2+HEARTBEAT);
      c->counter = -1;
      c->time = timestamp;
    }
    else if (c->dataIn && c->numblksIn > c->blknumIn)
    {
      ReturnToClientCycle = c->numblksIn - c->blknumIn;
      c->blknumIn++;
      c->nextDataPtr += c->nextDataSize;
      c->BytesRemaining = c->BytesTotal - c->nextDataPtr;
    }
    c->nextDataSize = MIN(c->datamtu,c->BytesRemaining);

    if ((totalsize + stepsize) > pktlen) break;  /* no more packing possible */
  }

  h->TotalSize = totalsize;
  memcpy(h->UserName,gUserName,16);
  h->Protocol = prot;
  h->EqpSize = j;
  PrepOutgoingHeader(h,data);
  if (FECindex >= 0) 
  {
    if (SendSubBuffer(FECindex,sck,transport,clnWorkArea) != 0)
    {
      if (transport != PIPE) c->heartbeat = SUB_PENDING;  /* set it back */
    }
  }
  return FECindex;
}
int SendSubBuffer(int FECindex,int sck,int transport,BYTE *data)
{
  int cc;
  UINT16 poff;
  static CLN cln;
# if defined(IPXNW) || defined(IPXBSD)
  static BYTE reg_sock[2];
# endif
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  static UINT32 addr = 0;
# endif

  memset(&cln,0,sizeof(CLN));
  strcpy(cln.userName,RPCFec[FECindex].Name);
  cln.transport = transport;
  poff = GetPortOffset(&RPCFec[FECindex]);
  switch (transport)
  {
   case IPX:
     if (!IPXloaded) return ipx_not_supported;
#    if defined(IPXNW) || defined(IPXBSD)
     reg_sock[0] = REG_SOCKET + (BYTE)poff;
     reg_sock[1] = REG_SOCKET + (BYTE)poff;
     memcpy(&cln.IPXaddress,RPCFec[FECindex].Network,10);
     memcpy(cln.IPXaddress.socket,reg_sock,2);
     memcpy(cln.node,RPCFec[FECindex].ImmediateAddress,6);
     cln.sck = IPXConsumerSocket;
#    endif
     break;
   case UDP:
     if (!TCPloaded) return tcp_not_supported;
     memset(&prod_name,0,sizeof(prod_name));
     memcpy(&prod_name.sin_addr,&RPCFec[FECindex].IPh_addr,sizeof(struct in_addr));
     prod_name.sin_family=AF_INET;
     prod_name.sin_port = htons((UINT16)(gtUDPPort + poff));
     memcpy(&cln.IPaddress,&prod_name,sizeof(prod_name));
     cln.sck = ConsumerSocket;
     break;
   case TCP:
     if (!TCPloaded) return tcp_not_supported;
     memset(&prod_name,0,sizeof(prod_name));
     memcpy(&prod_name.sin_addr,&RPCFec[FECindex].IPh_addr,sizeof(struct in_addr));
     prod_name.sin_family=AF_INET;
     prod_name.sin_port = htons((UINT16)(gtTCPPort + poff));
     memcpy(&cln.IPaddress,&prod_name,sizeof(prod_name));
     cln.sck = sck;
     break;
#  if defined(UNIX) && !defined(NO_TINE_PIPE)
   case PIPE:
     if (addr == 0) addr = getmyipaddr();
     cln.sck = sck;
     memcpy(&cln.IPaddress.sin_addr,&addr,4);
     break;
#  endif
   case   0:
   default:
     if (NGdebug > 1) dbglog("FEC %s: acquire protocol\n>",RPCFec[FECindex].Name);
     return not_implemented;
   }
   if ((cc=SendToPeer(&cln,data)) != 0)
   {
     feclog("Send Subscription: %s\n>",erlst[cc]);
#    if defined(UNIX) && !defined(NO_TINE_PIPE)
     close(sck); freeBucket(sck);
     memset(RPCFec[FECindex].Network,0,4);
     RPCFec[FECindex].Prot = 0;
#    endif
     return cc;
   }
   return 0;
}
extern struct sockaddr_in IncomingIPfec;
extern IPXAddress IncomingIPXfec;
UINT32 IncomingNetworkProtocol;
int getIncomingFecIndex(void)
{
  int i,idx;
  UINT16 poff;

  if (IncomingNetworkProtocol == UDP || IncomingNetworkProtocol == TCP)
  {
    for (i=0; i<nConnectionTableEntries; i++) if (connTable[i]->protocol == UDP)
    {
      idx = connTable[i]->FECindex;
      poff = GetPortOffset(&RPCFec[idx]);
      if (IncomingIPfec.sin_port == htons((UINT16)(gtUDPPort + poff))
          && !memcmp(&IncomingIPfec.sin_addr.s_addr,&RPCFec[idx].IPh_addr,4) )
        return idx;
    }
    return -1;
  }
  if (IncomingNetworkProtocol == IPX)
  {
    for (i=0; i<nConnectionTableEntries; i++) if (connTable[i]->protocol == IPX)
    {
      idx = connTable[i]->FECindex;
      if (!memcmp(RPCFec[idx].Network,&IncomingIPXfec,10)) return idx;
    }
    return -1;
  }
  return -1;
}
extern int fecNameId;
#define TO_THRESHOLD 10
void resetLinkTOcounters(int FECindex)
{
  int i;
  ConnTableEntry *c;
  for (i=0; i<nConnectionTableEntries; i++)
    if ((c=connTable[i])->FECindex == FECindex) c->tocounter = 0;
}
int allLinksAboveThreshold(int FECindex)
{
  int i;
  ConnTableEntry *c;
  for (i=0; i<nConnectionTableEntries; i++)
    if ((c=connTable[i])->FECindex == FECindex && 
         BASEMODE(c->mode) > CM_SINGLE &&
         c->tocounter < TO_THRESHOLD) return 0;
  return 1;
}
UINT32 nPartialTransfers = 0;
void CheckConnections(void)
{
  int i;
  time_t timestamp = time(NULL);
  static time_t lasttime = 0, lastaddrreq = 0;
  ConnTableEntry *c;
  IDLE_CONNECTION *ic=NULL;
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  int pipe;
# endif

  for (i=0; i<nConnectionTableEntries; i++) if ((c=connTable[i])->RPCprotocol >= 4)
  {
    if (BASEMODE(c->mode) && c->heartbeat &&  /* heartbeat = 0 -> REGISTER */
        timestamp > (long)(c->time + c->heartbeat))
    {
      if (NGdebug > 1)
        dbglog("%s %s %s : %ld > %ld + %d",
          c->name,c->deviceName,c->prop,timestamp,c->time,(int)c->heartbeat);
      RPCLastErrorSource = RPC_ERROR_LCL;
      memcpy(RPCLastErrorString,erlst[link_not_open],RPCERR_SIZE-1);
      connectionTableID = i; c->linkStatus = c->dataFlg = link_not_open;
      NotifyClient(i,link_not_open);
      if (c->bytesin > 0) nPartialTransfers++;
      c->dtimestamp = 0L; c->dtimestampMSEC = 0;
      c->blksin = 0; c->bytesin = 0;
      nConnectionTimeouts++;
      if (BASEMODE(c->mode) > CM_SINGLE) c->heartbeat = SUB_PENDING;
      else
      {
        c->heartbeat = 0;
        c->mode = CM_CANCEL;
      }
      c->time = timestamp;
#     if defined(UNIX) && !defined(NO_TINE_PIPE)
      if (RPCFec[c->FECindex].Prot == PIPE) 
      {
        memcpy(&pipe,RPCFec[c->FECindex].Network,4);
        memset(RPCFec[c->FECindex].Network,0,4);
        if (pipe > 0)
        {
          close(pipe); freeBucket(pipe);
        }
        if (NGdebug)
        {
          feclog("FEC %s : reset pipe due to timeout",RPCFec[c->FECindex].Name);
        }
      }
#     endif
      if (strncmp(c->name,CSHOSTS,16)) /* not the name server */
      {                                /* locateFecName will call CheckConnections too */
        c->tocounter++;
        if (c->tocounter > TO_THRESHOLD*3 &&
            allLinksAboveThreshold(c->FECindex) &&
            fecNameId == -1 && c->redirectID == 0 && timestamp != lastaddrreq)
        {
          resetLinkTOcounters(c->FECindex);
          feclog("Re-acquire (%d) %s",i,RPCFec[c->FECindex].Name);
          RPCFec[c->FECindex].Prot = 0; /* forces address re-aquisition */
          if (NameServerLoaded) asyncLocateTagName(RPCData[c->RPCindex].ExportName,RPCData[c->RPCindex].Context);
          else locateTagName(RPCData[c->RPCindex].ExportName,RPCData[c->RPCindex].Context,NULL);
          lastaddrreq = timestamp;
        }
      }
      else
      {
        feclog("%s : timed out!",CSHOSTS);
      }
    }
  }
  for (ic=tcpIdleConnectionList; ic != NULL; )
  {
    ic->ttl -= (timestamp - lasttime)*1000;
    if (ic->ttl < 0) 
    {
      closeIPConnectionToFec(ic->idx);
      ic = RemoveIdleConnection(ic); 
    }
    else ic = ic->nxt;
  }
  for (i=0; i<nglobals; i++)             /* raise alarm for missing heartbeat */
  {
    if ((NGlobalList[i]->fcn != NULL) &&
        timestamp > NGlobalList[i]->time + NGlobalList[i]->maxPeriod*GLOBALGRACE)
    {
      (NGlobalList[i]->fcn)((int)NGlobalList[i]->callbackID,link_not_open);
      NGlobalList[i]->time = timestamp;
    }
  }
  if (timestamp != lasttime)
  {
    lasttime = timestamp;
  }
  return;
}
void doConsumer(void)
{
  int i,saveClientCycle;

  if (isNameServerRequest)      /* leave door open for sync ens requests */
  {
    CheckSubscriptions(i);
    ConsumerCheck(TRUE);
    CheckConnections();                       /* missing heartbeats ?   */
  }
  if (!doConsumerSemaphore)
  {
    doConsumerSemaphore = 1;
    ConsumerCheck(FALSE);                  /* flush input on the socket */
    CheckSubscriptions(i);
    ConsumerCheck(TRUE);
    CheckConnections();                       /* missing heartbeats ?   */
    doConsumerSemaphore = 0;
  }
  checkClientNotification(); /* alwasy check, irrespective of Semaphore */
  return;
}
void checkClientNotification(void)
{
  int i;
  ConnTableEntry *c;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    c = connTable[i];
    if (c->needsNotification) NotifyClient(i,c->linkStatus);
  }
}
void applyErrorValue(ConnTableEntry *c,void *errValue)
{
  int i,fmt = c->formatOut,len = fmtsizeof(fmt);
  BYTE *ptr;
  if (errValue == NULL) return;
  if (len < 1 || len > 64) return;
  for (i=0; i<(int)c->sizeOut; i++) 
  {
    ptr = (BYTE *)(c->dataOut + i*len);
    memcpy(ptr,errValue,len);
  }
}
void NotifyClient(int i,int cc)
{
  ConnTableEntry *c = connTable[i];
  GroupTableEntry *g = NULL;
  REDIRECTED_LINK *r;

  if (NGdebug > 2)
    dbglog("%s %s %s <%d>: %s",
		  c->name,c->deviceName,c->prop,cc,c->pending ? "notification" : "not pending");
  if (!c->pending) return;
  c->dataFlg = cc;
  connectionTableID = i;
  c->pending = 0;
  c->needsNotification = FALSE;
  if (cc != 0)
  {
    if ((cc == illegal_property || cc == non_existent || 
         cc == non_existent_elem || cc == non_existent_fec) &&
        (r=c->redirectedLink) != NULL)
    {
      c->redirectedLink = NULL;
      removeRedirectedLink(r,cc);
    }
    if (!(cc & CE_SENDDATA) && c->useErrValue)
    {
      applyErrorValue(c,c->errValue);
    }
  }
  if (c->grouped)
  {
    for (g=grpTable; g && g->GrpID != (UINT32)c->fcn; g = g->next);
    if (g == NULL) c->grouped = 0;
    else
    {
      g->GrpCc |= cc;
      if (--(g->numPending) > 0) return;
      if (BASEMODE(c->mode) <= CM_SINGLE) /* will be marked for cancellation */
           g->numInGrp = 0;               /* so reduce the group size to 0 */
      g->numPending = g->numInGrp;
      cc = g->GrpCc;
      g->GrpCc = 0;
    }
  }
  LastCompletionDataType = c->formatOut;
  LastCompletionDataSize = c->lastDataSize;
  gDataTimeStamp = putDataTimeStamp(0,c->dtimestamp,c->dtimestampMSEC);
  if (c->fcn != NULL) (c->fcn)((int)c->callbackID,cc);
  for (i=0; i<nConnectionTableEntries; i++) if (connTable[i]->fcn == c->fcn)
    connTable[i]->pending = 1;
}
int _datcmp(char *d1,char *d2,int dsz,short fmt,DataToleranceFactor *dtf)
{
  int i,n;
  float delta = 0.0, epsilon = 0.0;

  if (dtf == NULL) return memcmp(d1,d2,dsz);
  if (fmt < CF_DOUBLE || fmt > CF_FLOAT || fmt == CF_TEXT) return memcmp(d1,d2,dsz);
  n = dsz/fmtsizeof(fmt);
  switch (fmt)
  {
    case CF_DOUBLE:
      for (i=0; i<n; i++) 
      {
        delta = (float)fabs(((double *)d1)[i] - ((double *)d2)[i]);
        epsilon = (float)fabs(dtf->percent * ((double *)d2)[i]) + dtf->absolute;
        if (delta > epsilon) return TRUE;
      }
      break;
    case CF_FLOAT:
      for (i=0; i<n; i++) 
      {
        delta = (float)fabs(((float *)d1)[i] - ((float *)d2)[i]);
        epsilon = (float)fabs(dtf->percent * ((float *)d2)[i]) + dtf->absolute;
        if (delta > epsilon) return TRUE;
      }
      break;
    case CF_SHORT:
      for (i=0; i<n; i++) 
      {
        delta = (float)abs(((short *)d1)[i] - ((short *)d2)[i]);
        epsilon = (float)abs((int)(dtf->percent * (float)((short *)d2)[i])) + dtf->absolute;
        if (delta > epsilon) return TRUE;
      }
      break;
    case CF_LONG:
      for (i=0; i<n; i++) 
      {
        delta = (float)abs(((SINT32 *)d1)[i] - ((SINT32 *)d2)[i]);
        epsilon = (float)abs((int)(dtf->percent * (float)((SINT32 *)d2)[i])) + dtf->absolute;
        if (delta > epsilon) return TRUE;
      }
      break;
    case CF_BYTE:
    default:
      return memcmp(d1,d2,dsz);
  }
  return FALSE;
}
#define bowout(txt) {\
   if (NGdebug) { sprintf(dbgbuf,"subscription %d %ld",i,PheaderEx->timestamp);\
     if (c != NULL) sprintf(&dbgbuf[strlen(dbgbuf)],"(%s %s %s %ld)",c->name,c->deviceName,c->prop,c->dtimestamp);\
     sprintf(&dbgbuf[strlen(dbgbuf)],": %s",txt);\
     dbglog(dbgbuf);\
   }\
   goto bow;\
}
void consumeData(BYTE *buf)
{
  short i,m,cc=0,rc=0,fmtsize,totalsize,bytesread = sizeof(totalsize);
  short idx,prot,stepsize,baseMode,renewalMultiplier,supressNotify = 0;
  long gapsize, gapptr;
  UINT32 datasize,dcmpsize,dataPtr,doutsize,dretsize;
  ProducerHeaderStruct4 pTmp, *Pheader = &pTmp;
  ProducerHeaderStruct pTmpEx, *PheaderEx = &pTmpEx;
  ConnTableEntry *c = NULL;
  time_t timestamp = time(NULL);
  BYTE *srcbuf;
  char namstr[64],devstr[64],str[32];
  DTYPE dout, din;
  struct structStruct *s=NULL;

  nArrivals++;
  if ((totalsize = PeekWord(*(short *)buf)) <= 0) return;

  switch (cc=PeekWord(*(short *)&buf[6]))
  {
    case server_redirection:
      if ((i=PeekWord(*(short *)&buf[4])) < 0 || i >= nConnectionTableEntries) ccerr(non_existent);
      if (connTable[i]->mode == CM_CANCEL) ccerr(non_existent);
      memcpy(namstr,&buf[bytesread+PRODUCER_SIZE],RPCERR_SIZE);
      c = connTable[i]; strncpy(str,namstr,16); str[16] = 0;
      strncpy(c->redirectName,c->name,16);
      if (strlen(c->context) > 0) sprintf(devstr,"/%s/%s/%s",c->context,str,c->deviceName);
      else sprintf(devstr,"%s/%s",str,c->deviceName);
      strncpy(str,&namstr[16],16);
      if (strlen(str) == 0) strncpy(str,c->prop,32);
      strncpy(c->redirectProp,c->prop,16);
      dout.dArrayLength = c->sizeOut;
      dout.dFormat = c->formatOut;
      memcpy(dout.dTag,c->strTagOut,8);
      dout.data.cptr = c->dataOut;
      din.dArrayLength = c->sizeIn;
      din.dFormat = c->formatIn;
      memcpy(din.dTag,c->strTagIn,8);
      din.data.cptr = c->dataIn;
      if (NGdebug) dbglog("redirect %s/%s %s to %s %s",c->redirectName,c->deviceName,c->redirectProp,devstr,str);
      c->redirectID = redirectedEntrySlot = i;
      c->pending = 0;
      markTCPConnectionAsIdle(i);
      c->TCPsocket = 0;
      m = c->mode;
      if ((i=AttachLinkEx(devstr,str,&dout,&din,c->access,(int)c->pollRate,c->fcn,m,c->callbackID)) >= 0)
      {
        c->redirectedLink = appendRedirectedLinkList(c->context,c->redirectName,c->redirectProp,c->deviceName,namstr,&namstr[16]);
      }
      return;
    case illegal_protocol:
      if ((idx=getIncomingFecIndex()) < 0) ccerr(invalid_index);
      for (i=0; i<nConnectionTableEntries; i++) if ((c=connTable[i])->FECindex == idx && c->RPCprotocol == 5)
      {
        c->RPCprotocol = 4;
        c->heartbeat = SUB_PENDING;
        if (c->formatOut == CF_TEXT) c->formatOut = 1;
        if (NGdebug) dbglog("%s %s %s : switch to protocol 4",c->name,c->deviceName,c->prop);
      }
      return;
    case get_subscription_id:
      if ((idx=getIncomingFecIndex()) < 0) ccerr(invalid_index);
      if ((i=PeekWord(*(short *)&buf[4])) < 0 || i >= nConnectionTableEntries) ccerr(non_existent);
      connTable[i]->srvID = PeekWord(*(short *)&buf[bytesread+PRODUCER_SIZE]);
      if (NGdebug) dbglog("BCAST sub id %d from %s",connTable[i]->srvID,RPCFec[idx].Name);
      return;
    default:
      break;
  }

  if (PeekWord(*(short *)&buf[2+PRODUCER_SIZE-4]) == BCAST_ID) prot = 5;
  else if ((i=PeekWord(*(short *)&buf[4])) >= 0 && i < nConnectionTableEntries)
  {
    if ((prot=connTable[i]->RPCprotocol) < 4 || prot > 5) ccerr(illegal_protocol);
  }
  else ccerr(un_allocated);

  stepsize = prot == 4 ? PRODUCER4_SIZE : PRODUCER_SIZE;

  while (bytesread < totalsize)
  {
    PrepIncomingSubscription(prot,&buf[bytesread]);
    if (PheaderEx->ClnStarttime == BCAST_ID)   /* BCAST or MCAST contract */
    {
      if ((idx=getIncomingFecIndex()) < 0) return;         /* which FEC ? */
      for (i=0; i<nConnectionTableEntries; i++)        /* cross reference */
        if (connTable[i]->srvID == PheaderEx->SubscriptionID &&
           connTable[i]->FECindex ==  idx)
      {
        PheaderEx->SubscriptionID = i;        /* sync with local contract */
        PheaderEx->ClnStarttime = connTable[i]->starttime;
        break;
      }
      if (i == nConnectionTableEntries) return;/* not interested in this one */
    }
    i = PheaderEx->SubscriptionID;
    if (i < 0 || i >= nConnectionTableEntries) bowout("invalid id");
    if ((c=connTable[i])->starttime != PheaderEx->ClnStarttime) bowout("invalid starttime");
    if ((baseMode=BASEMODE(c->mode)) == CM_CANCEL) bowout("cancelled");
    if (PheaderEx->timestamp < c->dtimestamp) bowout("stale");
    c->numblks = PheaderEx->numblks;
    fmtsize = fmtsizeof(c->formatOut);
    doutsize = PheaderEx->msgsize - stepsize;
    datasize = fmtsize*c->sizeOut;
    if (c->formatOut == CF_SPECTRUM) datasize += 96;
    dcmpsize = datasize;
    if (dcmpsize%2) dcmpsize++;                         /* even boundaries */
    if (NGdebug > 2 || (NGdebug && baseMode <= CM_SINGLE && PheaderEx->numblks <= 4))
    {
      dbglog("%s (p%d %d of %d)-> %s %s [%s] (+%d) %d vals (%d)<%d>",
	      fecprot,prot,PheaderEx->blknum,PheaderEx->numblks,
	      c->name,c->prop,fecaddr,PheaderEx->counter,
	      (int)c->sizeOut,c->formatOut,
	      (int)(PheaderEx->CompletionCode & 0xfff));
    }
    RPCLastError = 0;
    RPCLastErrorSource = RPC_ERROR_RMT;
    memset(RPCLastErrorString,0,RPCERR_SIZE);
    c->lastDataSize = 0; gapsize = 0;
    switch (cc=PheaderEx->CompletionCode)
    {
      case non_existent_elem:
        if (strncmp(c->name,CSHOSTS,16))
        {
          c->tocounter++;
          if (c->tocounter > TO_THRESHOLD*3 &&
              allLinksAboveThreshold(c->FECindex) &&
              fecNameId == -1)
          {
            resetLinkTOcounters(c->FECindex);
            c->tocounter = 0;
            RPCFec[c->FECindex].Prot = 0; /* forces address re-aquisition */
            if (NameServerLoaded) asyncLocateTagName(RPCData[c->RPCindex].ExportName,RPCData[c->RPCindex].Context);
            else locateTagName(RPCData[c->RPCindex].ExportName,RPCData[c->RPCindex].Context,NULL);
          }
        }
        if (c->tocounter == 1)
        {
          feclog("%s/%s %s -> %s [%s] %s [%s]: %s",
            c->name,c->deviceName,c->prop,
            RPCData[c->RPCindex].ExportName,RPCData[c->RPCindex].EqName,
            RPCFec[c->FECindex].Name,fecaddr,erlst[cc]);
        }
      default:
        RPCLastError = cc;
        if (prot == 4) break;
        if (doutsize < RPCERR_SIZE && c->blksin == 0)
        {
          c->bytesin = 0;
          bowout("partial error block");
        }
        if (!(PheaderEx->CompletionCode & CE_SENDDATA)) datasize = 0;
        gapsize = 0;
        if (PheaderEx->blknum == PheaderEx->numblks || datasize == 0)
        {
          c->RPCLastErrorSource = RPC_ERROR_RMT;
		      memset(c->RPCLastErrorString,0,RPCERR_SIZE);
          gapsize = RPCERR_SIZE;
          if (doutsize < RPCERR_SIZE)
          {
            if (PheaderEx->numblks == 1) bowout("phantom block");
            gapsize = doutsize;
          }
          gapptr = RPCERR_SIZE - gapsize;
          memcpy(&c->RPCLastErrorString[(int)gapptr],&buf[bytesread+PheaderEx->msgsize-(int)gapsize],(size_t)gapsize);
          doutsize -= (UINT16)gapsize;
          c->bytesin += (UINT16)gapsize;
        }
        if (!(PheaderEx->CompletionCode & CE_SENDDATA)) break;
      case 0:
        if (fmtsize == 0) break;
        if (dcmpsize < doutsize)
        {
          cc = illegal_format;
          if (NGdebug) dbglog("data size mismatch: %lu vs %lu",datasize,doutsize);
        }
        else
        {
          if (PheaderEx->numblks == 1 && c->counter == PheaderEx->counter) bowout("phantom block");
          if (!c->blksin)                                    /* new set */
          {
            c->counter = PheaderEx->counter;
            if (c->bufOutBlks) memset(c->bufOutBlks,0,c->numblks);
          }
          if (c->counter != PheaderEx->counter)   /* we're being passed */
          {
            nPartialTransfers++;
            if (PheaderEx->counter < c->counter || PheaderEx->counter > c->counter + 30)
            {
              c->blksin = 0; c->bytesin = 0;
              c->counter = PheaderEx->counter;       /* more recent set */
              if (c->bufOutBlks) memset(c->bufOutBlks,0,c->numblks);
            }
            else bowout("old data block");       /* old stuff coming in */
          }
          if (PheaderEx->numblks > 1)               /* multiple blocks !*/
          {
            if (!c->bufOut && !(c->bufOut=(char *)SystemCalloc(1,(size_t)(datasize+RPCERR_SIZE)))) bowout("no local memory");
            if (!c->bufOutBlks && !(c->bufOutBlks=(BYTE *)SystemCalloc(c->numblks,1))) bowout("no local memory");
            if (PheaderEx->blknum > c->numblks) bowout("unexpected block");
            if (c->bufOutBlks[PheaderEx->blknum-1]) bowout("phantom block");
            c->bufOutBlks[PheaderEx->blknum-1] = 1;
            dataPtr = (PheaderEx->blknum-1)*PheaderEx->mtu;
            if (doutsize > 0) memcpy(&c->bufOut[(int)dataPtr],&buf[bytesread+stepsize],(size_t)doutsize);
            srcbuf = (BYTE *)c->bufOut;
            ReturnToClientCycle = PheaderEx->numblks - c->blksin;
          }
          else { dataPtr = 0; srcbuf = &buf[bytesread+stepsize]; }
          c->bytesin += doutsize;
          if (++(c->blksin) < PheaderEx->numblks) goto bow;
          c->RPCLastErrorSource = RPC_ERROR_RMT;
          memcpy(RPCLastErrorString,&c->RPCLastErrorString,32);
          ReturnToClientCycle = 0;
          dretsize = c->bytesin - (cc == 0 ? 0 : RPCERR_SIZE);
          if (dretsize <= 0) break;
          if (c->formatOut == CF_DEFAULT)
          {
            c->formatOut = LFMT(PheaderEx->EqpFormat); /* record the default */
            if ((fmtsize=fmtsizeof(c->formatOut)) == 0) bowout("illegal format");
            c->sizeOut = dretsize/fmtsize;             /* format and size */
          }
          PrepIncomingData(c->dataOut,srcbuf,(size_t)datasize,(size_t)dretsize,c->formatOut);
          c->lastDataSize = dretsize/fmtsize;
          supressNotify = CompIncomingData(c->dataOut,c->bufCmp,datasize,c->formatOut,c->dtf);
          if (!supressNotify && c->bufCmp != NULL) memcpy(c->bufCmp,c->dataOut,datasize);
          if (NGdebug > 3) DataDump("data received: ",c->dataOut,(int)c->sizeOut,c->formatOut,RTOL);
        }
        break;
      case use_stream_transport:
        if (baseMode > CM_SINGLE)
        {
	        c->RPCprotocol = 1;
	        bowout("convert to protocol 1");
        }
        break;
    }

    switch (baseMode)
    {
      default:
      case CM_SINGLE : c->mode = CM_CANCEL;  /* got reply */
		    break;
      case CM_REFRESH: c->heartbeat = 2 + HEARTBEAT;
		    break;
      case CM_POLL   : c->heartbeat = 2 + c->pollRate/1000;
		    break;
    }

    if (cc == non_existent_elem)
    {
      if (baseMode > CM_SINGLE && c->time != timestamp) c->heartbeat = SUB_PENDING; 
    }
    else c->tocounter = 0;
    c->linkStatus = cc;
    c->counter = PheaderEx->counter;
    c->dtimestamp = PheaderEx->timestamp;
    c->dtimestampMSEC = PheaderEx->timestampMSEC;
    c->time = timestamp;
    c->blksin = 0; c->bytesin = 0;
    if (c->bufOutBlks) memset(c->bufOutBlks,0,c->numblks);
    if (SystemLocalDataTimeSynchronization) /* don't let the incoming data be too far in the future ! */
    {
      gDataTimeStamp = makeDataTimeStamp();
      if (c->dtimestamp > (long)(gDataTimeStamp) + 1) 
      {
        feclog("adjusting incoming data timestamp for %s/%s[%s] from %ld to %ld",
          c->name,c->deviceName,c->prop,c->dtimestamp,(long)gDataTimeStamp);
        c->dtimestamp = (long)gDataTimeStamp;
        c->dtimestampMSEC = 0;
      }
    }
    if (!supressNotify) c->needsNotification = TRUE;
    if (!supressNotify && (c->mode & CM_SYNCNOTIFY)) NotifyClient(i,cc);

    if (baseMode > CM_SINGLE && PheaderEx->counter > 0)
    {
      renewalMultiplier = 1;
      if (c->pollRate < 1000 && c->pollRate >= 20) renewalMultiplier = (short)(1000/c->pollRate);
      if (PheaderEx->counter < REMINDER*renewalMultiplier) c->heartbeat = SUB_PENDING;
      else if (PheaderEx->counter <= ACK_REQUEST*renewalMultiplier) c->heartbeat = ACK_PENDING;
    }
bow:
    if ((bytesread += PheaderEx->msgsize) <= 0 || (bytesread%2 != 0)) return;
  }
  return;
err:
  if (cc && NGdebug)
  {
    rc = cc & 0xfff;
    if (strlen(RPCLastErrorString) == 0 && rc > 0 && rc < numErr)
      strncpy(RPCLastErrorString,erlst[rc],32);
    dbglog("ConsumeData : %s",RPCLastErrorString);
  }
  return;
}
double gSyncTimestamp = 0;
void cbGlobalSynchronization(int id, int cc)
{
  double delta_t = 0;
  static int last_cc = -1, ncorrections = 0;
  static double last_logged = 0;
  static double last_dataTimeOffset = 0, largest_correction = 0;

  if (cc != last_cc)
  {
    last_cc = cc;
    feclog("global synchronization : %s",erlst[cc]);
  }
  if (cc) return;
  if (gSyncTimestamp < 1000) return;                     /* this can't be */
  delta_t = gSyncTimestamp - gDataTimeStamp;
  if (delta_t < 0.5 && delta_t > -0.5) return;
  if (NGdebug) dbglog("global synchronization : %g sec",delta_t);
  gDataTimeStampOffset += delta_t;
  ncorrections++;
  if (fabs(delta_t) > largest_correction) largest_correction = fabs(delta_t);
  if (gDataTimeStamp < last_logged + 3600) return;
  last_logged = gDataTimeStamp;
  if (fabs(gDataTimeStampOffset-last_dataTimeOffset) < 0.01) return;
  feclog("global synchronization offset : %g sec",gDataTimeStampOffset);
  feclog("largest correction : %g sec  (nr. corrections %d)",largest_correction,ncorrections);
  last_logged = gDataTimeStamp; ncorrections = 0; largest_correction = 0;
  last_dataTimeOffset = gDataTimeStampOffset;
}
int SystemStartGlobalSynchronization(void)
{
  int cc = 0;
  DTYPE dout;

  dout.dArrayLength = 1;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = &gSyncTimestamp;
  dout.dTag[0] = 0;
  cc = recvNetGlobalEx("SYSTIME",&dout,cbGlobalSynchronization,0);
  if (cc < 0) feclog("could not start global synchronization : %s",erlst[-cc]);
  return cc < 0 ? -cc : 0;
}
void cbTimeSynchronization(int id,int cc)
{
  gDataTimeStamp = makeDataTimeStamp();
  cbGlobalSynchronization(id,cc);
}
int SystemStartTimeSynchronization(char *timesrv)
{
  int cc = 0;
  DTYPE dout;

  if (timesrv == NULL) ccerr(-argument_list_error);
  dout.dArrayLength = 1;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = &gSyncTimestamp;
  dout.dTag[0] = 0;
  cc = AttachLink(timesrv,"SYSTIME",&dout,NULL,CA_READ,5000,cbTimeSynchronization,CM_POLL);
err:
  if (cc < 0) feclog("could not start time synchronization : %s",erlst[-cc]);
  return cc < 0 ? -cc : 0;
}
void getGlobalUpdate(BYTE *buf)
{
  BrdcstData tmpBD, *brdcstdata = &tmpBD;
  HeaderStruct hTmp, *brdcstHeader = &hTmp;
  int num,size=HEADERSTRUCT_SIZE;
  int cc=0,i,j,fmtsize,dsize,datasize,dretsize;
  time_t t=time(NULL);
  ConnTableEntry *c = NULL;
  struct structStruct *s=NULL;

  BRDcount++; /* increment statistics counter */
  PrepIncomingHeader(brdcstHeader,buf);
  switch (brdcstHeader->Protocol)
  {
    case 3: break;
    default: ccerr(illegal_protocol);
  }
  num = brdcstHeader->EqpSize;
  for (i=0; i<num; i++)
  {
    PrepIncomingBrdcstData(brdcstdata,&buf[size])
    size += sizeof(BrdcstData);
    if ((fmtsize=fmtsizeof(brdcstdata->format)) < 0) ccerr(illegal_format);
    for (j=0; j<nglobals; j++)
    {
      if (!strcmp(brdcstdata->name,NGlobalList[j]->keyword))
      {
        NGlobalList[j]->time = t;
        dsize = MIN(NGlobalList[j]->size,brdcstdata->size);
        dretsize = fmtsize*dsize; 
        datasize = fmtsize*NGlobalList[j]->size;
        if (NGlobalList[j]->format != brdcstdata->format) 
        {
          if (NGlobalList[j]->refdata == NULL)
          {
            if ((NGlobalList[j]->refdata=(BYTE *)calloc(brdcstdata->size,fmtsize)) == NULL) 
            {
              if (NGlobalList[j]->fcn != NULL) (NGlobalList[j]->fcn)((int)NGlobalList[j]->callbackID,out_of_client_memory);
              continue;
            }
          }
          PrepIncomingData(NGlobalList[j]->refdata,&buf[size],datasize,dretsize,brdcstdata->format);
          cc = reformat(NGlobalList[j]->data,NGlobalList[j]->format,NGlobalList[j]->refdata,brdcstdata->format,brdcstdata->size);
        }
        else
        {
          PrepIncomingData(NGlobalList[j]->data,&buf[size],datasize,dretsize,brdcstdata->format);
          cc = 0;
        }
        gDataTimeStamp = makeDataTimeStamp();
        if (NGlobalList[j]->fcn != NULL) (NGlobalList[j]->fcn)((int)NGlobalList[j]->callbackID,cc);
      }
    }
    size += brdcstdata->size*fmtsize;
  }
err:
  if (NGdebug > 2)
  {
    sprintf(dbgbuf,"Received %s Broadcast packet",fecprot);
    if (cc) sprintf(&dbgbuf[strlen(dbgbuf)],"ERR: %s",erlst[cc]);
    dbglog(dbgbuf);
  }
  return;
}
void dumpGlobals(void)
{
  int i,j,dsize;
  char Buffer[256];

  for (i=0; i<nglobals; i++)
  {
    sprintf(Buffer,"\t%s : ",NGlobalList[i]->keyword);
    dsize = MIN(NGlobalList[i]->size,10);
    switch (NGlobalList[i]->format)
    {
      case CF_SHORT:
        for (j=0; j<dsize; j++) 
          sprintf(&Buffer[strlen(Buffer)],"%d ",((short *)NGlobalList[i]->data)[j]);
        break;
      case CF_LONG:
        for (j=0; j<dsize; j++) 
          sprintf(&Buffer[strlen(Buffer)],"%ld ",((long *)NGlobalList[i]->data)[j]);
        break;
      case CF_FLOAT:
        for (j=0; j<dsize; j++) 
          sprintf(&Buffer[strlen(Buffer)],"%g ",((float *)NGlobalList[i]->data)[j]);
        break;
      case CF_DOUBLE:
        for (j=0; j<dsize; j++) 
          sprintf(&Buffer[strlen(Buffer)],"%g ",((double *)NGlobalList[i]->data)[j]);
        break;
      case CF_NAME16:
        for (j=0; j<dsize; j++) 
          sprintf(&Buffer[strlen(Buffer)],"%s ",((NAME16 *)NGlobalList[i]->data)[j].name);
        break;
      case CF_NAME32:
        for (j=0; j<dsize; j++) 
          sprintf(&Buffer[strlen(Buffer)],"%s ",((NAME32 *)NGlobalList[i]->data)[j].name);
        break;
      case CF_TEXT:
        strncat(Buffer,(char *)NGlobalList[i]->data,MIN(NGlobalList[i]->size,200));
        break;
    }
    strcat(Buffer,"\n>");
    ttyoutput(Buffer);
  }
  return;
}
#endif /* NO_DOXYGEN */

int SetNotificationTolerance(int linkId, float toleranceAbsolute, float tolerancePercent)
{
  if (toleranceAbsolute <= 0 && tolerancePercent <= 0) return 0; /* no reason to allocate a dtf */
  if (linkId < 0 || linkId > nConnectionTableEntries-1) return out_of_range;
  connTable[linkId]->dtf = (DataToleranceFactor *)SystemCalloc(1,sizeof(DataToleranceFactor));
  if (connTable[linkId]->dtf == NULL) return out_of_local_memory;
  connTable[linkId]->dtf->absolute = toleranceAbsolute;
  connTable[linkId]->dtf->percent = tolerancePercent;
  return 0;
}
/* TINE Client API Calls follow: */
int AttachLinkEx(char *devName,char *devProperty,
	       DTYPE *dout,DTYPE *din,short access,
	       int pollingRate,void (*callback)(int,int),int mode,UINT32 callbackID)
{
  char fulname[72];
  int i=-1,grouped,supressDuplication,wait,baseMode,useOnError,dinsize=0,doutsize=0,cc=0;
  GroupTableEntry *gt = NULL;
  struct structStruct *s;
  char *cn = NULL,*en = NULL,*dn = NULL;
  if (devName == NULL || devProperty == NULL) ccerr(argument_list_error);
  strncpy(fulname,devName,71); fulname[71] = 0;
  if (getCdiLinkNameHook != NULL && (dn=getCdiLinkNameHook(devName)) != NULL)
  {
    cc = -not_implemented;
    if (cdiAttachLinkHook != NULL) cc = cdiAttachLinkHook(dn,devProperty,dout,din,access,pollingRate,callback,mode,callbackID);
    if (cc != -server_redirection) return cc;
    strncpy(fulname,dn,64);
    dn = NULL; cc = 0;
  }
  RPCLastErrorSource = RPC_ERROR_LCL;
  if (AttachLinkSemaphore && !strstr(devName,CSHOSTS)) return -semaphore_busy; else AttachLinkSemaphore = -1;

  supressDuplication = mode & CM_NODUPLICATES; 
  grouped = mode & CM_GROUPED; 
  wait = mode & CM_WAIT;
  useOnError = mode & CM_USEONERROR; 
  mode &= ~(CM_REMOVEVOLATILE); /* strip off these bits */
  if ((mode & CM_BCAST || mode & CM_MCAST) && !ListenForMCasts)
  {
    if ((cc=ListenForMulticasts(TRUE)) != 0) ccerr(cc);
  }
  baseMode = BASEMODE(mode);
  if (baseMode > CM_REGISTER  || baseMode < CM_CANCEL) ccerr(illegal_mode);
  if (baseMode == CM_REFRESH) grouped = 0;
  pollingRate = pollingRate ? pollingRate : 1000;
  if (SystemSupressHeartbeatNotification && (mode & CM_REFRESH)) supressDuplication = TRUE;

  if (din)
  {
    din->dTimeStamp = 0;
    dinsize = (int)(fmtsizeof(din->dFormat)*din->dArrayLength);
    if (din->dFormat == CF_STRUCT && (s=getStructFromTag(din->dTag)) != NULL) dinsize *= s->size;
  }
  if (dout)
  {
    dout->dTimeStamp = 0;
    doutsize = (int)(fmtsizeof(dout->dFormat)*dout->dArrayLength);
    if (dout->dFormat == CF_STRUCT && (s=getStructFromTag(dout->dTag)) != NULL) doutsize *= s->size;
    if (dout->dFormat == CF_SPECTRUM) doutsize += 96;
  }
  if (baseMode == CM_SINGLE && pollingRate < (doutsize+dinsize)/5)
    pollingRate += (doutsize+dinsize)/5;
  if (NGdebug)
  {
    feclog("Attach %s %s (%d msec): %s",devName,devProperty,pollingRate,modetoa[baseMode]);
  }
  if ((cn=strtok(fulname,DELIMITER)) == NULL) ccerr(argument_list_error);
  if ((en=strtok(NULL,DELIMITER)) == NULL) /* only one given */
  {
    en = cn; cn = NULL; dn = &fulname[71];
  }
  else if ((dn=strtok(NULL,DELIMITER)) == NULL) /* only two given */
  {
     if (cn == fulname) /* no leading delimiter */
     {
       dn = en; en = cn; cn = NULL;
     }
     else dn = &fulname[71];
  }
  if ((i=getConnectionEntry(dout,din,cn,en,dn,devProperty)) < 0)
  {
    if (i != -link_exists)
    {
      feclog("Unable to subscribe to %s %s: %s",devName,devProperty,erlst[-i]);
      if (callback != NULL && callbackID != 0) callback(callbackID,-i); /* in case of server_redirect */
      ccerr(-i);
    }
    if ((i=lastOutputDataLink) < 0) ccerr(link_exists);
    if (BASEMODE(connTable[i]->mode) == CM_SINGLE)  /* was a SINGLE link */
    {
      if (NGdebug) dbglog("attaching %s %s to link %d",devName,devProperty,i);
      if (dout->data.vptr != NULL) connTable[i]->dataOut = (char *)dout->data.vptr;
      else dout->data.vptr = (void *)connTable[i]->dataOut;
    }
    ccerr(link_exists);
  }
  if (callbackID == 0) callbackID = i;
  if (connTable[i]->RPCprotocol == 0) connTable[i]->RPCprotocol = 5;
  connTable[i]->pollRate  = pollingRate;
  connTable[i]->mode      = mode;
  connTable[i]->fcn       = callback;
  connTable[i]->callbackID= callbackID;
  connTable[i]->time      = time(NULL);
  connTable[i]->heartbeat = mode == CM_REGISTER ? 0 : SUB_PENDING;
  connTable[i]->access    = access;
  connTable[i]->pending   = 1;
  connTable[i]->needsNotification = FALSE;
  if (supressDuplication) connTable[i]->bufCmp = (char *)SystemCalloc(1,(size_t)doutsize);
  if (useOnError && dout->dArrayLength > 0) PutRPCErrorValue(i,dout->data.vptr);
  if (grouped && (gt=findGroup((UINT32)callback)) == NULL) grouped = 0;
  if (gt)
  {
    gt->numPending++;
    gt->numInGrp++;
    connTable[i]->grouped = 1;
  }
  if (wait)
  {
    connTable[i]->tocounter = -1;
    while (connTable[i]->tocounter == -1) WaitForSyncCount(syncCount);
  }
  ReturnToClientCycle = 1;
err:
  AttachLinkSemaphore = 0;
  return cc ? -cc : i;
}
int AttachLink(char *devName,char *devProperty,
	       DTYPE *dout,DTYPE *din,short access,
	       int pollingRate,void (*callback)(int,int),int mode)
{
  return AttachLinkEx(devName,devProperty,dout,din,access,pollingRate,callback,mode,0);
}
void CloseLink(int i)
{
  ConnTableEntry *c;
  if (cdiCloseLinkHook != NULL && cdiCloseLinkHook(i) == 0) return;
  if (i < 0 || i > nConnectionTableEntries) return; /* not a valid link table entry ! */
  c = connTable[i];

  if (NGdebug) feclog("Close Link [%d] %s %s %s",i,c->name,c->prop,c->deviceName);
  if (BASEMODE(c->mode) > CM_SINGLE)
  {
    if (c->mode & (CM_BCAST|CM_MCAST)) c->mode = CM_CANCEL | CM_MCAST;
    else c->mode = CM_CANCEL;
    if (c->TCPsocket > 0) c->mode |= CM_CONNECT;
    c->heartbeat = SUB_PENDING;
  }
  return;
}
int ExecLinkEx(char *devName,char *devProperty,
	     DTYPE *dout,DTYPE *din,short access,UINT16 timeout)
{
  int i,cc = 0,id;
  int mode,to,cnt=1,isRetry=0;
  struct timeval tv0,tv;
  char *dn=devName;
  if (devName == NULL || devProperty == NULL) return argument_list_error;
  if (getCdiLinkNameHook != NULL && (dn=getCdiLinkNameHook(devName)) != NULL)
  {
    cc = not_implemented;
    if (cdiExecLinkHook != NULL) cc = cdiExecLinkHook(dn,devProperty,dout,din,access,timeout);
    if (cc != server_redirection) return cc;
    cc = 0;
  }
  RPCLastErrorSource = RPC_ERROR_LCL;
  if (doConsumerSemaphore) return semaphore_error; /* no sync link inside callback! */
  if (ExecLinkSemaphore && !strstr(dn,CSHOSTS)) return semaphore_busy; else ExecLinkSemaphore = -1;
retry:
  gettimeofday(&tv0,(struct timezone *)NULL);
  tv = tv0;

  mode = CM_SINGLE;
  if (access & CA_CONNECT) mode |= CM_CONNECT;
  if (access & CA_SYNCNOTIFY) mode |= CM_SYNCNOTIFY;
  if (isRetry) mode |= CM_RETRY;
  if ((id=AttachLink(dn,devProperty,dout,din,access,timeout,SyncAsync,mode)) < 0)
  {
    if (id == -link_exists) /* redirect output */
    {
      if ((i=lastOutputDataLink) >= 0)
      {
        if (NGdebug) dbglog("redirecting %s %s to link %d",dn,devProperty,i);
        if (dout->data.vptr == NULL)
	        dout->data.vptr = (void *)connTable[i]->dataOut;
        else if (dout->data.vptr != connTable[i]->dataOut)
	        memcpy(dout->data.vptr,connTable[i]->dataOut,(size_t)connTable[i]->sizeOut*fmtsizeof(connTable[i]->formatOut));
        ccerr(0);
      }
    }
    ccerr(-id);
  }
  to = (int)connTable[id]->pollRate;
  for (syncCount=1; syncCount>0 && MSECS(tv,tv0) < to;)
  {
    WaitForSyncCount(syncCount);
    gettimeofday(&tv,(struct timezone *)NULL);
  }
  if (syncCount)                                  /* didn't come in ! */
  {
		/* if the caller wants this to be quick, don't retry! */
    if (cnt-- > 0 && to >= 1000)
    {
      connTable[id]->heartbeat = 0;
      connTable[id]->mode = CM_CANCEL;  /* CANCEL it in the link table ! */
      isRetry = TRUE;
      goto retry;
    }
    connTable[id]->dataFlg = connection_timeout;
    RPCFec[connTable[id]->FECindex].Prot = 0;/* forces address re-aquisition */
  }
  connTable[id]->heartbeat = 0;
  connTable[id]->mode = CM_CANCEL;        /* make sure it is CANCELled ! */
  cc = connTable[id]->dataFlg;
  if (cc != 0 && connTable[id]->redirectID > 0) cc |= CE_REDIRECTED;
  if (dout != NULL) 
  {
    dout->dTimeStamp = putDataTimeStamp(0,connTable[id]->dtimestamp,connTable[id]->dtimestampMSEC);
    if (dout->dFormat == CF_DEFAULT)
    {
      dout->dFormat = connTable[id]->formatOut;
      dout->dArrayLength = connTable[id]->sizeOut;
    }
  }
  if (NGdebug && cc) dbglog("ExecLink: %s",RPCLastErrorString);
err:
  ExecLinkSemaphore = 0;
  return cc;
}
int ExecLink(char *devName,char *devProperty,
	     DTYPE *dout,DTYPE *din,short access)
{
  char lclDevName[64];
  if (!strncmp(devName,"/LOCAL/",7) || !strncmp(devName,"\\LOCAL\\",7))
  {
    if (strlen(devName) > 64) return non_existent_elem;
    strcpy(lclDevName,&devName[7]);
    return ExecLocalLink(lclDevName,devProperty,dout,din,access);
  }
  return ExecLinkEx(devName,devProperty,dout,din,access,DEFAULT_CLIENT_POLLRATE);
}
char *GetRPCLastError(short cc)
{
  static char s[56];

  if (RPCLastError == cc) sprintf(s,"RMT: %s",RPCLastErrorString);
  else
  {
    if (cc < 0) cc = -cc;
    if (cc > numErr) sprintf(s,"LCL: (ERR %d undefined)",cc);
    else sprintf(s,"LCL: %s",erlst[cc]);
  }
  return s;
}
char *GetCompletionStatus(int i)
{
  if (i < 0 || i > nConnectionTableEntries) return NULL;
  return connTable[i]->RPCLastErrorString;
}
short GetCompletionSource(int i)
{
  if (i < 0 || i > nConnectionTableEntries) return -1;
  return connTable[i]->RPCLastErrorSource;
}
char *GetCompletionStatusFromCallbackId(int id)
{
  int i;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (connTable[i]->callbackID == (UINT32)id) return connTable[i]->RPCLastErrorString;
  }
  return NULL;
}
short GetCompletionSourceFromCallbackId(int id)
{
  int i;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (connTable[i]->callbackID == (UINT32)id) return connTable[i]->RPCLastErrorSource;
  }
  return -1;
}
UINT32 LastCompletionDataSize = 0;
int LastCompletionDataType = CF_NULL;
int GetCompletionDataSize(int i)
{
  if (i == -1) return (int)LastCompletionDataSize;
  if (i < 0 || i > nConnectionTableEntries) return -out_of_range;
  return (int)connTable[i]->lastDataSize;
}
int GetCompletionDataSizeFromCallbackId(int id)
{
  int i;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (connTable[i]->callbackID == (UINT32)id)
      return (int)connTable[i]->lastDataSize;
  }
  return -link_not_open;
}
int GetCompletionDataType(int i)
{
  if (i == -1) return (int)LastCompletionDataType;
  if (i < 0 || i > nConnectionTableEntries) return -out_of_range;
  return (int)connTable[i]->formatOut;
}
int GetCompletionDataTypeFromCallbackId(int id)
{
  int i;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (connTable[i]->callbackID == (UINT32)id)
      return (int)connTable[i]->formatOut;
  }
  return -link_not_open;
}
int recvNetGlobalEx(char *keyword,DTYPE *dout,void (*callback)(int,int),UINT32 callbackID)
{
  static int initialized = 0;
  int cc = 0,i = 0,rc;
  UINT16 dsize;

  if (recvNetGlobalSemaphore) return -semaphore_busy; else recvNetGlobalSemaphore = -1;

  if (!ListenForGlobals)
  {
    if ((cc=ListenForNetGlobals(TRUE)) != 0) ccerr(cc);
  }
  if (!initialized)
  {
    initialized = TRUE;
    if ((NGlobalList=(GlobalListStruct **)SystemCalloc(MaxNumGlobals,sizeof(GlobalListStruct *))) == NULL)
      ccerr(out_of_client_memory);
  }
  if (nglobals >= MaxNumGlobals) ccerr(resources_exhausted);
  for (i=0; i<nglobals; i++) if (!strcmp(NGlobalList[i]->keyword,keyword)) break;
  if (i < nglobals) return i; /* already in list */
  /* request for new global ... */
  if ((NGlobalList[i]=(GlobalListStruct *)SystemCalloc(1,sizeof(GlobalListStruct))) == NULL)
    ccerr(out_of_client_memory);
  nglobals++;
  if (callbackID == 0) callbackID = i;
  strncpy(NGlobalList[i]->keyword,keyword,16);
  NGlobalList[i]->size = (short)dout->dArrayLength;
  NGlobalList[i]->format = dout->dFormat;
  NGlobalList[i]->fcn = callback;
  NGlobalList[i]->callbackID = callbackID;
  NGlobalList[i]->maxPeriod = GlobalHeartbeat;
  dsize = (UINT16)(dout->dArrayLength*fmtsizeof(dout->dFormat));
  if (dout->data.vptr == NULL)
    if ((dout->data.vptr=(BYTE *)SystemCalloc(1,dsize)) == NULL)
      ccerr(out_of_client_memory);
  NGlobalList[i]->data = (BYTE *)dout->data.vptr;

  NGlobalList[i]->time = time(NULL);
  rc = ExecLinkEx(GLOBALSRV,NGlobalList[i]->keyword,dout,NULL,CA_READ,100);
  NGlobalList[i]->time = rc ? 0 : time(NULL);
  if (StartupDebug)
  {
    printf("Listening for global parameter : %s\n>",NGlobalList[i]->keyword);
  }
err:
  recvNetGlobalSemaphore = 0;
  return cc ? -cc : i;
}
int recvNetGlobal(char *keyword,DTYPE *dout,void (*callback)(int,int))
{
  return recvNetGlobalEx(keyword,dout,callback,0);
}
double getCurrentDataTimeStamp(int i)
{
  if (i < 0 || i > nConnectionTableEntries) return out_of_range;
  return putDataTimeStamp(0,connTable[i]->dtimestamp,connTable[i]->dtimestampMSEC);
}
double getCurrentDataTimeStampFromCallbackId(int id)
{
  int i;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (connTable[i]->callbackID == (UINT32)id)
      return putDataTimeStamp(0,connTable[i]->dtimestamp,connTable[i]->dtimestampMSEC);
  }
  return 0;
}
int GetCurrentLinkStatus(int i)
{
  if (i < 0 || i > nConnectionTableEntries) return out_of_range;
  return connTable[i]->linkStatus;
}
int GetCurrentLinkStatusFromCallbackId(int id)
{
  int i;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (connTable[i]->callbackID == (UINT32)id)
      return (int)connTable[i]->linkStatus;
  }
  return -link_not_open;
}
#ifndef SMALL_TINE_LIB /* for MSDOS, etc. ... */
int ModifyLinkAttributes(int i,short access,int pollingRate,void (*callback)(int,int),int mode,UINT32 callbackID)
{
  int isGrouped,wasGrouped,baseMode,cc = 0;
  GroupTableEntry *gt = NULL;

  if (i < 0 || i >= nConnectionTableEntries) return invalid_index;
  if (connTable[i]->mode == CM_CANCEL) return un_allocated;
  if ((wasGrouped=connTable[i]->grouped) != 0)
  {
    if ((gt=findGroup((UINT32)connTable[i]->fcn)) == NULL) wasGrouped = 0;
  }

  if (mode == 0)
  {
    mode = connTable[i]->mode;                                     /* keep it */
    isGrouped = connTable[i]->grouped;                             /* keep it */
  }
  else
  {
    isGrouped = mode & GROUPED; mode &= ~(GROUPED);
    if ((mode & CM_BCAST || mode & CM_MCAST) && !ListenForMCasts)
    {
      if ((cc=ListenForMulticasts(TRUE)) != 0) return cc;
    }
    baseMode = BASEMODE(mode);
    if (baseMode > CM_REGISTER  || baseMode < CM_CANCEL) return illegal_mode;
    if (baseMode == CM_REFRESH) isGrouped = 0;
  }
  if (pollingRate == 0) pollingRate = (int) connTable[i]->pollRate; /* keep it */
  if (access == 0) access = connTable[i]->access;                   /* keep it */
  if (callbackID == 0) callbackID = connTable[i]->callbackID;       /* keep it */
  if (callback == NULL) callback = connTable[i]->fcn;               /* keep it */

  connTable[i]->pollRate  = pollingRate;
  connTable[i]->mode      = mode;
  connTable[i]->fcn       = callback;
  connTable[i]->callbackID= callbackID;
  connTable[i]->heartbeat = mode == CM_REGISTER ? 0 : SUB_PENDING;
  connTable[i]->access    = access;
  connTable[i]->pending   = 1;

  if (!wasGrouped && isGrouped)
  {
    if ((gt=findGroup((UINT32)callback)) != NULL)
    {
      gt->numPending++;
      gt->numInGrp++;
      connTable[i]->grouped = 1;
    }
  }
  else if (wasGrouped && !isGrouped)
  {
    if (gt != NULL)
    {
      gt->numPending--;
      gt->numInGrp--;
    }
    connTable[i]->grouped = 0;
  }
  return 0;
}
#endif /* SMALL_TINE_LIB */
