/** \file srvdbase.c
 * \brief TINE server engine and registration API routines
 *
 * The API routines presented here deal almost exclusively with server-side registration calls.
 * All TINE servers must export their behavior in the form of properties, devices, server names,
 * etc.  The routines offered here provide ways to do this.  In most cases there are also 
 * alternatives based on startup configuration files such as fecid.csv, exports.csv, etc.  On
 * front-end systems where a local disk is available, this is the suggested registration 
 * method, since it avoids the hard coding of 'names' inside of the server programs themselves.
 *
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include "cprolog.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "servdef.h"
#include "threader.h"
#include "srvip.h"
#include "srvipx.h"
#include "database.h"
#include "errors.h"
#include "history.h"
#include "almsrv.h"
#include "brdcast.h"
#if defined(UNIX) && !defined(NOREMOTE)
#  define UNIX_TTY
#  include "unixcmd.h"
#endif

time_t appmktime = 0;
char appmkversion[64] = "1.00.0000";
char gFecName[20] =  "*unknown*";
char gDeviceContext[20] = "";
extern char *vFeclogBuffer;
int queryDeviceNamesFromProperties = FALSE;
int acquireDeviceNamesFromProperties = FALSE;

void haltServer(void)
{
  feclog("Stopping RPC Server ...\n>");
  ServerExitCondition = 4;
  closeServers(0);
  ExitEQPModules();
# ifndef MSDOS
  freeServerMemory();
# endif
  exit(0);
}
void ParseRedirector(char *str,ExportPropertyListStruct *prp)
{
  char *c = str, *sstr = NULL, *pstr = str, *dstr = NULL;

  if (str == NULL) return;
  if (str[0] == '/')                                         /* context */
  {
    if ((c=strchr(++c,'/')) == NULL) return;/* device server must be there !*/
    c++;
  }
  sstr = c;                                                /* device server */
  if ((c=strchr(sstr,'/')) != NULL)
  {
    *c++ = 0; dstr = c;                                      /* device name */
  }
  else
  {
    c = sstr;                                                /* put it back */
  }
  if ((pstr=strtok(c,"["))!= NULL) pstr = strtok(NULL,"]");
  strncpy(prp->prd->prpRedirection[0].name,sstr,16);
  if (pstr != NULL) strncpy(prp->prd->prpRedirection[1].name,pstr,16); /* redirected property (note 16-char len) */
  if (dstr != NULL) strncpy(prp->prd->prpRedirection[2].name,dstr,16); /* device name */
}
void ParsePropertyDescription(char *str,ExportPropertyListStruct *prp)
{
  char *c=NULL, *c1, tmp[64], *s = str;

  if (str == NULL) return;
  if ((c=strchr(str,'[')) != NULL)
  {
    if ((c1=strchr(++c,']')) != NULL)               /* 1st token: egu */
    {
      *c1++ = 0;
      str = c1;         
      if ((c1=strstr(c,"!")) != NULL)                  /* graph type */
      {
        *c1++ = 0; prp->egu.graph = getGraphType(c1);
      }
      if ((c1=strchr(c,':')) != NULL)               /* prp egu given */
      {
        c1++;
        prp->egu.min = (float)strtod(c,&c); c = c1;
        prp->egu.max = (float)strtod(c,&c); c++;
        if (c != str)
        {
          strncpy(tmp,c,64);
          strncpy(prp->egu.units,tmp,16);
        }
      }
    }
  }
  if (str != NULL && (c=strchr(str,'[')) != NULL)
  {
    if ((c1=strchr(++c,']')) != NULL)             /* 2nd token: x egu */
    {
      *c1++ = 0;
      str = c1;         
      if ((c1=strstr(c,"!")) != NULL)                  /* graph type */
      {
        *c1++ = 0; prp->xegu.graph = getGraphType(c1);
      }
      if ((c1=strchr(c,':')) != NULL)                 /* x egu given */
      {
        c1++;
        prp->xegu.min = (float)strtod(c,&c); c = c1;
        prp->xegu.max = (float)strtod(c,&c); c++;
        if (c != str)
        {
          strncpy(tmp,c,64);
          strncpy(prp->xegu.units,tmp,16);
        }
      }
    }
  }
  if (str == NULL) str = s;
  strncpy(prp->prpDescription,str,32);  /* 3rd token: description part */
}
#ifdef VXWORKS
# ifndef VXWBKG_FLAGS
#   define VXWBKG_FLAGS VX_FP_TASK
# endif
# ifndef VXWBKG_STACK
#   define VXWBKG_STACK 10000
# endif
# ifndef VXWBKG_PRIORITY
#   define VXWBKG_PRIORITY 100
# endif
void VXWbkgtsk(void (*tsk)(void),int rate);
#endif
void makeEqpModuleHashTable(ExportListStruct *el)
{
  HashXRefStruct *xref;
  char lclname[36];
  int cc=0,idx,i;

  if (el == NULL || el->EqpModuleXRef == NULL) return;
  lclname[32] = 0;
  for (i=0; i<el->EqpNumModules; i++)
  {
    strncpy(lclname,el->EqpModuleName[i].name,32);
    idx = (int)(ElfHash((unsigned char *)lclname) % MODULE_HASH_SIZE);
    for (xref=el->EqpModuleXRef[idx]; xref && xref->index != i; xref=xref->next);
    if (xref == NULL) /* it's a new one*/
    {
      if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) ccerr(out_of_server_memory);
      xref->index = i;
      xref->next = el->EqpModuleXRef[idx];
      el->EqpModuleXRef[idx] = xref;
    }
  }
err:
  if (cc) feclog("makeEqpModuleHashTable: %s\n",erlst[cc]);
  return;
}
int csvMatchExport(char *exp)
{
  ExportListStruct *el = NULL;

  if (exp == NULL) return 0;                    /* not given */
  for (el=ExportList; el != NULL; el=el->next)
    if (!strncmp(exp,el->tagName,ELEM_NAME_SIZE)) return 0; /* found it */
  return -1;                                    /* not a match */
}
void csvFecNameDone(struct CSVDBStruct *db)
{
  db = db;  /* avoid compiler warning */
  strncpy(gUserName,gFecName,16);
  FecNameRegistered = 1;
  return;
}
CSVCOLUMN csvFecNameColumns[] =
{
  { "FEC_NAME","",0,CF_TEXT,16,(BYTE *)gFecName,NULL,0 },
  { "OS",OS_TAG,0,CF_TEXT,12,(BYTE *)RPCFecEx.os,NULL,0 },
  { "PORT_OFFSET","0",0,CF_SHORT,1,(BYTE *)&PortOffset,NULL,0 },
  { "CONTEXT","TEST",0,CF_TEXT,16,(BYTE *)gDeviceContext,NULL,0 },
  { "EXPORT_NAME",NULL,0,CF_TEXT,16,NULL,csvMatchExport,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE *)RPCFecEx.desc,NULL,0 },
  { "LOCATION","",0,CF_TEXT,32,(BYTE *)RPCFecEx.loc,NULL,0 },
  { "HARDWARE","none",0,CF_TEXT,32,(BYTE *)RPCFecEx.hdw,NULL,0 },
  { "SUBSYSTEM","",0,CF_TEXT,16,(BYTE *)RPCFecEx.sub,NULL,0 },
  { "RESPONSIBLE","",0,CF_TEXT,32,(BYTE *)RPCFecEx.resp,NULL,0 },
};
CSVDB csvFecNameDb =
{
  csvFecNameColumns,
  sizeof(csvFecNameColumns)/sizeof(csvFecNameColumns[0]),
  0,
  NULL,
  csvStdIterator,
  csvFecNameDone,
  0,
  NULL,
  0
};
int getFecName(void)
{
  char *c,scratch[20];
  int cc=0;
  if (FecNameRegistered) return 0;
  cc = csvReadFile(FecDBpath,FECID_FILE,&csvFecNameDb,NULL);
  if (strlen(RPCFecEx.sub) == 0)
  {
    if ((c=strchr(RPCFecEx.desc,'[')) != 0)
    {
      strncpy(scratch,++c,20);
      if ((c=strchr(scratch,']')) != 0) *c=0;
      strncpy(RPCFecEx.sub,c,16);
    }
  }
  return cc;
}
void checkAccessLocks(time_t t)
{
  if (currentAccessLock.lockDuration < 0) return;  /* forever */
  if (t > currentAccessLock.lockStart + currentAccessLock.lockDuration)
    memset(&currentAccessLock,0,sizeof(currentAccessLock));
}
int IsMemberLockSet(CLN *cln)
{
  if (!currentAccessLock.lockType) return TRUE; /* unlocked */
  if (strncmp(currentAccessLock.cln.userName,cln->userName,16)) return FALSE;
  if (memcmp(&currentAccessLock.cln.IPXaddress,&cln->IPXaddress,sizeof(IPXAddress)))
    return FALSE;
  if (memcmp(&currentAccessLock.cln.IPaddress,&cln->IPaddress,sizeof(struct sockaddr_in)))
    return FALSE;
  return TRUE;
}
int DENYaccess = 1;
int WRITEaccess = 0;
int NrOfUsers = 0;
int NrOfDeniedUsers = 0;
UserListStruct *UserList = NULL;
UserListStruct *DeniedUserList = NULL;
int IsDeniedUser(char *usr)
{
  int i;
  for (i=0; i<NrOfDeniedUsers && strnicmp(usr,DeniedUserList[i].UserName,16); i++);
  return i == NrOfDeniedUsers ? 0 : -1;
}
int removeRegisteredUser(char *user)
{
  FILE *fp = NULL, *tmpfp = NULL;
  int i,len,cc=0;
  char *c,str[256],userfilename[128],tmpfilename[128];

  if (user == NULL || strlen(user) == 0) ccerr(argument_list_error);
  for (i=0; i<NrOfUsers;i++)
    if (!strnicmp(UserList[i].UserName,user,16)) break;
  if (i == NrOfUsers) ccerr(0);
  sprintf(userfilename,"%s%s",FecDBpath,USERS_FILE);
  strncpy(tmpfilename,userfilename,128);
  if ((c=strstr(tmpfilename,".csv")) == NULL) ccerr(file_error);
  *c = 0; strcat(tmpfilename,".bak");
  unlink(tmpfilename);
  rename(userfilename,tmpfilename);

  if ((tmpfp=fopen(tmpfilename,"r")) == NULL) ccerr(no_such_file);
  if ((fp=fopen(userfilename,"w")) == NULL) ccerr(no_such_file);

  while (fgets(str,255,tmpfp) != NULL)
  {
    if ((len=strlen(str)-1) <= 0) continue;
    if (!strnicmp(str,user,len)) continue;
    fwrite(str,strlen(str),1,fp);
  }
  fclose(fp); fp = NULL;
  fclose(tmpfp); tmpfp = NULL;
  getRegisteredUsers(&UserList,USERS_FILE,&NrOfUsers,"WRITE",&WRITEaccess);
err:
  if (fp != NULL) fclose(fp);
  if (tmpfp != NULL) fclose(tmpfp);
  return cc;
}
int appendRegisteredUserList(char *user)
{
  FILE *fp = NULL;
  int cc=0;
  char userfilename[128],usr[32];
  if (user == NULL || strlen(user) == 0) ccerr(argument_list_error);
  sprintf(userfilename,"%s%s",FecDBpath,USERS_FILE);
  if ((fp=fopen(userfilename,"a")) == NULL) ccerr(no_such_file);
  fseek(fp,0,SEEK_END);
  if (ftell(fp) == 0)
  {
    fprintf(fp,"USERNAME\n");  /* file did not exist before */
  }
  strncpy(usr,user,16); usr[16] = 0;
  fprintf(fp,"%s\n",usr);
  fclose(fp);
  getRegisteredUsers(&UserList,USERS_FILE,&NrOfUsers,"WRITE",&WRITEaccess);
err:
  return cc;
}
CSVCOLUMN csvUserFileColumns[] = 
{
  { "USERNAME","",0,CF_TEXT,16,NULL,NULL,0 }
};
CSVDB csvUserFileDb =
{
  csvUserFileColumns,
  sizeof(csvUserFileColumns)/sizeof(csvUserFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(UserListStruct)
};
int getRegisteredUsers(UserListStruct **ul,char *fn,int *nr,char *accstr,int *accflg)
{
  int i,cc = 0;
  char usr[32];

  cc = csvReadFile(FecDBpath,fn,&csvUserFileDb,(void **)ul);
  *nr = csvUserFileDb.siz;
  *accflg = cc == 0 ? 0 : 1;
  if (*accflg == 0)
  {
    for (i=0; i<csvUserFileDb.siz; i++)
    {
      strncpy(usr,(*ul)[i].UserName,16);
      feclog("%s has %s access",usr,accstr);
    }
    if (csvUserFileDb.siz == 0)
    {
      strcpy(usr,!strcmp(accstr,"WRITE") ? "NONE" : "EVERYONE");
      feclog("%s has %s access",usr,accstr);
    }
  }
  else
  {
    strcpy(usr,!strcmp(accstr,"WRITE") ? "EVERYONE" : "NONE");
    feclog("%s has %s access",usr,accstr);
  }
  return cc;
}  
void dumpRegisteredUsers(void)
{
  int i;
  char usr[32];
  if (WRITEaccess)
  {
    dbglog("Write Access open to all users\n");
    return;
  }
  dbglog("Write Access open to :\n");
  for (usr[16]=0,i=0; i<NrOfUsers; i++)
  {
    strncpy(usr,UserList[i].UserName,16);
    dbglog("\t%s\n",usr);
  }
}
int RegisterExport(char *tagname,char *eqpname,int numModules)
{
  int i, cc = 0;
  ExportListStruct *el;
  FunHead *header;
  int ftype;
  static int inside = 0;

  if (inside) return -1; else inside = -1;

  if (eqpname == NULL || strlen(eqpname) == 0)
  {
    feclog("RegisterExport: missing local equipment module name");
    ccerr(argument_list_error);
  }
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqpname,el->EqpName,EQP_NAME_SIZE)) break;
  if (el != NULL) ccerr(0);        /* already in list */
  el = (ExportListStruct *)SystemCalloc(1,sizeof(ExportListStruct));
  if (el == NULL) ccerr(out_of_server_memory);
  if (tagname == NULL || strlen(tagname) == 0)
  {
    feclog("RegisterExport: equipment module name %s not found",eqpname);
    ccerr(argument_list_error);
  }
  strncpy(el->tagName,tagname,ELEM_NAME_SIZE);
  strncpy(el->EqpName,eqpname,EQP_NAME_SIZE);
  strncpy(el->Context,gDeviceContext,EQP_CONTEXT_SIZE);
  if ((i=getAliasTableIndexFromName(el->EqpName)) >= 0)
    strncpy(el->EqpAlias,gAliasTable[i].alias,EQP_NAME_SIZE);
  if (!search_function(el->EqpName,&header,&ftype))
  {
    el->NdlFcn = (EQPCALL)header->F_entry;
  }
  el->EqpNumModules = numModules;
  el->EqpModuleName = GetEqpModuleNames(el->EqpName,el->EqpNumModules);
  if (el->EqpNumModules >= MODULE_HASH_PAYOFF)
  {
   if ((el->EqpModuleXRef == NULL) &&
       (el->EqpModuleXRef=(HashXRefStruct **)SystemCalloc(MODULE_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
     ccerr(out_of_server_memory);
    makeEqpModuleHashTable(el);
  }
  el->EqpNumProperties = 0;
  if ((el->almLst=(ALARM **)SystemCalloc(el->EqpNumModules,sizeof(ALARM *))) == NULL)
    ccerr(out_of_server_memory);
  el->exportId = NrOfEqpTypes++;
  el->next = ExportList;
  ExportList = el;

  getAlarmInfoTable(eqpname);

err:
  inside = 0;
  return cc;
}
int RegisterNodalModule(
    char *tagname,
    char *eqpname,
    int NumModules,
#   ifdef NODAL
    int (*fcn)(double *,int,int,char *),
#   else
    int (*fcn)(double *,short,short,short,char *),
#   endif
    void (*ini)(void),
    void (*tsk)(void),
    int rate,
    void (*exi)(void)
)
{
  int cc;
  ExportListStruct *el;
  FunHead *header;
  int ftype;
# ifdef VXWORKS
  char tasktag[16];
# endif
  if ((cc=RegisterExport(tagname,eqpname,NumModules)) != 0) return cc;
  if ((el=getExportListItem(eqpname)) == NULL) return non_existent_elem;

  el->NdlFcn  = fcn;
# ifndef VXWORKS
  el->eqpBkg  = tsk;
# else
  if (tsk != NULL)
  {
    sprintf(tasktag,"%s_task",eqpname);
    if (taskSpawn(tasktag,VXWBKG_PRIORITY,VXWBKG_FLAGS,VXWBKG_STACK,(FUNCPTR)VXWbkgtsk,(int)tsk,(int)rate,0,0,0,0,0,0,0,0) == ERROR)
      cc = not_running;
    feclog("task %s : %s\n>",tasktag,cc ? "not spawned !" : "spawned");
  }
  el->eqpBkg  = NULL;
# endif
  el->rateBkg = rate;
  if (el->eqpBkg && rate && ServerCycleMode != POLLING && rate < MaxPollingRate)
  {
    SysPoll = rate;
    MaxPollingRate = rate;
    if (rate < MinPollingRate) MinPollingRate = rate;
  }
  el->eqpIni = ini;
  el->eqpExi = exi;

  feclog("%12s legacy task registration: %s %s %s",eqpname,
          ini ? "INIT" : "----",
          tsk ? "BCKG" : "----",
          exi ? "EXIT" : "----");

  if (search_function(eqpname,&header,&ftype)) /* if not in NODAL list */
  {
    if (!search_function("DUMMY",&header,&ftype))       /* find a slot */
    {
      header->F_entry = el->NdlFcn;
      strncpy(header->F_name,eqpname,EQP_NAME_SIZE);
    }
  }
  if (SystemRunning && el->eqpIni) el->eqpIni();
  return 0;
}
int getRegisteredExports(void)
{
  FILE *fp;
  int i,idx,cc=0,done=0;
  int export_name_col=-1,local_name_col=-1,property_col=-1,property_size_col=-1,
      num_modules_col=-1,format_col=-1,access_col=-1,description_col=-1,
      property_id_col=-1,property_insize_col=-1,informat_col=-1,redir_col=-1;
  char str[256],hdr[256],scratch[64],expfilename[128], *c;
  FunHead *header;
  int ftype;
  long prpId;
  ExportPropertyListStruct *prp;
  ExportListStruct *el;
  PrpRedirBlk *prd = NULL;

  if (ExportList != NULL) SystemFree((char *)ExportList);

  sprintf(expfilename,"%s%s",FecDBpath,EXPORT_FILE);
  if ((fp=fopen(expfilename,"r")) == NULL) ccerr(no_such_file);
  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;    /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((export_name_col=findcol(hdr,"EXPORT_NAME")) < 0) ccerr(no_such_column);
      if ((local_name_col=findcol(hdr,"LOCAL_NAME")) < 0) ccerr(no_such_column);
      if ((num_modules_col=findcol(hdr,"NUM_MODULES")) < 0) ccerr(no_such_column);
      if ((property_col=findcol(hdr,"PROPERTY")) < 0) ccerr(no_such_column);
      if ((property_size_col=findcol(hdr,"PROPERTY_SIZE")) < 0) ccerr(no_such_column);
      property_insize_col = findcol(hdr,"PROPERTY_INSIZE");
      property_id_col = findcol(hdr,"PROPERTY_ID");
      if ((format_col=findcol(hdr,"FORMAT")) < 0) ccerr(no_such_column);
      informat_col = findcol(hdr,"INFORMAT");
      if ((access_col=findcol(hdr,"ACCESS")) < 0) ccerr(no_such_column);
      if ((description_col=findcol(hdr,"DESCRIPTION")) < 0) ccerr(no_such_column);
      redir_col = findcol(hdr,"REDIRECTION");
      done = 1;                              /* read header info */
      continue;
    }
    strncpy(scratch,colptr(export_name_col,str),EQP_DEVICE_SIZE);

    for (el=ExportList; el != NULL; el=el->next) if (!strncmp(scratch,el->tagName,EQP_DEVICE_SIZE)) break;
    if (el == NULL)               /* new eqp type */
    {
      el = (ExportListStruct *)SystemCalloc(1,sizeof(ExportListStruct));
      if (el == NULL) ccerr(out_of_server_memory);
      strncpy(el->tagName,scratch,ELEM_NAME_SIZE);
      strncpy(el->EqpName,colptr(local_name_col,str),EQP_NAME_SIZE);
      strncpy(el->Context,gDeviceContext,EQP_CONTEXT_SIZE);
      if (!search_function(el->EqpName,&header,&ftype)) /* compile-time registration */
      {
        el->NdlFcn = (EQPCALL)header->F_entry;
      }
      strncpy(scratch,colptr(num_modules_col,str),16);
      el->EqpNumModules = atoi(scratch);
      if (el->EqpNumModules == 0) el->EqpNumModules = 1;
      el->EqpModuleName = GetEqpModuleNames(el->EqpName,el->EqpNumModules);
      if (el->EqpNumModules >= MODULE_HASH_PAYOFF)
      {
        if ((el->EqpModuleXRef == NULL) &&
            (el->EqpModuleXRef=(HashXRefStruct **)SystemCalloc(MODULE_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
          ccerr(out_of_server_memory);
        makeEqpModuleHashTable(el);
      }
      el->EqpNumProperties = 0;
      if ((el->almLst=(ALARM **)SystemCalloc(el->EqpNumModules,sizeof(ALARM *))) == NULL)
        ccerr(out_of_server_memory);
      el->exportId = NrOfEqpTypes++;
      el->next = ExportList;
      ExportList = el;
    }
    if ((prp=(ExportPropertyListStruct *)SystemCalloc(1,sizeof(ExportPropertyListStruct))) == NULL) ccerr(out_of_server_memory);
    strncpy(prp->prpName,colptr(property_col,str),EQP_PROP_SIZE);
    if (strlen(prp->prpName) > 16 && longDeviceNames) ccerr(string_too_long);
    if ((i=getAliasTableIndexFromName(prp->prpName)) >= 0)
      strncpy(prp->prpAlias,gAliasTable[i].alias,EQP_PROP_SIZE);
    strncpy(scratch,colptr(property_size_col,str),16);
    prp->prpSize = atoi(scratch);
    strncpy(scratch,colptr(format_col,str),64);
    if (!strncmp(scratch,"STRUCT.",7)) /* tag coming ... */
    {
      if ((c=strstr(scratch,".")) == NULL) ccerr(code_failure);
      *c++ = 0;
      strncpy(prp->prpFormatTag.name,c,8);
    }
    prp->prpFormat = ftoi(scratch);
    prp->prpArrayType = getArrayType(scratch);
    if (!(prp->prpArrayType & AT_DOUBLE)) prp->prpArrayType |= prp->prpSize > 1 ? AT_SINGLE : AT_SCALAR;
    prp->rowSize = getArrayRowLength(scratch);
    if (prp->rowSize > 0) prp->numRows = (UINT16)(prp->prpSize/prp->rowSize); else prp->rowSize = (short)prp->prpSize;
    strncpy(scratch,colptr(access_col,str),16);
    prp->prpAccessMode = mtoi(scratch);
    if (property_insize_col >= 0)
    {
      strncpy(scratch,colptr(property_insize_col,str),16);
      prp->prpSizeIn = atoi(scratch);
    }
    else
    {
      if (prp->prpAccessMode & CA_WRITE) prp->prpSizeIn = prp->prpSize;
    }
    prp->prpFormatIn = CF_NULL;
    if (informat_col >= 0)
    {
      strncpy(scratch,colptr(informat_col,str),16);
      if (!strncmp(scratch,"STRUCT.",7)) /* tag coming ... */
      {
        if ((c=strstr(scratch,".")) == NULL) ccerr(code_failure);
        *c++ = 0;
        strncpy(prp->prpFormatTagIn.name,c,8);
      }
      prp->prpFormatIn = ftoi(scratch);
    }
    else
    {
      if (prp->prpAccessMode & CA_WRITE) prp->prpFormatIn = prp->prpFormat;
    }
    ParsePropertyDescription(colptr(description_col,str),prp);
    if (redir_col >= 0)
    {
      strncpy(scratch,colptr(redir_col,str),64);
      if (strlen(scratch) > 0)
      {
        if ((prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) ccerr(out_of_server_memory);
        prd->next = prp->prd;
        prp->prd = prd;
        ParseRedirector(scratch,prp);
      }
    }
    if (strchr(prp->prpName,'*')) idx = PROPERTY_HASH_SIZE;
    else idx = (int)(ElfHash((unsigned char *)prp->prpName) % PROPERTY_HASH_SIZE);
    if (property_id_col >= 0)
    {
      strncpy(scratch,colptr(property_id_col,str),16);
      prpId = atoi(scratch);
    }
    else prpId = el->EqpNumProperties;
    prp->prpId = MAKE_PROPERTY_ID(prpId,idx);
    prp->next = el->EqpProperty[idx];
    el->EqpProperty[idx] = prp;
    if ((i=strlen(prp->prpName)) > 4 && !strcmp(&prp->prpName[i-4],".NAM")) 
    {
      el->hasPropertyDeviceNames = TRUE;
      if (queryDeviceNamesFromProperties) acquireDeviceNamesFromProperties = TRUE;
    }
    el->EqpNumProperties++;
  }

  getAlarmInfo();

err:
  if (fp) fclose(fp);
  feclog("exports file %s : %s\n",expfilename,cc ? erlst[cc] : "found");
  return cc;
}
ExportListStruct *getExportListItem(char *eqm)
{
  ExportListStruct *el;

  if (eqm == NULL) return NULL;
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqpName,EQP_NAME_SIZE)) return el;
  return NULL;
}
FILE *devFileOpen(char *eqm)
{
  char fname[32],devFile[128];
  int i;
  FILE *fp=NULL;
  for (i=0; i<2; i++)
  {
    memset(fname,0,sizeof(fname));
    strncpy(fname,eqm,6);
    strlwr(fname);
#   if !defined(MSDOS)
    strcat(fname,i == 0 ? "-devices.csv" : "dv.csv");
#   else
    strcat(fname,i == 0 ? "-d.csv" : "dv.csv");
#   endif
    sprintf(devFile,"%s%s",FecDBpath,fname);
    if ((fp=fopen(devFile,"r")) != NULL) return fp;
  }
  return NULL;
}
NAME32 *GetEqpModuleNames(char *eqname,int num)
{
  NAME32 *eqpModuleName = NULL;
  int i,cc=0,done=0,name_col=-1,num_col=-1;
  char str[256],hdr[256],scratch[36];
  FILE *fp = NULL;

  if (num <= 0) ccerr(out_of_range);
  if ((eqpModuleName=(NAME32 *)SystemCalloc(num,sizeof(NAME32))) == NULL) ccerr(out_of_server_memory);
  if ((fp=devFileOpen(eqname)) == NULL) ccerr(no_such_file);

  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((name_col=findcol(hdr,"MODULE_NAME")) < 0 &&
          (name_col=findcol(hdr,"DEVICE_NAME")) < 0 ) ccerr(no_such_column);
      if ((num_col=findcol(hdr,"MODULE_NUMBER")) < 0 &&
          (num_col=findcol(hdr,"DEVICE_NUMBER")) < 0 ) ccerr(no_such_column);
      done = 1;                              /* read header info */
      continue;
    }
    memset(scratch,0,33);
    strncpy(scratch,colptr(num_col,str),32);
    if ((i=atoi(scratch)) < num)
      strncpy(eqpModuleName[i].name,colptr(name_col,str),32);
  }

err:
  if (fp) fclose(fp);
  if (cc && NGdebug) printf("GetEqpModuleNames: %s - %s\n>",eqname,erlst[cc]);
  feclog("device file for %s : %s",eqname, cc ? erlst[cc] : "found");
  return eqpModuleName;
}
char tRedirectedServer[64];
int lookupPropertyName(ExportListStruct *el,char *prpName,char *devName)
{
  ExportPropertyListStruct *prp;
  PrpRedirBlk *prd = NULL;
  int idx, redir = 0;

  if (el == NULL || prpName == NULL) return -1;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);
  for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
    if (!strncmp(prp->prpName,prpName,32)) 
    {
      if (prp->prd != NULL)
      {
        for (prd=prp->prd; prd != NULL; prd=prd->next)
        {
          if (prd->prpRedirection[2].name[0] == 0) redir = -1;
          if (prd->prpRedirection[2].name[0] != 0 &&
              !strncmp(prd->prpRedirection[2].name,devName,16)) redir = -1;
          if (redir)
          {
            memcpy(tRedirectedServer,prd->prpRedirection[0].name,16*4);
            return -server_redirection;
          }
        }
      }
      return (int)PROPERTY_HASH_INDEX(prp->prpId);
    }
  return -1;
}
char *lookupPropertyId(ExportListStruct *el,short id)
{
  ExportPropertyListStruct *prp;
  int idx;

  if (el == NULL) return (char *)NULL;
  idx = PROPERTY_HASH_INDEX(id);
  for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
    if (prp->prpId == id) return prp->prpName;
  return (char *)NULL;
}
int RegisterLocalAlias(char *eqpname,char *alias)
{
  int cc=0;
  ExportListStruct *el;

  if ((el=getExportListItem(eqpname)) == NULL) ccerr(non_existent_elem);
  strncpy(el->EqpAlias,alias,EQP_NAME_SIZE);
  feclog("Register local alias %s for %s",alias,eqpname);
err:
  return cc;
}
int sndRegisteredExports(struct timeval *tv)
{
  int cc=0;
  static time_t lastAttempt = 0;
  static int toCount = 0, errCount = 0;
  RPCDataStruct rpcdata;
  RPCFecStruct  rpcfec;
  static BYTE fecAddressBuf[sizeof(RPCDataStruct)+sizeof(RPCFecStruct)+sizeof(RPCFecExStruct)];
  ExportListStruct *el;

  if (!FecNameRegistered)
  {
    /* Fec Name not available, so don't try again */
    ExportsRegistered = 1;
    return database_not_loaded;
  }
  if ((cc=GetNameServerAddress()) != 0)
  {
    /* Name Server Address not available, so don't try again */
    ExportsRegistered = 1;
    return cc;
  }
  if (toCount > 3)
  {
    /* Name Server not available, so don't try again */
    ExportsRegistered = 1;
    return link_not_open;
  }
  if (tv->tv_sec < lastAttempt + 10) return 0;
  lastAttempt = tv->tv_sec;
  memset(&rpcfec,0,sizeof(RPCFecStruct));
  strncpy(rpcfec.Name,gFecName,FEC_NAME_SIZE);
  memcpy(rpcfec.Network,&IPXFecAddress,12);
  rpcfec.Port[0] = PortOffset&0xff;
  rpcfec.Port[1] = (PortOffset>>8)&0xff;
  rpcfec.PortOffset = (BYTE)PortOffset; /* legacy setting only 1 byte long ! */
  PrepOutgoingFecaddr(&rpcfec,fecAddressBuf);
  memcpy(&fecAddressBuf[RPCDATASTRUCT_SIZE],fecAddressBuf,RPCFECSTRUCT_SIZE);

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (!strcmp(el->tagName,"LOCAL")) el->registered = TRUE;
    if (el->registered) continue;
    memset(&rpcdata,0,sizeof(RPCDataStruct));
    strncpy(rpcdata.FecName,gFecName,FEC_NAME_SIZE);
    strncpy(rpcdata.ExportName,el->tagName,ELEM_NAME_SIZE);
    strncpy(rpcdata.EqName,el->EqpName,EQP_NAME_SIZE);
    rpcdata.alarmSystem = el->alarmSystem;
    strncpy(el->Context,gDeviceContext,EQP_CONTEXT_SIZE);
    strncpy(rpcdata.Context,gDeviceContext,EQP_CONTEXT_SIZE);
    PrepOutgoingEqpaddr(&rpcdata,fecAddressBuf);
    memcpy(&fecAddressBuf[RPCDATASTRUCT_SIZE+RPCFECSTRUCT_SIZE],&RPCFecEx,sizeof(RPCFecExStruct));
    switch (cc=tExecLink(0,(char *)fecAddressBuf,CSHOSTS,gDeviceContext,"EXPORT",-(short)(RPCFECSTRUCT_SIZE+RPCDATASTRUCT_SIZE+sizeof(RPCFecExStruct)),CF_STRUCT,100))
    {
      case 0:
        feclog("FEC %s, Server %s (%s) registered with equipment name server",
          gFecName,el->tagName,el->EqpName);
        feclog("Context is %s",strlen(gDeviceContext) > 0 ? gDeviceContext : "not given");
        errCount = 0;
        break;
      case server_name_in_use:
        feclog("FEC name %s or Server name %s is in use",gFecName,el->tagName);
        break;
      case operation_busy: ccerr(cc);
      case link_not_open: toCount++;
      default: errCount++;
        if (errCount < 30) ccerr(cc);
        feclog("register FEC %s, Server %s : %s",gFecName,el->tagName,erlst[cc]);
        break;
    }
    el->registered = 1;
  }
  ExportsRegistered = 1;
err:
  gettimeofday(tv,(struct timezone *)NULL);
  return cc;
}
int outputConsumerList(void)
{
  int k;
  char Buffer[256];

  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," CLIENT"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[16],"ADDRESS"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[32],"PROTOCOL"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[48],"Contracts\n");
  ttyoutput(Buffer);
  for (k=0; k<nconsumer; k++)
  {
    memset(Buffer,' ',80);
    sprintf(&Buffer[0]," %s",Consumers[k]->userName); Buffer[strlen(Buffer)] = ' ';
    ntoa(Consumers[k],&Buffer[16]);
    Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[32],"%d",Consumers[k]->protocol);
    sprintf(&Buffer[32],"%s",getProtocolAsString(Consumers[k]->transport));
    Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[48],"%d\n",Consumers[k]->count);
    ttyoutput(Buffer);
  }
  return 0;
}
int outputContractList(void)
{
  int i;
  char Buffer[256];
  CONSUMER *c;
  CONTRACT *con;

  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," CONTRACT"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[50],"POLL"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[60],"TO\n");
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  for (i=0; i<ncontract; i++)
  {
    con = &ContractList[i]->contract;
    c   = ContractList[i]->consumer;
    sprintf(&Buffer[0],"[%d] %s %s <%s> (%lu elements)",
      i,con->EqpName,con->EqpProperty,con->EqpDeviceName,con->EqpSizeOut);
    Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[50],"%ld msec",ContractList[i]->PollingRate);
    Buffer[strlen(Buffer)] = ' ';
    if (ContractList[i]->expired)
    {
      sprintf(&Buffer[60]," [expired]\n");
      ttyoutput(Buffer);
      memset(Buffer,' ',80);
    }
    for (c=ContractList[i]->consumer; c != NULL; c=c->next)
    {
      sprintf(&Buffer[60],"%s\n",c->client->userName);
      ttyoutput(Buffer);
      memset(Buffer,' ',80);
    }
  }
  return 0;
}
int outputContractStats(int i)
{
  char Buffer[256];
  CONSUMER *c;
  CONTRACT *con;

  if (i < 0 || i >= ncontract) return out_of_range;

  con = &ContractList[i]->contract;

  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," CONTRACT\n");
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," %s %s <%s> (%lu elements) status <%d> returned <%d>\n\n",
    con->EqpName,con->EqpProperty,con->EqpDeviceName,con->EqpSizeOut,
    ContractList[i]->compStatus,ContractList[i]->returnCode);
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  sprintf(&Buffer[0],"POLL"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[12],"LAST SENT"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[40],"LAST STALE"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[63],"MISSES DELAYS\n"); 
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  sprintf(&Buffer[0],"%ld msec",ContractList[i]->PollingRate);
  Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[12],"%ld sec %ld usec",
   ContractList[i]->lasttime.tv_sec,ContractList[i]->lasttime.tv_usec);
  Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[40],"%ld sec %ld usec",
   ContractList[i]->lastmarked.tv_sec,ContractList[i]->lastmarked.tv_usec);
  Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[63],"%6ld %6ld\n\n",
   ContractList[i]->misses,ContractList[i]->delays);
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  sprintf(&Buffer[0],"CLIENTS\n");
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  for (c=ContractList[i]->consumer; c != NULL; c=c->next)
  {
    sprintf(&Buffer[0],c->client->userName); Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[9]," [%s] %ld ms +%d : %s %ld s %ld us %lu miss %lu rec\n",
            SubType[c->status.mode],
            c->status.PollingRate,
            c->status.counter,
            c->status.Stale ? "STALE" : "FRESH",
            c->status.lasttime.tv_sec,
            c->status.lasttime.tv_usec,
            c->status.misses,
            c->status.reconnects);
    ttyoutput(Buffer);
    memset(Buffer,' ',80);
  }
  return 0;
}
long CMDcount = 0;
int outputServerStats(void)
{
  char strbuf[256];

  sprintf(strbuf,"Running since %s",ctime(&startuptime)); ttyoutput(strbuf);
  sprintf(strbuf,"Total SPX requests  : %lu\n",SPXcount); ttyoutput(strbuf);
  sprintf(strbuf,"Total IPX requests  : %lu\n",IPXcount); ttyoutput(strbuf);
  sprintf(strbuf,"Total TCP requests  : %lu\n",TCPcount); ttyoutput(strbuf);
  sprintf(strbuf,"Total UDP requests  : %lu\n",UDPcount); ttyoutput(strbuf);
  sprintf(strbuf,"Total KBD commands  : %lu\n",CMDcount); ttyoutput(strbuf);
  sprintf(strbuf,"Socket RCV Buffers  : %d\n",max_rcvbuf); ttyoutput(strbuf);
  sprintf(strbuf,"Socket SND Buffers  : %d\n",max_sndbuf); ttyoutput(strbuf);
  if (NrOfEqpTypes) /* non-servers must cover these variables */
  {
    sprintf(strbuf,"Server Work Area    : %lu bytes\n",srvWorkAreaSize); ttyoutput(strbuf);
    sprintf(strbuf,"Registered clients  : %d\n",nconsumer); ttyoutput(strbuf);
    sprintf(strbuf,"Registered contracts: %d\n",ncontract); ttyoutput(strbuf);
    sprintf(strbuf,"Contract misses     : %lu\n",gContractMisses); ttyoutput(strbuf);
    sprintf(strbuf,"Contract delays     : %lu\n",gContractDelays); ttyoutput(strbuf);
    sprintf(strbuf,"Client misses       : %lu\n",gClientMisses); ttyoutput(strbuf);
    sprintf(strbuf,"Client reconnects   : %lu\n",gClientReconnects); ttyoutput(strbuf);
    sprintf(strbuf,"Client retries      : %lu\n",gClientRetries); ttyoutput(strbuf);
    sprintf(strbuf,"Synchronous calls   : %lu\n",gSingleLinkCount); ttyoutput(strbuf);
    sprintf(strbuf,"Bursts              : %lu\n",gBurstLimitReachedCount); ttyoutput(strbuf);
  }
  if (nConnectionTableEntries)                 /* non-clients must cover these variables */
  {
    sprintf(strbuf,"Connection arrivals : %lu\n",nArrivals); ttyoutput(strbuf);
    sprintf(strbuf,"Connection timeouts : %lu\n",nConnectionTimeouts); ttyoutput(strbuf);
    sprintf(strbuf,"Incomplete transfers: %lu\n",nPartialTransfers); ttyoutput(strbuf);
    sprintf(strbuf,"Client Work Area    : %lu bytes\n",clnWorkAreaSize); ttyoutput(strbuf);
  }
  sprintf(strbuf,"System Polling rate : %ld msec\n",SysPoll); ttyoutput(strbuf);
  sprintf(strbuf,"CPU usage           : %d percent\n", AveBusyTime); ttyoutput(strbuf);
  sprintf(strbuf,"Average Cycles/sec  : %d Hz\n", gCycleCounts); ttyoutput(strbuf);
  sprintf(strbuf,"Max Cycles/sec      : %d Hz\n", gMaxCycleCounts); ttyoutput(strbuf);
  return 0;
}
int outputServerTime(void)
{
  char sstr[128],tstr[64];
  time_t t;

  t = time(NULL); sprintf(tstr,"%s",ctime(&t)); tstr[24] = 0;
  sprintf(sstr,"current FEC time: %s (UNIX: %ld)\n>",tstr,t);
  ttyoutput(sstr);
  gDataTimeStamp = makeDataTimeStamp();
  t = RPCtimestamp; sprintf(tstr,"%s",ctime(&t)); tstr[24] = 0;
  sprintf(&tstr[strlen(tstr)]," %d msec",RPCtimestampMSEC);
  sprintf(sstr,"current Data timestamp: %s (UNIX: %10.4f)\n>",tstr,gDataTimeStamp);
  ttyoutput(sstr);
  sprintf(sstr,"current synchronization offset: %10.4f secs\n>",gDataTimeStampOffset);
  ttyoutput(sstr);
  return 0;
}
int dumpVersion(void)
{
  BYTE *ver = SystemVersion();
  char strbuf[256];
  sprintf(strbuf,"TINE version %d.%02d.%04d\n>",ver[0],ver[1],ver[2]*256 + ver[3]);
  ttyoutput(strbuf);
  sprintf(strbuf,"Appl version %s compiled on ",appmkversion);
  strcat(strbuf,ctime(&appmktime)); strcat(strbuf,">");
  ttyoutput(strbuf);
  return 0;
}
int dumpConsumerList(void) { return outputConsumerList(); }
int dumpContractList(void) { return outputContractList(); }
int dumpContractStats(int i) { return outputContractStats(i); }
int dumpServerStats(void) { return outputServerStats(); }
int dumpServerTime(void) { return outputServerTime(); }

ALIAS_TABLE *gAliasTable = NULL;
int nAliasTableEntries = 0;
HashXRefStruct **AliasXRefTable = NULL;

CSVCOLUMN csvAliasFileColumns[] = 
{
  { "NAME","",0,CF_TEXT,32,(BYTE *)OFFSETIN(ALIAS_TABLE,name),NULL,0 },
  { "ALIAS","",0,CF_TEXT,32,(BYTE *)OFFSETIN(ALIAS_TABLE,alias),NULL,0 }
};
CSVDB csvAliasFileDb =
{
  csvAliasFileColumns,
  sizeof(csvAliasFileColumns)/sizeof(csvAliasFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(ALIAS_TABLE)
};
int GetAliasTable(void)
{
  int cc;
  if ((cc=csvReadFile(FecDBpath,ALIAS_FILE,&csvAliasFileDb,(void **)&gAliasTable)) != 0) ccerr(cc);
  nAliasTableEntries = csvAliasFileDb.siz;
  if (nAliasTableEntries > ALIAS_HASH_PAYOFF && (cc=makeAliasXRefTable()) != 0) ccerr(cc);
err:
  feclog("alias file : %s", cc ? erlst[cc] : "found");
  return cc;
}
int makeAliasXRefTable(void)
{
  int k,idx;
  HashXRefStruct *xref;

  if ((AliasXRefTable=(HashXRefStruct **)SystemCalloc(ALIAS_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
    return out_of_local_memory;
  for (k=0; k<nAliasTableEntries; k++)
  {
    idx = (int)(ElfHash((unsigned char *)gAliasTable[k].alias) % ALIAS_HASH_SIZE);
    for (xref=AliasXRefTable[idx]; xref && xref->index != k; xref = xref->next)
    {
      if (!strncmp(gAliasTable[k].alias,gAliasTable[xref->index].alias,32)) break;
    }
    if (xref == NULL) 
    {
      if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) return out_of_server_memory;
      xref->index = k;
      xref->next = AliasXRefTable[idx];
      AliasXRefTable[idx] = xref;
    }
  }
  return 0;
}
int getTableIndex(char *str,HashXRefStruct **xrf,int hshsiz,char *tbl,int tblsiz,int off,int reclen)
{
  int i,idx;
  HashXRefStruct *xref;

  if (tbl == NULL) return -1;
  if (xrf != NULL && hshsiz > 0) /* this should be quick */
  {
    idx = (int)(ElfHash((unsigned char *)str) % hshsiz);
    for (xref=xrf[idx]; xref; xref = xref->next)
    {
      i = xref->index;
      if (!strncmp(str,tbl+reclen*i+off,32)) return i;
    }
  }
  else
  {
    for (i=0; i<tblsiz; i++)
    {
      if (!strncmp(str,tbl+reclen*i+off,32)) return i;
    }
  }
  return -1;
}
#ifndef MSDOS
void freeServerMemory(void)
{
  ExportListStruct *el, *el0;
  HashXRefStruct *xref, *xref0;
  ExportPropertyListStruct *prp, *prp0;
  ALARM *alm, *alm0;
  ALARM_WATCH *aw, *aw0;
  PrpRedirBlk *prd, *prd0;
  struct structStruct *s, *s0;
  GroupTableEntry *gt, *gt0;
  int i;

  if (cdiSystemFreeHook != NULL) cdiSystemFreeHook();

  if (TCPsck != NULL) SystemFree(TCPsck); TCPsck = NULL;
  
  if (IPNetList != NULL) SystemFree(IPNetList); IPNetList = NULL;
  if (IPBcastList != NULL) SystemFree(IPBcastList); IPBcastList = NULL;

  if (vFeclogBuffer != NULL) SystemFree(vFeclogBuffer);
  vFeclogBuffer = NULL;

  if (gAliasTable != NULL) 
  {
    for (i=0; i<ALIAS_HASH_SIZE; i++)
    {
      for (xref=AliasXRefTable[i]; xref != NULL; )
      {
        xref0 = xref->next;
        SystemFree(xref);
        xref = xref0;
      }
    }
    SystemFree(gAliasTable);
    gAliasTable = NULL;
  }
  if (StockPropXRefTable != NULL)
  {
    for (i=0; i<STOCKPROP_HASH_SIZE; i++)
    {
      for (xref=StockPropXRefTable[i]; xref != NULL; )
      {
        xref0 = xref->next;
        SystemFree(xref);
        xref = xref0;
      }
    }
    SystemFree(StockPropXRefTable);
  }
  if (historyRecords != NULL)
  {
    for (i=0; i<maxHistoryRecords && i<nHistoryRecords; i++)
    {
      if (historyRecords[i].timestamp != NULL) SystemFree(historyRecords[i].timestamp);
      if (historyRecords[i].data != NULL) SystemFree(historyRecords[i].data);
    }
    SystemFree(historyRecords); historyRecords = NULL;
  }
  for (aw=almWatchList; aw != NULL; )
  {
    aw0 = aw->nxt;
    SystemFree(aw);
    aw = aw0;
  }
  for (el=ExportList; el != NULL; )
  {
    for (i=0; i<PROPERTY_HASH_SIZE; i++)
    {
      for (prp=el->EqpProperty[i]; prp != NULL; )
      {
        prp0 = prp->next;
        if (prp->prd != NULL) 
        {
          for (prd=prp->prd; prd != NULL; )
          {
            prd0 = prd->next;
            SystemFree(prd);
            prd = prd0;
          }
        }
        SystemFree(prp);
        prp = prp0;
      }
    }
    if (el->EqpModuleXRef != NULL)
    {
      for (i=0; i<MODULE_HASH_SIZE; i++)
      {
        for (xref=el->EqpModuleXRef[i]; xref != NULL; )
        {
          xref0 = xref->next;
          SystemFree(xref);
          xref = xref0;
        }
      }
    }
    if (el->adsTable != NULL) SystemFree(el->adsTable);
    for (i=0; i<el->EqpNumModules; i++) if (el->almLst[i] != NULL)
    {
      for (alm=el->almLst[i]; alm != NULL; )
      {
        alm0 = alm->next;
        SystemFree(alm);
        alm = alm0;
      }
      SystemFree(el->almLst[i]);
    }
    SystemFree(el->almLst);

    el0 = el->next;
    SystemFree(el);
    el = el0;
  }
  ExportList = NULL;
  for (i=0; i<nbroadcasts; i++)
  {
    if (BrdcstList[i]->refdata != NULL) SystemFree(BrdcstList[i]->refdata);
    if (BrdcstList[i] != NULL) SystemFree(BrdcstList[i]);
  }

  if (ContractList != NULL)
  {
    for (i=0; i<MaxNumContracts; i++)
    {
      if (ContractList[i] == NULL) continue;
      if (ContractList[i]->DataIn != NULL) SystemFree(ContractList[i]->DataIn);
      if (ContractList[i]->DataOut != NULL) SystemFree(ContractList[i]->DataOut);
      SystemFree(ContractList[i]);
    }
    SystemFree(ContractList); ContractList = NULL;
  }
  if (Consumers != NULL)
  {
    for (i=0; i<MaxNumClients; i++)
    {
      if (Consumers[i] == NULL) continue;
      SystemFree(Consumers[i]);
    }
    SystemFree(Consumers); Consumers = NULL;
  }

  if (connTable != NULL)
  {
    for (gt=grpTable; gt != NULL; )
    {
      gt0 = gt->next;
      SystemFree(gt);
      gt = gt0;
    }
    for (i=0; i<MaxNumConnections; i++)
    {
      if (connTable[i] != NULL)
      {
        if (connTable[i]->dataIn != NULL) SystemFree(connTable[i]->dataIn);
        if (connTable[i]->dataOut != NULL) SystemFree(connTable[i]->dataOut);
        if (connTable[i]->bufOut != NULL) SystemFree(connTable[i]->bufOut);
        if (connTable[i]->bufOutBlks != NULL) SystemFree(connTable[i]->bufOutBlks);
        SystemFree(connTable[i]);
      }
    }
    SystemFree(connTable); connTable = NULL;
  }

  if (NGlobalList!= NULL)
  {
    for (i=0; i<nglobals; i++)
    {
      if (NGlobalList[i] == NULL) continue;
      if (NGlobalList[i]->data != NULL) SystemFree(NGlobalList[i]->data);
      if (NGlobalList[i]->refdata != NULL) SystemFree(NGlobalList[i]->refdata);
      SystemFree(NGlobalList[i]);
    }
    SystemFree(NGlobalList); NGlobalList = NULL;
    nglobals = 0;
  }

  if (RPCData != NULL) SystemFree(RPCData); RPCData = NULL;
  if (RPCFec != NULL) SystemFree(RPCFec); RPCFec = NULL;
  if (NameServer != NULL) SystemFree(NameServer); NameServer = NULL;

# ifndef INTEL
  if (tmpWorkArea != NULL && tmpWorkArea != srvWorkArea) SystemFree(tmpWorkArea); tmpWorkArea = NULL;
# endif
# ifndef srvWorkArea
  SystemFree(srvWorkArea); srvWorkArea = NULL;
# else
  SystemFree(WorkArea); WorkArea = NULL;
# endif
# ifndef clnWorkArea
  if (clnWorkArea != NULL) SystemFree(clnWorkArea); clnWorkArea = NULL;
# endif
# ifndef MSDOS
  freeStruct(HeaderStructStruct);
  freeStruct(EqpInfoStructStruct);
  freeStruct(CONTRACT4Struct);
  freeStruct(CONTRACTStruct);
  freeStruct(ProducerHeaderStruct4Struct);
  freeStruct(ProducerHeaderStructStruct);
  freeStruct(SUBSCRIPTION4Struct);
  freeStruct(SUBSCRIPTIONStruct);
  freeStruct(RPCDataStructStruct);
  freeStruct(RPCFecStructStruct);
  freeStruct(AMSStructStruct);
  freeStruct(ADSStructStruct);
  freeStruct(PropertyQueryStructStruct);
  freeStruct(PropertyQueryStructExStruct);
  freeStruct(ActivityQueryStructStruct);
  freeStruct(ServerStateStructStruct);
  freeStruct(BrdcstDataStruct);
  freeStruct(ClientQueryStructStruct);
  freeStruct(SpectrumHeaderStruct);

  for (s=structList; s != NULL; )
  {
    s0 = s->next;
    if (s->sdata != NULL) SystemFree(s->sdata);
    SystemFree(s);
    s = s0;
  }
  structList = NULL;
# endif
  return;
}
int GetPrpId(char *eqm,char *prpName)
{
  ExportListStruct *el = getExportListItem(eqm);
  ExportPropertyListStruct *prp = NULL;
  int idx;

  if (el == NULL || prpName == NULL) return -1;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);

  for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
  {
    if (!strncmp(prp->prpName,prpName,EQP_PROP_SIZE)) return prp->prpId;
  }
  return -1;
}
ExportPropertyListStruct *GetDeviceListFromProperty(char *eqm,char *prp)
{
  int cc = 0,pid;
  ExportListStruct *el;
  ExportPropertyListStruct *pl;
  char namProperty[64];
  DTYPE dout;
  static int inside = 0;

# ifdef MULTITHREADED_XXXX
  if (gEqpFcnBusySemaphore) return NULL;
# endif
  if ((el=getExportListItem(eqm)) == NULL) return NULL;
  if (el->EqpFcn == NULL) return NULL;
  if ((pid=GetPrpId(eqm,prp)) < 0) return NULL;
  if (strlen(prp) > 28) return NULL;
  pid = PROPERTY_HASH_INDEX(pid);
  for (pl=el->EqpProperty[pid]; pl != NULL && strncmp(pl->prpName,prp,32); pl=pl->next);
  if (pl == NULL) return NULL;
  if (pl->devNames == NULL)
  {
    if (inside) return NULL;
    inside = TRUE;
    sprintf(namProperty,"%s.NAM",prp);
    dout.dArrayLength = pl->prpSize;
    dout.dFormat = CF_NAME16;
    dout.data.vptr = tmpWorkArea;
    gEqpFcnBusySemaphore = 1;
    cc = (*el->EqpFcn)("#0",namProperty,&dout,NULL,CA_READ);
    gEqpFcnBusySemaphore = 0;
    if (cc != 0) ccerr(cc);
    if ((pl->devNames=(NAME16 *)SystemCalloc(dout.dArrayLength,16)) == NULL) ccerr(out_of_server_memory);
    memcpy(pl->devNames,dout.data.vptr,dout.dArrayLength*16);
    pl->numDevices = dout.dArrayLength;
  }
err:
  inside = FALSE;
  return cc ? NULL : pl;
}
#else
ExportPropertyListStruct *GetDeviceListFromProperty(char *eqm,char *prp)
{
  eqm = eqm; prp = prp;
  return NULL;
}
#endif /* MSDOS */
int (*lookupRedirectionNameStub)(char *eqm,char *prpName,char *devName) = NULL;
#endif /* NO_DOXYGEN */

/* Documented API calls follow: */
int AssignNumModules(char *eqname,int num)
{
  ExportListStruct *el;
  int cc=0;

  for (el=ExportList; el != NULL; el=el->next)
    if (!strncmp(eqname,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);

  if (num < 0 || num >= el->EqpNumModules) ccerr(out_of_range);
  el->EqpNumModules = num;      /* this is a one-way street  */
err:
  return cc;
}
int ExecLocalLink(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  ContractListStruct cl;
  CLN cln;
  BYTE *datain = NULL,*dataout = NULL;
  char fulname[72];
  char *en = NULL,*dn = NULL;
  int cc = 0;

  strncpy(fulname,devName,71); fulname[71] = 0;
  if ((en=strtok(fulname,DELIMITER)) == NULL) ccerr(argument_list_error);
  if ((dn=strtok(NULL,DELIMITER)) == NULL) dn = &fulname[71];

  strncpy(cln.userName,gFecName,16);

  memset(&cl.contract,0,sizeof(CONTRACT));
  strncpy(cl.contract.EqpName,en,EQP_NAME_SIZE);
  strncpy(cl.contract.EqpDeviceName,dn,EQP_DEVICE_SIZE);
  strncpy(cl.contract.EqpProperty,devProperty,EQP_PROP_SIZE);

  if (strlen(dn) > EQP_DEVICE_SIZE && strlen(devProperty) <= EQP_DEVICE_SIZE)
  {
    cl.contract.EqpDeviceName[15] = '&';
    cl.contract.EqpProperty[16] = 0;
    strncpy(&cl.contract.EqpProperty[17],&dn[15],15);
  }

  if (din != NULL)
  {
    cl.contract.EqpSizeIn = din->dArrayLength;
    cl.contract.EqpFormatIn = BFMT(din->dFormat);
    cl.contract.strTagIn[0] = 0;
    datain = din->data.bptr;
  }
  if (dout != NULL)
  {
    cl.contract.EqpSizeOut = dout->dArrayLength;
    cl.contract.EqpFormatOut = BFMT(dout->dFormat);
    cl.contract.strTagOut[0] = 0;
    dataout = dout->data.bptr;
  }
  cl.contract.EqpAccess = (BYTE)access;

  if ((cc=IsIllegalProperty(cl.contract.EqpName,
        cl.contract.EqpProperty,cl.contract.EqpDeviceName,
        &cln,access,&cl.exp)) != 0) ccerr(cc);

  setDataTimeStamp(makeDataTimeStamp());
  cl.DataOut = dataout;
  cl.DataIn = datain;
  cl.DataBuf = srvWorkArea;
  if ((cc=doContract(&cl)) != 0) goto err;
err:
  return cc;
}
char *GetCaller(void)
{
  static char name[48];

  if (EQPconsumerlist == NULL || 
      EQPconsumerlist->client == NULL || 
      EQPconsumerlist->client->userName == NULL) return "LOCAL";
  strncpy(name,EQPconsumerlist->client->userName,16);
  return name;
}
int GetCallerInfo(NAME16 *un,BYTE *ipx,UINT32 *ip,short *prot,int *num)
{
  CONSUMER *cl;
  char scratch[48];
  int i;

  if (num == NULL) return buffer_too_small;
  for (i=0,cl=EQPconsumerlist; cl; cl=cl->next,i++)
  {
    if (i < *num)
    {
      if (un != NULL) strncpy(un[i].name,cl->client->userName,16);
      if (ipx != NULL) memcpy(&ipx[i*6],cl->client->IPXaddress.network,6);
      if (ip != NULL)
      {
        inet_ntoa_b(cl->client->IPaddress.sin_addr,scratch);
        ip[i] = inet_addr(scratch);
      }
      if (prot != NULL) prot[i] = cl->client->protocol;
    }
  }
  *num = i;
  return 0;
}
char *GetExportedName(char *eqpname)
{
  ExportListStruct *el;
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqpname,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) return NULL;
  return el->tagName;
}
char *GetDeviceName(char *eqm,int devnr)
{
  ExportListStruct *el = getExportListItem(eqm);
  static char lclname[36];

  sprintf(lclname,"DEVICE %d",devnr);
  if (el == NULL || devnr < 0 || devnr >= el->EqpNumModules) return lclname;
  return el->EqpModuleName[devnr].name;
}
NAME32 *GetExportedDeviceList(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el != NULL) return el->EqpModuleName;
  return NULL;
}
int GetDeviceNumberEx(char *eqm,char *devname,char *prpname)
{
  char lclname[36];
  int idx,i;
  HashXRefStruct *xref;
  ExportListStruct *el = getExportListItem(eqm);
  ExportPropertyListStruct *prp;

  if (eqm == NULL || devname == NULL) return -argument_list_error;
  if (prpname != NULL && (prp=GetDeviceListFromProperty(eqm,prpname)) != NULL)
  {
    for (idx=0; idx<prp->numDevices; idx++) if (!strncmp(devname,prp->devNames[idx].name,EQP_DEVICE_SIZE)) return idx;
  }
  if (el == NULL) return -non_existent_elem;
  memset(lclname,0,33);
  strncpy(lclname,devname,32);
  if (devname[0] == '#') return atoi(&lclname[1]);
  if (el->EqpNumModules < MODULE_HASH_PAYOFF)
  {
    for (i=0; i<el->EqpNumModules; i++)
      if (!strncmp(lclname,el->EqpModuleName[i].name,32)) return i;
    return -device_not_connected;
  }
  if (el->EqpModuleXRef == NULL) return -device_not_connected;
  idx = (int)(ElfHash((unsigned char *)lclname) % MODULE_HASH_SIZE);
  for (xref=el->EqpModuleXRef[idx]; xref; xref=xref->next)
    if (!strncmp(lclname,el->EqpModuleName[xref->index].name,32)) return xref->index;
  return -device_not_connected;
}
int GetDeviceNumber(char *eqm,char *devname)
{
  return GetDeviceNumberEx(eqm,devname,NULL);
}
int GetNumContracts(char *eqname)
{
  int i,n=0;
  if (eqname == NULL || strlen(eqname) == 0) return ncontract;
  for (i=0; i<ncontract; i++)
   if (!strncmp(ContractList[i]->contract.EqpName,eqname,EQP_NAME_SIZE) && !ContractList[i]->expired) n++;
  return n;
}
int GetNumCallers(char *eqname)
{
  return GetNumConsumers(eqname);
}
int GetNumConsumers(char *eqname)
{
  int i,n=0;
  if (eqname == NULL || strlen(eqname) == 0) return nconsumer;
  for (i=0; i<ncontract; i++)
   if (!strncmp(ContractList[i]->contract.EqpName,eqname,EQP_NAME_SIZE) &&
       ContractList[i]->nconsumer > n) n = ContractList[i]->nconsumer;
  return n;
}
int GetPropertyId(char *eqm,char *prpName)
{
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  int idx;

  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL || prpName == NULL) return -1;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);
  for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
    if (!strncmp(prp->prpName,prpName,EQP_PROP_SIZE)) return (int)PROPERTY_NUM_INDEX(prp->prpId);
  return -1;
}
int RegisterDeviceName(char *eqname,char *devname,int devnr)
{
  ExportListStruct *el;
  HashXRefStruct *xref;
  char lclname[36];
  int cc=0,idx;

  for (el=ExportList; el != NULL; el=el->next)
    if (!strncmp(eqname,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);

  if (devnr < 0 || devnr >= el->EqpNumModules) ccerr(out_of_range);
  memset(lclname,0,33);
  strncpy(lclname,devname,32);
  strncpy(el->EqpModuleName[devnr].name,lclname,32);

  if (el->EqpNumModules < MODULE_HASH_PAYOFF) return 0;
  if (el->EqpModuleXRef == NULL) return 0;
  idx = (int)(ElfHash((unsigned char *)lclname) % MODULE_HASH_SIZE);
  for (xref=el->EqpModuleXRef[idx]; xref && xref->index != devnr; xref=xref->next);
  if (xref == NULL) /* it's a new one*/
  {
    if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) ccerr(out_of_server_memory);
    xref->index = devnr;
    xref->next = el->EqpModuleXRef[idx];
    el->EqpModuleXRef[idx] = xref;
  }
err:
  return cc;
}
int RegisterEquipmentModule(
    char *tagname,
    char *eqpname,
    int NumModules,
    int (*fcn)(char *,char *,DTYPE *,DTYPE *,short),
    void (*ini)(void),
    void (*tsk)(void),
    int rate,
    void (*exi)(void)
)
{
  int cc;
  ExportListStruct *el;
# ifdef VXWORKS
  char tasktag[16];
# endif
# ifdef MULTITHREADED
  BkgThreadFcn *btf;
# endif

  if (NumModules == 0) NumModules = 1;
  if ((cc=RegisterExport(tagname,eqpname,NumModules)) != 0) return cc;
  if ((el=getExportListItem(eqpname)) == NULL) return non_existent_elem;

  el->EqpFcn  = fcn;
  el->eqpBkg  = tsk;

# if defined(VXWORKS)
  if (tsk != NULL)
  {
    sprintf(tasktag,"%s_task",eqpname);
    if (taskSpawn(tasktag,VXWBKG_PRIORITY,VXWBKG_FLAGS,VXWBKG_STACK,(FUNCPTR)VXWbkgtsk,(int)tsk,(int)rate,0,0,0,0,0,0,0,0) == ERROR)
      cc = not_running;
    feclog("task %s : %s\n>",tasktag,cc ? "not spawned !" : "spawned");
  }
  el->eqpBkg  = NULL;        /* let VxWorks handle the task */
# elif defined(MULTITHREADED)
  if (tsk != NULL && useMultiThreadedBackgroundTasks)
  {
    feclog("FEC: module %s is using multi-threaded background tasks",eqpname);
    if ((btf=(BkgThreadFcn *)SystemCalloc(1,sizeof(BkgThreadFcn))) == NULL) return out_of_local_memory;
    btf->rate = rate;   btf->tsk = tsk;
    el->bkgbusy = TRUE;
    el->hEqpBkg = bkgCreateThread(btf);
    el->eqpBkg  = NULL;     /* let the pthread handle the task */
  }
# endif

  el->rateBkg = rate;
  if (el->eqpBkg && rate && ServerCycleMode != POLLING && rate < MaxPollingRate)
  {
    SysPoll = rate;
    MaxPollingRate = rate;
    if (rate < MinPollingRate) MinPollingRate = rate;
  }
  el->eqpIni = ini;
  el->eqpExi = exi;

  feclog("%12s task registration: %s %s %s",eqpname,
          ini ? "INIT" : "----",
          tsk ? "BCKG" : "----",
          exi ? "EXIT" : "----");

  if (SystemRunning && el->eqpIni) el->eqpIni();
  ExportsRegistered = 0;
  return 0;
}
int RegisterFecInformation(char *name,char *sub,char *cntxt,char *dsc,char *loc,char *hdw,char *resp,UINT16 poff)
{
  if (FecNameRegistered)
  {
    feclog("FEC Name %s already register for this server",gFecName);
    return unauthorised_action;
  }
  strncpy(gFecName,name,FEC_NAME_SIZE);
  strncpy(RPCFecEx.os,OS_TAG,12);
  strncpy(RPCFecEx.desc,dsc,64);
  strncpy(RPCFecEx.loc,loc,32);
  strncpy(RPCFecEx.hdw,hdw,32);
  strncpy(RPCFecEx.resp,resp,32);
  strncpy(gUserName,gFecName,16);
  PortOffset = poff;
  if (cntxt != NULL) strncpy(gDeviceContext,cntxt,EQP_CONTEXT_SIZE);
  if (sub != NULL) strncpy(RPCFecEx.sub,sub,16);
  FecNameRegistered = 1;
  feclog("FEC Name %s registered from API",name);
  return 0;
}
int RegisterFecNameEx(char *name,char *desc,char *os,char *loc,char *hdw,char *resp,UINT16 poff,char *context)
{
  char *c,sub[16] = "";
  if ((c=strchr(desc,'[')) != NULL) 
  {
    strncpy(sub,++c,16);
    if ((c=strchr(sub,']')) != NULL) *c = 0;
  }
  return RegisterFecInformation(name,sub,context,desc,loc,hdw,resp,poff);
}
int RegisterFecName(char *name,char *desc,char *os,char *loc,char *hdw,char *resp,UINT16 poff)
{
  return RegisterFecNameEx(name,desc,os,loc,hdw,resp,poff,NULL);
}
/* Appends property list for eqm ...*/
int RegisterPropertyInformation(char *eqm,char *prop,DTYPE *dout,DTYPE *din,short acc,short atype,UINT16 rowlength,char *dsc,int propId,char *rdr) 
{
  int i,idx,cc=0,isNew=0;
  long prpId = propId, prpId0 = -1;
  ExportPropertyListStruct *prp = NULL;
  ExportListStruct *el;
  PrpRedirBlk *prd;
  char strbuf[128];

  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  if ((i=strlen(prop)) > 4 && !strcmp(&prop[i-4],".NAM")) 
  {
    el->hasPropertyDeviceNames = TRUE;
    if (queryDeviceNamesFromProperties) acquireDeviceNamesFromProperties = TRUE;
  }
  if (strlen(prop) > 16 && longDeviceNames) ccerr(string_too_long);
  if (strchr(prop,'*')) idx = PROPERTY_HASH_SIZE;
  else idx = (int)(ElfHash((unsigned char *)prop) % PROPERTY_HASH_SIZE);

  for (prp=el->EqpProperty[idx]; prp; prp=prp->next) if (!strncmp(prp->prpName,prop,EQP_PROP_SIZE)) break;
  if (prp != NULL) 
  {
    prpId0 = prp->prpId;
    if (dout != NULL)
    {
      if (dout->dArrayLength != prp->prpSize || dout->dFormat != prp->prpFormat) isNew = -1;
    }
    else
    {
      if (prp->prpSize > 0) isNew = -1;
    }
    if (din != NULL)
    {
      if (din->dArrayLength != prp->prpSizeIn || din->dFormat != prp->prpFormatIn) isNew = -1;
    }
    else
    {
      if (prp->prpSizeIn > 0) isNew = -1;
    }
    if (acc != prp->prpAccessMode) isNew = -1;
  }
  else
  {
    isNew = -1;
  }

  if (!isNew) ccerr(0);

  if ((prp=(ExportPropertyListStruct *)SystemCalloc(1,sizeof(ExportPropertyListStruct))) == NULL) ccerr(out_of_server_memory);
  strncpy(prp->prpName,prop,EQP_PROP_SIZE);
  if ((i=getAliasTableIndexFromName(prop)) >= 0)
    strncpy(prp->prpAlias,gAliasTable[i].alias,EQP_PROP_SIZE);
  prp->next = el->EqpProperty[idx];
  el->EqpProperty[idx] = prp;
  if (prpId < 0) prpId = el->EqpNumProperties;
  prp->prpId = prpId0 >= 0 ? prpId0 : MAKE_PROPERTY_ID(prpId,idx);
  el->EqpNumProperties++;

  if (dout != NULL)
  {
    prp->prpSize = dout->dArrayLength;
    prp->prpFormat = dout->dFormat; 
    prp->prpArrayType = atype;
    if (!(prp->prpArrayType & AT_DOUBLE)) prp->prpArrayType |= prp->prpSize > 1 ? AT_SINGLE : AT_SCALAR;
    if (strlen(dout->dTag)) strncpy(prp->prpFormatTag.name,dout->dTag,8);
    if (rowlength > 0)
    {
      prp->rowSize = rowlength;
      prp->numRows = (UINT16)(prp->prpSize/rowlength);
    }
    else
    {
      prp->rowSize = (UINT16)prp->prpSize;
      prp->numRows = 1;
    }
  }
  if (din != NULL)
  {
    prp->prpSizeIn = din->dArrayLength;
    prp->prpFormatIn = din->dFormat;
    if (strlen(din->dTag)) strncpy(prp->prpFormatTagIn.name,din->dTag,8);
  }
  prp->prpAccessMode = acc;
  strncpy(strbuf,dsc,127); strbuf[127] = 0;
  ParsePropertyDescription(strbuf,prp);
  if (rdr != NULL && strlen(rdr) != 0)
  {
    if ((prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) ccerr(out_of_server_memory);
    prd->next = prp->prd;
    prp->prd = prd;
    strncpy(strbuf,rdr,127);
    ParseRedirector(strbuf,prp);
  }
err:
  return cc ? -cc : (int)PROPERTY_NUM_INDEX(prp->prpId);
}
int RegisterPropertyEx(char *eqm,char *prop,int siz,short fmt,short acc,char *dsc,int propId) /* Appends property list for eqm */
{
  DTYPE dout,din,*dinPtr = NULL;
  dout.dArrayLength = siz;
  dout.dFormat = fmt;
  dout.dTag[0] = 0;
  if (acc & CA_WRITE)
  {
    din.dArrayLength = siz;
    din.dFormat = fmt;
    din.dTag[0] = 0;
    dinPtr = &din;
  }
  return RegisterPropertyInformation(eqm,prop,&dout,dinPtr,acc,AT_UNKNOWN,0,dsc,propId,NULL);
}
int RegisterProperty(char *eqm,char *prop,int siz,short fmt,short acc,char *dsc) /* Appends property list for eqm */
{
  return RegisterPropertyEx(eqm,prop,siz,fmt,acc,dsc,-1);
}
short RegisterPropertyQueryFunction(char *eqm,int (*fcn)(char *devName,PropertyQueryStruct **pqs),int numprops)
{
  short cc=0;
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  el->EqpPropertyQueryFcn = fcn;
  el->EqpNumProperties = numprops;
err:
  return cc;
}
short RegisterXPropertyQueryFunction(char *eqm,int (*fcn)(char *devName,XPropertyQueryStruct **xpqs),int numprops)
{
  short cc=0;
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  el->XEqpPropertyQueryFcn = fcn;
  el->EqpNumProperties = numprops;
err:
  return cc;
}
int sendNetGlobal(char *keyword,DTYPE *din,void (*callback)(int,int),
		  int minPeriod,int maxPeriod,double tolerance)
{
  int i=0, cc=0, isRegistered=FALSE;
  UINT16 dsize;
  static int inside = 0;

  if (inside) return -semaphore_busy; else inside = -1;
  if (!BroadcastServer)
  {
    feclog("Broadcasting not enabled !");
    ccerr(un_allocated);
  }
  if (din == NULL || din->data.vptr == NULL) ccerr(argument_list_error);
  for (i=0; i<nbroadcasts; i++) if (!strcmp(BrdcstList[i]->keyword,keyword)) break;
  if (i == nbroadcasts)
  {
    if (nbroadcasts >= MaxNumGlobals) ccerr(resources_exhausted);
    if ((BrdcstList[i]=(GlobalListStruct *)SystemCalloc(1,sizeof(GlobalListStruct))) == NULL)
      ccerr(out_of_client_memory);
    nbroadcasts++;
    if (minPeriod <= 60) minPeriod *= 1000; /* convert to msec */
    if (maxPeriod <= 60) maxPeriod *= 1000;
    strncpy(BrdcstList[i]->keyword,keyword,16);
    BrdcstList[i]->size = (short)din->dArrayLength;
    BrdcstList[i]->format = din->dFormat;
    BrdcstList[i]->fcn = callback;
    BrdcstList[i]->maxPeriod = maxPeriod;
    BrdcstList[i]->minPeriod = minPeriod;
    BrdcstList[i]->tolerance = tolerance;
    dsize = BrdcstList[i]->size*fmtsizeof(BrdcstList[i]->format);
    if ((BrdcstList[i]->refdata=(BYTE *)SystemCalloc(1,dsize)) == NULL) ccerr(out_of_server_memory);
    BrdcstList[i]->time = time(NULL);
    feclog("GLOBAL %s (%d bytes) : %s",BrdcstList[i]->keyword,dsize,cc ? erlst[cc] : "registered");
  }
  else
  {
    isRegistered = TRUE;
  }
  dsize = BrdcstList[i]->size*fmtsizeof(BrdcstList[i]->format);
  BrdcstList[i]->data = (BYTE *)din->data.vptr;
  if (isRegistered)
  {
    BrdcstList[i]->tv.tv_sec = 0;
    doBroadcast(&thisCycleTime);       /* send it out immediately if already registered */
  }
err:
  inside = 0;
  return cc ? -cc : i;
}
int SetNameServerAddress(char *ens)
{
  int i, n = 1, poff, cc = 0;
  char s[256],*c, *c1, *ip, *po;
  DWORD haddr;

  if (gotNameServerAddress) return 0;
  if (ens == NULL) ccerr(argument_list_error);
  strncpy(s,ens,256);
  for (i=0; i<256; i++) if (s[i] == ',') n++;
  if ((NameServer=(RPCFecStruct *)SystemCalloc(n,sizeof(RPCFecStruct))) == NULL) ccerr(out_of_server_memory);
  for (i=0, c=strtok(s,","); i<n && c != NULL; i++, c=strtok(NULL,","))
  {
    ip = c; po = NULL;
    if ((c1=strchr(c,':')) != NULL)
    {
      *c1++ = 0; po = c1;
    }
    haddr = inet_addr(ip);
    sprintf(NameServer[i].Name,"ENS#%d",i);
    strncpy(NameServer[i].IP,ip,16);
    memcpy(NameServer[i].IPh_addr,&haddr,4);
    NameServer[i].Prot = UDP;
    if (po != NULL) 
    {
      poff = atoi(po);
      NameServer[i].Port[0] = poff&0xff;
      NameServer[i].Port[1] = (poff>>8)&0xff;
    }
  }
  numNameServers = i;
  gotNameServerAddress = 1;
err:
  return cc;
}
void SetRPCCompletion(char *errstr)
{
  memset(RPCLastErrorString,0,RPCERR_SIZE);
  if (errstr != NULL && strlen(errstr) > 0)
    strncpy(RPCLastErrorString,errstr,RPCERR_SIZE-1);
}
void SetRPCRedirection(char *rdrSrv,char *rdrPrp)
{
  memset(tRedirectedServer,0,64);
  if (rdrSrv != NULL) memcpy(tRedirectedServer,rdrSrv,16);
  if (rdrPrp != NULL) memcpy(&tRedirectedServer[16],rdrPrp,16);
}
void SetlookupRedirectionNameStub(int (*fcn)(char *eqm,char *prpName,char *devName))
{
  lookupRedirectionNameStub = fcn;
}
void SetAppVersion(int major,int minor,int revision)
{
  sprintf(appmkversion,"%d.%02d.%04d",major,minor,revision);
  return;
}
void SetAppDate(time_t appdate)
{
  appmktime = appdate;
  return;
}
